﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Algorithms.Utils;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Factorizing the observed rating values using a feature matrix for users and one for items.
    /// This class can update the factorization online.
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    public class MatrixFactorization : RatingPredictor
    {
        Matrix<double> user_feature;
        Matrix<double> item_feature;
        double bias;

        /// <summary>
        /// Feature count
        /// </summary>
        public int feature_count = 10;
        /// <summary>
        /// Learn rate
        /// </summary>
        public double learn_rate = 0.01;
        /// <summary>
        /// Regularization
        /// </summary>
        public double regularization = 0.015;
        /// <summary>
        /// F mean
        /// </summary>
        public double init_f_mean = 0;
        /// <summary>
        /// F standard deviation
        /// </summary>
        public double init_f_stdev = 0.01;
        /// <summary>
        /// Iterations
        /// </summary>
        public int iter = 30;

        /// <summary>
        /// Item features
        /// </summary>
        public bool update_item_features = true;
        /// <summary>
        /// User features
        /// </summary>
        public bool update_user_features = true;

        Data data;

        /// <inheritdoc />
        protected override void _Train()
        {            
            // (1.1) create data structure
            data = new Data(true, true, true, this.max_data_value, this.min_data_value);
            data.add(backend.GetRelation(RelationType.Rated));
            
            // (1.2) Init feature matrizes
            user_feature = new Matrix<double>((int) data.userRatings.Count, feature_count);
            item_feature = new Matrix<double>((int) data.itemRatings.Count, feature_count);
            MatrixUtils.initNormal(user_feature, init_f_mean, init_f_stdev);
            MatrixUtils.initNormal(item_feature, init_f_mean, init_f_stdev);

            // (2) learn model parameters
            bias = data.allRatings.avg();
            learnFeatures(data.allRatings, true, true);

            // (3) release All-index
            data.release(data.allRatings);
        }

        /// <summary>
        /// Retrains the user.
        /// </summary>
        /// <param name="user_id">The user_id.</param>
        protected void retrainUser(int user_id)
        {
            if (!update_user_features) { return; }
            MatrixUtils.initNormal(user_feature, init_f_mean, init_f_stdev, user_id);
            learnFeatures(data.userRatings[(int)user_id], true, false);
        }

        /// <summary>
        /// Retrains the item.
        /// </summary>
        /// <param name="item_id">The item_id.</param>
        protected void retrainItem(int item_id)
        {
            if (!update_item_features) { return; }
            MatrixUtils.initNormal(item_feature, init_f_mean, init_f_stdev, item_id);
            learnFeatures(data.itemRatings[(int)item_id], false, true);
        }
        
        private void learnFeatures(Ratings rating_set, bool update_user, bool update_item)
        {
            for (int current_iter = 0; current_iter < iter; current_iter++)
            {
                foreach (RatingEvent rating in rating_set.items)
                {

                    int u = rating.user_id;
                    int i = rating.item_id;
                    double r = rating.rating;

                    double p = predictRating(u, i, false);
                    double err = r - p;

                    // Adjust features
                    for (int f = 0; f < feature_count; f++)
                    {
                        double u_f = user_feature.get(u, f);
                        double i_f = item_feature.get(i, f);

                        double delta_u = (err * i_f - regularization * u_f);
                        double delta_i = (err * u_f - regularization * i_f);

                        if (update_user) MatrixUtils.inc(user_feature, u, f, learn_rate * delta_u);
                        if (update_item) MatrixUtils.inc(item_feature, i, f, learn_rate * delta_i);
                    }
                }
            }
        }

        /// <inheritdoc />
        protected double predictRating(int user_id, int item_id, bool bound)
        {
            double result = bias;

            // U*V
            for (int f = 0; f < feature_count; f++) {
                result += user_feature.get(user_id, f) * item_feature.get(item_id, f);
            }

            if (bound) {
                if (result > data.max_value) result = data.max_value;
                if (result < data.min_value) result = data.min_value;
            }
            return result;
        }

        /// <inheritdoc />
        public override double PredictRating(int user_id, int item_id)
        {
            return predictRating((int)user_id, (int)item_id, true);
        }

        /// <summary>
        /// </summary>
        /// <param name="user_id"></param>
        /// <param name="item_id"></param>
        /// <param name="rating"></param>
        /// <inheritdoc/>
        public override void AddRating(int user_id, int item_id, double rating)
        {
            RatingEvent r = new RatingEvent();
            r.user_id = (int)user_id;
            r.item_id = (int)item_id;
            r.rating = rating;
            data.addRating(r);
         
            retrainUser(r.user_id);   
            retrainItem(r.item_id);
        }

        /// <summary>
        /// </summary>
        /// <param name="user_id"></param>
        /// <param name="item_id"></param>
        /// <param name="rating"></param>
        /// <inheritdoc/>
        public override void UpdateRating(int user_id, int item_id, double rating) 
        {
            RatingEvent r = data.findRating((int)user_id, (int)item_id);
            if (r == null)
            {
                throw new Exception("rating not found");
            }
            data.changeRating(r, rating);
            
            retrainUser(r.user_id);
            retrainItem(r.item_id);
        }

        /// <summary>
        /// </summary>
        /// <param name="user_id"></param>
        /// <param name="item_id"></param>
        /// <inheritdoc/>
        public override void RemoveRating(int user_id, int item_id) 
        {
            RatingEvent r = data.findRating((int)user_id, (int)item_id);
            if (r == null)
            {
                throw new Exception("rating not found");
            }
            data.removeRating(r);

            retrainUser(r.user_id);
            retrainItem(r.item_id);
        }


        /// <summary>
        /// </summary>
        /// <param name="user_id"></param>
        /// <inheritdoc/>
        public override void AddUser(int user_id) 
        {
            // (1) add user to Ratings-Data storage
            data.addUser(user_id);
            // (2) add user-feature vector and initialize with small numbers
            // (2) add item-feature vector and initialize with small numbers        
            int old_c = user_feature.dim2;
            user_feature.addRow(data.userRatings.Count);
            for (int i = old_c; i < user_feature.dim2; i++)
            {
                MatrixUtils.initNormal(user_feature, init_f_mean, init_f_stdev, user_id);
            }
        }

        /// <summary>
        /// </summary>
        /// <param name="item_id"></param>
        /// <inheritdoc/>
        public override void AddItem(int item_id)
        {
            // (1) add item to Ratings-Data storage
            data.addItem(item_id);
            // (2) add item-feature vector and initialize with small numbers        
            int old_c = item_feature.dim2;
            item_feature.addRow(data.itemRatings.Count);
            for (int i = old_c; i < item_feature.dim2; i++)
            {
                MatrixUtils.initNormal(item_feature, init_f_mean, init_f_stdev, item_id);
            }
        }

        /// <summary>
        /// </summary>
        /// <param name="user_id"></param>
        /// <inheritdoc/>
        public override void RemoveUser(int user_id)
        {
            // (1) remove all ratings of this user from Ratings-Data storage
            data.removeUser(user_id);
            // (2) initialize user-features with small numbers
            MatrixUtils.initNormal(user_feature, init_f_mean, init_f_stdev, user_id);
        }

        /// <summary>
        /// </summary>
        /// <param name="item_id"></param>
        /// <inheritdoc/>
        public override void RemoveItem(int item_id) 
        {
            // (1) remove all ratings of this item from Ratings-Data storage
            data.removeItem(item_id);
            // (2) initialize item-features with small numbers
            MatrixUtils.initNormal(item_feature, init_f_mean, init_f_stdev, item_id);
        }
    }
}
