﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{
    /// <summary>
    /// Value object for Correlation data.
    /// </summary>
    public class Correlation {
        private double correlation;
        private int sampleSize;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="correlation">Correlation</param>
        /// <param name="sampleSize">Size of the sample on which the correlation was based</param>
        public Correlation(double correlation, int sampleSize) {
            this.correlation = correlation;
            this.sampleSize = sampleSize;
        }

        /// <summary>
        /// Gets the correlation factor
        /// </summary>
        public double CorrelationFactor {
            get
            {
                return correlation;
            }
        }

        /// <summary>
        /// Gets the sample size
        /// </summary>
        public int SampleSize {
            get
            {
                return sampleSize;
            }
        }
    }

    /// <summary>
    /// Object that represents a rating of a content item. A rating can not be instantiated directly.
    /// </summary>
    public class Rating {

        private float value;

        /// <summary>
        /// Constructor
        /// </summary>
	    public Rating() {}

        /// <summary>
        /// Constructor is protected. Ratings are constructed through the RatingModel interface.
        /// </summary>
        /// <param name="userId">the id of the user who gave this rating</param>
        /// <param name="itemId">the id of the content item that the rating has been given to</param>
        /// <param name="value">the actual rating value</param>
	    public Rating(int userId, int itemId, float value) {
		    UserId = userId;
		    ItemId = itemId;
		    Value = value;
	    }

        /// <summary>
        /// Gets/sets the user Id
        /// </summary>
        public int UserId { get; set; }
            
        /// <summary>
        /// Gets/sets the item id
        /// </summary>
        public int ItemId { get; set; }

        /// <summary>
        /// Gets/sets the value (rating)
        /// </summary>
        public float Value { 
            get {
                return this.value;
            }
            set {
		        if (float.IsNaN(value)) {
			        value = 0.0f;
		        }
		        this.value = value;
            }
        }

        /// <inheritdoc />
        public override string ToString()
        {
		    StringBuilder sb = new StringBuilder();
		    sb.Append("Rating(user=");
		    sb.Append(UserId);
		    sb.Append(", item=");
		    sb.Append(ItemId);
		    sb.Append(", value=");
		    sb.Append(")");
		    return sb.ToString();
	    }

        /// <summary>
        /// Gets the hash code
        /// </summary>
        /// <returns>The hash code for this instance</returns>
        public override int GetHashCode()
        {
		    const int prime = 31;
		    int result = 1;
		    result = prime * result + ItemId.GetHashCode();
            result = prime * result + UserId.GetHashCode();
		    return result;
	    }

        /// <inheritdoc />
        public override bool Equals(object obj)
        {
		    if (this == obj) return true;
		    if (obj == null) return false;
		    if (!(obj is Rating)) return false;

            Rating other = (Rating) obj;
		    if (ItemId != other.ItemId)
			    return false;
		    if (UserId != other.UserId)
			    return false;
		    return true;
	    }
    }

    /// <summary>
    /// Abstract base class for storage of a collection of Ratings. This class maintains average and stddev values automatically when the collection is
    /// updated.
    /// </summary>
    public abstract class AbstractRatings {

        /// <summary>
        /// Flag if average and stdev need to be recomputed
        /// </summary>
        protected bool stale = false;

        /// <summary>
        /// Maintains (caches) the average value of the ratings. Is recalculated upon addition of new ratings.
        /// </summary>
	    private double average = 0.0;

        /// <summary>
        /// Maintains (caches) the standard deviation of the ratings. Is recalculated upon addition of new ratings.
        /// </summary>
	    private double stdev = 0.0;

        /// <summary>
        /// Constructor
        /// </summary>
	    public AbstractRatings() {
	    }

        /// <summary>
        /// Adds a collection of ratings to this collection. An existing rating for an item will be replaced. The values of the average and stdev are
	    /// automatically updated.
        /// </summary>
        /// <param name="ratings">The ratings to add</param>
	    public abstract void Add(List<Rating> ratings);

        /// <summary>
        /// Adds a rating to the collection. An existing rating for an item will be replaced. The values of the average and stdev are automatically 
        /// updated.
        /// </summary>
        /// <param name="rating">The rating to store</param>
        /// <returns>The previous value associated with the rating, or null if there was no mapping for the rating</returns>
        public abstract Rating Add(Rating rating);

        /// <summary>
        /// Calculate the average and the standard deviation of the ratings in this collection.
        /// </summary>
	    protected void calculate() {
		    double[] values = new double[getNoRatings()];
            int i = 0;
            foreach (Rating r in getRatings()) {
			    values[i] = r.Value;
			    i++;
		    }
		    average = Statistics.avg(values);
		    stdev = Statistics.stdvp(values);

            stale = false;
	    }

        /// <summary>
        /// Gets the average of the ratings in this cache.
        /// </summary>
        /// <returns>The standard deviation</returns>
        public double getAverage() {
            if (stale) calculate();
		    return average;
	    }

        /// <summary>
        /// Gets the standard deviation of the ratings in this cache.
        /// </summary>
        /// <returns>The standard deviation</returns>
	    public double getStdev() {
            if (stale) calculate();
            return stdev;
	    }

        /// <summary>
        /// Gets the number of ratings in this collection.
        /// </summary>
        /// <returns>The number of Ratings</returns>
	    public abstract int getNoRatings();

        /// <summary>
        /// Gets all ratings in this cache.
        /// </summary>
        /// <returns>A collection with all ratings in this cache. If no ratings are available then an empty collection is returned.</returns>
	    public abstract List<Rating> getRatings();
    }

    /// <summary>
    /// Storage of a collection of Ratings
    /// </summary>
    public class RatingList : AbstractRatings
    {
        /// <summary>
        /// A list with the ratings.
        /// </summary>
        protected List<Rating> ratings = new List<Rating>();

        /// <summary>
        /// Creates a new Ratings instance.
        /// </summary>
        public RatingList()
        {
        }

        /// <summary>
        /// Gets the number of ratings in this collection.
        /// </summary>
        /// <returns>The number of Ratings</returns>
        public override int getNoRatings()
        {
            return ratings.Count;
        }

        /// <summary>
        /// Gets all ratings in this list
        /// </summary>
        /// <returns>A collection with all ratings in this list. If no ratings are available then an empty collection is returned</returns>
        public override List<Rating> getRatings()
        {
            return ratings;
        }

        /// <inheritdoc />
        public override void Add(List<Rating> ratings)
        {
            this.ratings.AddRange(ratings);
        }

        /// <inheritdoc />
        public override Rating Add(Rating rating)
        {
            ratings.Add(rating);
            return rating;
        }

        /// <summary>
        /// Remove the specified rating
        /// </summary>
        /// <param name="rating">Rating to remove</param>
        public void Remove(Rating rating)
        {
            ratings.Remove(rating);
        }
    }

    /// <summary>
    /// Abstract base class for storage of a collection of Ratings for an entity
    /// </summary>
    public abstract class EntityRatings : AbstractRatings
    {
        /// <summary>
        /// Entity ID for all the ratings
        /// </summary>
        protected int entityId;

        /// <summary>
        /// A dictionary with the entityIds and associated ratings.
        /// </summary>
        protected Dictionary<int, Rating> indexedRatings = new Dictionary<int, Rating>();

        /// <summary>
        /// Creates a new EntityRatings instance.
        /// </summary>
        /// <param name="entityId">The entity id</param>
        public EntityRatings(int entityId)
        {
            this.entityId = entityId;
        }

        /// <summary>
        /// Removes a rating by the given entity. The values of the average and stdev are automatically updated.
        /// </summary>
        /// <param name="entityId">The id of the entity to remove the rating for</param>
        /// <returns>Rating The previous value associated with the specified entity, or null if there was no mapping for the id</returns>
        public Rating Remove(int entityId)
        {
            Rating result = null;
            if (indexedRatings.ContainsKey(entityId))
            {
                result = indexedRatings[entityId];
                indexedRatings.Remove(entityId);
                stale = true;
            }
            return result;
        }

        /// <summary>
        /// Gets the number of ratings in this collection.
        /// </summary>
        /// <returns>The number of Ratings</returns>
        public override int getNoRatings()
        {
            return indexedRatings.Count;
        }

        /// <summary>
        /// Gets all ratings in this list
        /// </summary>
        /// <returns>A collection with all ratings in this list. If no ratings are available then an empty collection is returned</returns>
        public override List<Rating> getRatings()
        {
            return indexedRatings.Values.ToList();
        }
    }

    /// <summary>
    /// Class that holds all ratings of an item, with a fast lookup by userId function for ratings.
    /// </summary>
    public class ItemRatings : EntityRatings
    {

        /// <summary>
        /// Creates a new ItemRatings instance.
        /// </summary>
        /// <param name="itemId">The item id</param>
	    public ItemRatings(int itemId) : base (itemId) {
	    }

        /// <summary>
        /// Adds a rating to the collection. An existing rating for an item will be replaced. The values of the average and stdev are 
        /// automatically updated.
        /// </summary>
        /// <param name="rating">The rating to store</param>
        /// <returns>The replaced rating (if any)</returns>
        public override Rating Add(Rating rating)
        {
            Rating result = null;
            if (indexedRatings.ContainsKey(rating.UserId)) result = indexedRatings[rating.UserId];
            indexedRatings[rating.UserId] = rating;
            stale = true;
            return result;
        }

        /// <summary>
        /// Adds a collection of ratings to this collection. An existing rating for an item will be replaced. The values of the average and stdev 
        /// are automatically updated.
        /// </summary>
        /// <param name="ratings">The ratings to add</param>
        public override void Add(List<Rating> ratings)
        {
            foreach (Rating rating in ratings)
            {
                indexedRatings[rating.UserId] = rating;
            }
            stale = true;
        }

        /// <summary>
	    /// Lists all users that have rated the item.
        /// </summary>
        /// <returns>A Set with all users. If no users are available then an empty set is returned</returns>
	    public List<int> getUserIds() {
            return indexedRatings.Keys.ToList(); 
	    }

        /// <summary>
        /// Retrieves the rating object associated with the user.	     
        /// </summary>
        /// <param name="userId">The user to lookup the rating for</param>
        /// <returns>The rating object for the user. If not available then null is returned</returns>
	    public Rating getRating(int userId) {
            if (indexedRatings.ContainsKey(userId))
            {
                return indexedRatings[userId];
            }
            return null;
        }
        
        /// <summary>
        /// Gets the item Id
        /// </summary>
        /// <returns>The item Id</returns>
	    public int getItemId() {
		    return entityId;
	    }

        /// <summary>
        /// Get a string representation of this instance
        /// </summary>
        /// <returns>A string representation of this instance</returns>
        public override string ToString()
        {
		    StringBuilder sb = new StringBuilder();
		    sb.Append("itemId=").Append(entityId);
		    sb.Append(", noRatings=").Append(getNoRatings());
		    sb.Append(", average=").Append(getAverage());
		    sb.Append(", stdev=").Append(getStdev());
		    return sb.ToString();
	    }

        /// <summary>
        /// Compares this object to another object
        /// </summary>
        /// <param name="obj">Object to compare with</param>
        /// <returns>True if the objects are equal, false otherwise</returns>
        public override bool Equals(object obj)
        {
		    if (this == obj)
			    return true;
		    if (obj == null)
			    return false;
		    if (!(obj is ItemRatings))
			    return false;
		    ItemRatings other = (ItemRatings) obj;
		    if (entityId != other.getItemId())
			    return false;
		    return true;
	    }

        /// <summary>
        /// Gets the hash code
        /// </summary>
        /// <returns>The hash code for this instance</returns>
        public override int GetHashCode()
        {
            const int prime = 31;
            int result = 1;
            result = prime * result + entityId.GetHashCode();
            return result;
        }
    }

    /// <summary>
    /// Class that holds all ratings of a user, with a fast lookup by entityId function for ratings.
    /// </summary>
    public class UserRatings : EntityRatings
    {

        /// <summary>
        /// Creates a new UserRatings instance.
        /// </summary>
        /// <param name="userId">The user id</param>
        public UserRatings(int userId)
            : base(userId)
        {
        }

        /// <summary>
        /// Adds a rating to the collection. An existing rating for an item will be replaced. The values of the average and stdev are 
        /// automatically updated.
        /// </summary>
        /// <param name="rating">The rating to store</param>
        /// <returns>The replaced rating (if any)</returns>
        public override Rating Add(Rating rating)
        {
            Rating result = null;
            if (indexedRatings.ContainsKey(rating.ItemId)) result = indexedRatings[rating.ItemId];
            indexedRatings[rating.ItemId] = rating;
            stale = true;
            return result;
        }

        /// <summary>
        /// Adds a collection of ratings to this collection. An existing rating for an item will be replaced. The values of the average and stdev 
        /// are automatically updated.
        /// </summary>
        /// <param name="ratings">The ratings to add</param>
        public override void Add(List<Rating> ratings)
        {
            foreach (Rating rating in ratings)
            {
                indexedRatings[rating.ItemId] = rating;
            }
            stale = true;
        }

        /// <summary>
        /// Lists all items that have been rated by this user
        /// </summary>
        /// <returns>A Set with all items. If no items are available then an empty set is returned</returns>
        public List<int> getItemIds()
        {
            return indexedRatings.Keys.ToList(); ;
        }

        /// <summary>
        /// Retrieves the rating object associated with the item.	     
        /// </summary>
        /// <param name="itemId">The user to lookup the rating for</param>
        /// <returns>The rating object for the item. If not available then null is returned</returns>
        public Rating getRating(int itemId)
        {
            if (indexedRatings.ContainsKey(itemId))
            {
                return indexedRatings[itemId];
            }
            return null;
        }

        /// <summary>
        /// Gets an iterator over the union of items rated in this and the specified UserRatings instance
        /// </summary>
        /// <param name="otherUserRatings">Other UserRatings instance</param>
        /// <returns>An iterator over the union of rated item IDs</returns>
        public IEnumerator<int> getKeyIteratorUnion(UserRatings otherUserRatings)
        {
            // need to merge the key sets
            return otherUserRatings.getItemIds().Union(otherUserRatings.getItemIds()).GetEnumerator();
        }

        /// <summary>
        /// Gets the user Id
        /// </summary>
        /// <returns>The user Id</returns>
        public int getUserId()
        {
            return entityId;
        }

        /// <summary>
        /// Get a string representation of this instance
        /// </summary>
        /// <returns>A string representation of this instance</returns>
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("userId=").Append(entityId);
            sb.Append(", noRatings=").Append(getNoRatings());
            sb.Append(", average=").Append(getAverage());
            sb.Append(", stdev=").Append(getStdev());
            return sb.ToString();
        }

        /// <summary>
        /// Compares this object to another object
        /// </summary>
        /// <param name="obj">Object to compare with</param>
        /// <returns>True if the objects are equal, false otherwise</returns>
        public override bool Equals(object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (!(obj is UserRatings))
                return false;
            UserRatings other = (UserRatings)obj;
            if (entityId != other.getUserId())
                return false;
            return true;
        }

        /// <summary>
        /// Gets the hash code
        /// </summary>
        /// <returns>The hash code for this instance</returns>
        public override int GetHashCode()
        {
            const int prime = 31;
            int result = 1;
            result = prime * result + entityId.GetHashCode();
            return result;
        }
    }

    /// <summary>
    /// Rating model interface
    /// </summary>
    public interface IRatingModel
    {
        /// <summary>
        /// Gets/sets the IEntityRelationDataProvider to be used by this model
        /// </summary>
        IEntityRelationDataProvider EntityRelationDataProvider { get; set; }

        /// <summary>
        /// Fetches a collection containing all ratings available on <c>itemId</c>, keyed by the users.
        /// </summary>
        /// <param name="itemId">The item</param>
        /// <returns>A ItemRatings object, containing all ratings for the item, indexed by the users</returns>
        ItemRatings getItemRatings(int itemId);

        /// <summary>
        /// Fetches a collection containing all ratings available on <c>userId</c>, keyed by the rated items.
        /// </summary>
        /// <param name="userId">The user</param>
        /// <returns>A ItemRatings object, containing all ratings for the user, indexed by the items</returns>
        UserRatings getUserRatings(int userId);

        /// <summary>
        /// Fetches a collection containing all ratings available.
        /// </summary>
        /// <returns>A Ratings object containing all ratings</returns>
        RatingList getAllRatings();

        /// <summary>
        /// Gets the list of items that both of the specified users have rated
        /// </summary>
        /// <param name="userIdA">User A</param>
        /// <param name="userIdB">User B</param>
        /// <returns>List of items rated by user A and user B</returns>
        List<int> getItemsRatedByBoth(int userIdA, int userIdB);

        /// <summary>
        /// Gets an enumerator over the set of existing users
        /// </summary>
        /// <returns>Enumerator over the set of existing users</returns>
        IEnumerator<int> getUserEnumerator();

        /// <summary>
        /// Gets an enumerator over the set of existing items
        /// </summary>
        /// <returns>Enumerator over the set of existing items</returns>
        IEnumerator<int> getItemEnumerator();

        /// <summary>
        /// Gets an enumerator over the set of existing ratings
        /// </summary>
        /// <returns>Enumerator over the set of existing ratings</returns>
        IEnumerator<Rating> getRatingEnumerator();

        /// <summary>
        /// Computes the correlation between the ratings of two users
        /// </summary>
        /// <param name="userId">User Id</param>
        /// <param name="otherUserId">Id of the user to correlate with</param>
        /// <returns>The correlation between the ratings of userId and otherUserId</returns>
        Correlation getUserCorrelation(int userId, int otherUserId);

        /// <summary>
        /// Computes the correlation between two sets of ratings
        /// </summary>
        /// <param name="ratingsA">Rating set A</param>
        /// <param name="ratingsB">Rating set B</param>
        /// <returns>The correlation between rating set A and rating set B</returns>
        Correlation getCorrelation(UserRatings ratingsA, UserRatings ratingsB);

        /// <summary>
        /// Gets the number of users
        /// </summary>
        /// <returns>Number of users</returns>
        int getUserCount();

        /// <summary>
        /// Gets the number of items
        /// </summary>
        /// <returns>Number of items</returns>
        int getItemCount();

        /// <summary>
        /// Gets the minimum rating value
        /// </summary>
        float MaxValue { get; }

        /// <summary>
        /// Gets the maximum rating value
        /// </summary>
        float MinValue { get; }

        /// <summary>
        /// Add rating
        /// </summary>
        /// <param name="userId">User Id</param>
        /// <param name="itemId">Item Id</param>
        /// <param name="rating">Rating</param>
        void addRating(int userId, int itemId, float rating);

        /// <summary>
        /// Remove rating
        /// </summary>
        /// <param name="userId">User Id</param>
        /// <param name="itemId">Item Id</param>
        void removeRating(int userId, int itemId);

        /// <summary>
        /// Remove all ratings for a user
        /// </summary>
        /// <param name="userId">User Id</param>
        void removeUser(int userId);

        /// <summary>
        /// Remove all ratings for an item
        /// </summary>
        /// <param name="itemId">Item id</param>
        void removeItem(int itemId);

        /// <summary>
        /// Get a particular rating
        /// </summary>
        /// <param name="userId">User ID</param>
        /// <param name="itemId">Item ID</param>
        /// <returns>Given rating for that user and item, or null if no such rating exists</returns>
        Rating getRating(int userId, int itemId);
    }

    /// <summary>
    /// Dummy rating model
    /// </summary>
    abstract class DummyRatingModel : IRatingModel
    {
        public DummyRatingModel()
        {
        }

        public IEntityRelationDataProvider EntityRelationDataProvider { get; set; }

        public void initialize()
        {
        }

        #region IReadOnlyRatingModel Members

        public ItemRatings getItemRatings(int itemId)
        {
            return new ItemRatings(itemId);
        }

        public UserRatings getUserRatings(int userId)
        {
            return new UserRatings(userId);
        }

        public List<int> getItemsRatedByBoth(int userIdA, int userIdB)
        {
            return new List<int>();
        }

        public Correlation getUserCorrelation(int userId, int otherUserId)
        {
            return new Correlation(0, 0);
        }

        public Correlation getCorrelation(UserRatings ratingsA, UserRatings ratingsB)
        {
            return new Correlation(0, 0);
        }

        public IEnumerator<Rating> getRatingEnumerator() {
            return new RatingList().getRatings().GetEnumerator();
        }

        public IEnumerator<int> getUserEnumerator()
        {
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.User);
            for (dr.Open(); dr.Read(); )
            {
                yield return dr.GetInt32(0);
            }
        }

        public IEnumerator<int> getItemEnumerator()
        {
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.CatalogItem);
            for (dr.Open(); dr.Read(); )
            {
                yield return dr.GetInt32(0);
            }
        }

        public int getUserCount()
        {
            // lazy impl
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.User);
            int count = 0;
            for (dr.Open(); dr.Read(); ) count++;

            return count;
        }

        public int getItemCount()
        {
            // lazy impl
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.CatalogItem);
            int count = 0;
            for (dr.Open(); dr.Read(); ) count++;

            return count;
        }

        public RatingList getAllRatings()
        {
            return new RatingList();
        }

        public float MaxValue
        {
            get
            {
                return 0f;
            }
        }

        public float MinValue
        {
            get
            {
                return 0f;
            }
        }
        #endregion

        public void addRating(int userId, int itemId, float rating)
        {
        }

        public void removeRating(int userId, int itemId)
        {
        }

        public void removeUser(int userId)
        {
        }

        public void removeItem(int itemId) 
        {
        }

        public Rating getRating(int userId, int itemId)
        {
            return null;
        }
    }

    class DuineRatingModel : IRatingModel
    {
        private float min;
        private float max;

        public DuineRatingModel(float min, float max)
        {
            this.min = min;
            this.max = max;
        }

        public IEntityRelationDataProvider EntityRelationDataProvider { get; set; }

        private Dictionary<int, UserRatings> userRatingsCache = new Dictionary<int, UserRatings>();
        private Dictionary<int, ItemRatings> itemRatingsCache = new Dictionary<int, ItemRatings>();
        private RatingList ratings = new RatingList();

        public void initialize()
        {
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetRelation(RelationType.Rated);
            for (dr.Open(); dr.Read(); )
            {
                addRating(dr.GetInt32(0), dr.GetInt32(1), (float)dr.GetDouble(2));
            }
        }

        public void addRating(int userId, int itemId, float rating)
        {

            Rating r = new Rating(userId, itemId, rating);
            ratings.Add(r);

            ItemRatings itemRatings;
            if (itemRatingsCache.ContainsKey(itemId))
            {
                itemRatings = itemRatingsCache[itemId];
            }
            else
            {
                itemRatings = new ItemRatings(itemId);
                itemRatingsCache[itemId] = itemRatings;
            }
            itemRatings.Add(r);

            UserRatings userRatings;
            if (userRatingsCache.ContainsKey(userId))
            {
                userRatings = userRatingsCache[userId];
            }
            else
            {
                userRatings = new UserRatings(userId);
                userRatingsCache[userId] = userRatings;
            }
            userRatings.Add(r);
        }

        public void removeRating(int userId, int itemId)
        {
            if (itemRatingsCache.ContainsKey(itemId))
            {
                Rating r = itemRatingsCache[itemId].Remove(userId);
                if (r != null) ratings.Remove(r);
            }
            if (userRatingsCache.ContainsKey(userId))
            {
                Rating r = userRatingsCache[userId].Remove(itemId);
                if (r != null) ratings.Remove(r);
            }
        }

        public void removeUser(int userId) {
            if (userRatingsCache.ContainsKey(userId))
            {
                UserRatings ratings = userRatingsCache[userId];
                userRatingsCache.Remove(userId);
                foreach (Rating r in ratings.getRatings()) {
                    removeRating(userId, r.ItemId);
                }
            }
        }

        public void removeItem(int itemId) {
            if (itemRatingsCache.ContainsKey(itemId))
            {
                ItemRatings ratings = itemRatingsCache[itemId];
                itemRatingsCache.Remove(itemId);
                foreach (Rating r in ratings.getRatings()) {
                    removeRating(r.UserId, itemId);
                }
            }
        }


        #region IReadOnlyRatingModel Members

        public ItemRatings getItemRatings(int itemId)
        {
            if (itemRatingsCache.ContainsKey(itemId))
            {
                return itemRatingsCache[itemId];
            }
            return new ItemRatings(itemId);
        }

        public UserRatings getUserRatings(int userId)
        {
            if (userRatingsCache.ContainsKey(userId))
            {
                return userRatingsCache[userId];
            }

            return new UserRatings(userId);
        }

        public List<int> getItemsRatedByBoth(int userIdA, int userIdB)
        {
            UserRatings ratingsA = getUserRatings(userIdA);
            UserRatings ratingsB = getUserRatings(userIdB);
            HashSet<int> setA = new HashSet<int>();
            foreach (int itemId in ratingsA.getItemIds())
            {
                setA.Add(itemId);
            }
            List<int> result = new List<int>();
            foreach (int itemId in ratingsB.getItemIds())
            {
                if (setA.Contains(itemId))
                {
                    result.Add(itemId);
                    setA.Remove(itemId); // this prevents duplicates
                }
            }
            return result;
        }

        public Correlation getUserCorrelation(int userId, int otherUserId)
        {
            UserRatings userRatings = getUserRatings(userId);
            UserRatings otherUserRatings = getUserRatings(otherUserId);

            return getCorrelation(userRatings, otherUserRatings);
        }

        public Correlation getCorrelation(UserRatings ratingsA, UserRatings ratingsB)
        {
            Correlation c = null;
            List<double> aResult = new List<double>();
            List<double> bResult = new List<double>();

            foreach (Rating rating in ratingsA.getRatings())
            {
                Rating otherRating = ratingsB.getRating(rating.ItemId);
                if (otherRating != null)
                {
                    aResult.Add(rating.Value);
                    bResult.Add(otherRating.Value);
                }
            }
            int n = aResult.Count;
            if (n > 0)
            {
                c = new Correlation(Statistics.pearson(aResult.ToArray(), bResult.ToArray()), n);
            }
            return c;
        }

        public IEnumerator<int> getUserEnumerator()
        {
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.User);
            for (dr.Open(); dr.Read(); )
            {
                yield return dr.GetInt32(0);
            }
        }

        public IEnumerator<int> getItemEnumerator()
        {
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.CatalogItem);
            for (dr.Open(); dr.Read(); )
            {
                yield return dr.GetInt32(0);
            }
        }

        public int getUserCount()
        {
            // lazy impl
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.User);
            int count = 0;
            for (dr.Open(); dr.Read(); ) count++;

            return count;
        }

        public int getItemCount()
        {
            // lazy impl
            IEntityRelationDataReader dr = EntityRelationDataProvider.GetEntity(EntityType.CatalogItem);
            int count = 0;
            for (dr.Open(); dr.Read(); ) count++;

            return count;
        }

        public IEnumerator<Rating> getRatingEnumerator()
        {
            return ratings.getRatings().GetEnumerator();
        }

        public RatingList getAllRatings()
        {
            return ratings;
        }


        public Rating getRating(int userId, int itemId)
        {
            UserRatings ratings = getUserRatings(userId);
            if (ratings == null) return null;
            return ratings.getRating(itemId);
        }

        public float MaxValue
        {
            get
            {
                return max;
            }
        }

        public float MinValue
        {
            get
            {
                return min;
            }
        }
        #endregion
    }
}
