﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{

    class RatingEvent
    {
        public int user_id;
        public int item_id;
        public double rating;
    }

    class Ratings
    {
        public HashSet<RatingEvent> items = new HashSet<RatingEvent>();
        private double sum_rating = 0;
        private double sum_sqr_rating = 0;
        private int count = 0;
        public double avg() { return sum_rating / count; }
        public void addRating(RatingEvent rating) 
        {
            count++;
            sum_rating += rating.rating;
            sum_sqr_rating += rating.rating*rating.rating;
            items.Add(rating);
        }
        public void changeRating(RatingEvent rating, double new_rating)
        {
            sum_rating += new_rating - rating.rating;
            sum_sqr_rating += new_rating*new_rating - rating.rating * rating.rating;
        }
        public void removeRating(RatingEvent rating)
        {
            items.Remove(rating);
            count--;
            sum_rating -= rating.rating;
            sum_sqr_rating -= rating.rating * rating.rating;            
        }

        public RatingEvent findRating(int user_id, int item_id)
        {
            foreach (RatingEvent rating in items)
            {
                if ((rating.user_id == user_id) && (rating.item_id == item_id))
                {
                    return rating;
                }
            }
            return null;
        }
    }

    /// <summary>
    /// Data storage for rating data.
    /// The rating events are accessible in user-wise, item-wise and unsorted triple-wise order.
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    class Data
    {
        public Ratings allRatings = null;
        public List<Ratings> userRatings = null;
        public List<Ratings> itemRatings = null;
        public double max_value;
        public double min_value;

        public Data(bool hasAll, bool hasUser, bool hasItem, double maxVal, double minVal) 
        {
            max_value = maxVal;
            min_value = minVal;

            if (hasAll) {
                allRatings = new Ratings();
            }
            if (hasUser) {
                userRatings = new List<Ratings>();
            }
            if (hasItem) {
                itemRatings = new List<Ratings>();
            }
        }

        public void release(Object o)
        {
            o = null;
        }

        public void add(IEntityRelationDataReader data)
        {
            while (data.Read())
            {
                RatingEvent rating = new RatingEvent();
                rating.user_id = (int) data.GetInt32(0);
                rating.item_id = (int) data.GetInt32(1);
                rating.rating = data.GetDouble(2);
                addRating(rating);
            }
        }

        public void addRating(RatingEvent rating)
        {
            if (itemRatings != null)
            {
                addUser(rating.user_id);
                userRatings[(int)rating.user_id].addRating(rating);
            }
            if (itemRatings != null)
            {
                addItem(rating.item_id);
                itemRatings[(int)rating.item_id].addRating(rating);
            }
            if (allRatings != null)
            {
                allRatings.addRating(rating);
            }

        }

        public void addUser(int entity_id)
        {
            if (userRatings != null)
            {
                while (entity_id >= userRatings.Count)
                {
                    Ratings ratings = new Ratings();
                    userRatings.Add(ratings);
                }
            }
        }

        public void addItem(int entity_id)
        {
            if (itemRatings != null)
            {
                while (entity_id >= itemRatings.Count)
                {
                    Ratings ratings = new Ratings();
                    itemRatings.Add(ratings);
                }
            }
        }

        public void removeRating(RatingEvent rating)
        {
            if ((userRatings != null) && (rating.user_id < userRatings.Count))
            {
                userRatings[(int)rating.user_id].removeRating(rating);
            }
            if ((itemRatings != null) && (rating.item_id < itemRatings.Count))
            {
                itemRatings[(int)rating.item_id].removeRating(rating);
            }
            if ((allRatings != null))
            {
                allRatings.removeRating(rating);
            }
        }

        public void removeUser(int user_id)
        {
            if (userRatings != null)
            {
                Ratings r = userRatings[(int)user_id];
                List<RatingEvent> temp = r.items.ToList();                
                while (temp.Count > 0)
                {
                    removeRating(temp[0]);
                }
            }
            else if ((itemRatings != null) || (allRatings != null)) 
            {
                throw new Exception("data storage out of sync");
            }
        }

        public void removeItem(int item_id)
        {
            if (itemRatings != null)
            {
                Ratings r = itemRatings[(int)item_id];
                List<RatingEvent> temp = r.items.ToList();
                while (temp.Count > 0)
                {
                    removeRating(temp[0]);
                }
            }
            else if ((userRatings != null) || (allRatings != null))
            {
                throw new Exception("data storage out of sync");
            }
        }

        public void changeRating(RatingEvent rating, double new_rating)
        {
            if ((userRatings != null) && (rating.user_id < userRatings.Count))
            {
                userRatings[(int)rating.user_id].changeRating(rating, new_rating);
            }
            if ((itemRatings != null) && (rating.item_id < itemRatings.Count))
            {
                itemRatings[(int)rating.item_id].changeRating(rating, new_rating);
            }
            if ((allRatings != null))
            {
                allRatings.changeRating(rating, new_rating);
            }
            rating.rating = new_rating;
        }
        
        public RatingEvent findRating(int user_id, int item_id)
        {
            int cnt_user = Int32.MaxValue;
            int cnt_item = Int32.MaxValue;
            if ((userRatings != null) && (userRatings.Count > user_id))
            {
                cnt_user = userRatings[(int)user_id].items.Count;
            }
            if ((itemRatings != null) && (itemRatings.Count > item_id))
            {
                cnt_item = itemRatings[(int)item_id].items.Count;
            }
            
            if (cnt_user < cnt_item) {
                return userRatings[(int)user_id].findRating(user_id, item_id);
            } else if (cnt_user > cnt_item) {
                return itemRatings[(int)item_id].findRating(user_id, item_id);
            } else if (cnt_user < Int32.MaxValue) {
                return userRatings[(int)user_id].findRating(user_id, item_id);
            }
            else if (allRatings != null)
            {
                return allRatings.findRating(user_id, item_id);
            } else {
                return null;
            }
        }
    }
}
