﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MyMediaProject.RecommenderSystem.Algorithms.Utils
{
    /// <summary>
    /// Sparse vector implementation (NOT thread safe)
    /// </summary>
    /// <typeparam name="T">Vector element type</typeparam>
    [Serializable]
    public class SparseVector<T> : IList<T>
    {
        /// <summary>
        /// Element processor that knows how to add two instances of the parameter type
        /// </summary>
        /// <typeparam name="U">Parameter type</typeparam>
        public interface IElementProcessor<U>  {
            /// <summary>
            /// Adds the two parameters 
            /// </summary>
            /// <param name="p1">Parameter 1</param>
            /// <param name="p2">Parameter 2</param>
            /// <returns>The sum of the two parameters</returns>
            U Add(U p1, U p2);
        }

        class IntElementProcessorImpl : IElementProcessor<int> {
            public int  Add(int p1, int p2)
            {
 	            return p1 + p2;
            }
        }

        class FloatElementProcessorImpl : IElementProcessor<float>
        {
            public float Add(float p1, float p2)
            {
                return p1 + p2;
            }
        }

        /// <summary>
        /// IElementProcessor for adding ints
        /// </summary>
        public static IElementProcessor<int> IntElementprocessor = new IntElementProcessorImpl();
        /// <summary>
        /// IElementProcessor for adding floats
        /// </summary>
        public static IElementProcessor<float> FloatElementprocessor = new FloatElementProcessorImpl();

        private Dictionary<int, T> data;
        private int size;

        /// <summary>
        /// Constructor
        /// </summary>
        public SparseVector()
        {
            data = new Dictionary<int, T>();
            size = 0;
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="capacity">Default capacity of the vector</param>
        public SparseVector(int capacity)
        {
            data = new Dictionary<int, T>(capacity);
            size = 0;
        }


        #region IList<T> Members

        /// <inheritdoc />
        public int IndexOf(T item)
        {
            // not great for performance
            foreach (int i in data.Keys) {
                if (data[i].Equals(item)) return i;
            }
            return -1;
        }

        /// <inheritdoc />
        public void Insert(int index, T item)
        {
            // very expensive
            Dictionary<int, T> newData = new Dictionary<int, T>();
            foreach (int i in data.Keys)
            {
                T value = data[i];
                if (i >= index)
                {
                    newData[i + 1] = value;
                }
                else
                {
                    newData[i] = value;
                }
            }
            newData[index] = item;
            data = newData;
            size++;
        }

        /// <inheritdoc />
        public void RemoveAt(int index)
        {
            data.Remove(index);
            if (index == size) 
            {
                if (data.Count == 0)
                {
                    size = 0;
                }
                else
                {
                    // very expensive
                    size = data.Keys.Max();
                }
            }
        }

        /// <inheritdoc />
        public T this[int index]
        {
            get
            {
                if (data.ContainsKey(index)) return data[index];
                return default(T);
            }
            set
            {
                data[index] = value;
                if (index >= size) size = index + 1;
            }
        }

        #endregion

        #region ICollection<T> Members

        /// <inheritdoc />
        public void Add(T item)
        {
            data[size] = item;
            size++;
        }

        /// <inheritdoc />
        public void Clear()
        {
            size = 0;
            data.Clear();
        }

        /// <inheritdoc />
        public bool Contains(T item)
        {
            return data.ContainsValue(item);
        }

        /// <inheritdoc />
        public void CopyTo(T[] array, int arrayIndex)
        {
            // how to implement this, do we want to create a 'packed' array
            // containing all elements, or a 'sparse' array where all the
            // intermediate elements are 0? The first does not seem very
            // useful, the second may quickly cause memory issues
            // 1:
            // data.Values.CopyTo(array, arrayIndex);
            // 2: (naive)
            //int aIndex = arrayIndex;
            //for (int i = 0; i < size; i++) {
            //    array[aIndex] = this[i];
            //    aIndex++;
            //}
            throw new NotImplementedException();
        }

        /// <inheritdoc />
        public int Count
        {
            get { return size; }
        }

        /// <inheritdoc />
        public bool IsReadOnly
        {
            get { return false; }
        }

        /// <inheritdoc />
        public bool Remove(T item)
        {
            int index = IndexOf(item);
            if (index < 0) return false;

            RemoveAt(index);
            return true;
        }

        #endregion

        #region IEnumerable<T> Members

        /// <inheritdoc />
        public IEnumerator<T> GetEnumerator()
        {
            return this.data.Values.GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return data.Values.GetEnumerator();
        }

        #endregion

        /// <summary>
        /// Gets en enumerator over the set of keys of this vector (the keys having non-empty values)
        /// </summary>
        /// <returns></returns>
        public IEnumerator<int> getKeyIterator()
        {
            return data.Keys.GetEnumerator();
        }

        /// <summary>
        /// Gets an enumerator over the combined set of keys (unique) of this vector and the specified other one
        /// </summary>
        /// <param name="other">Other vector</param>
        /// <returns>Enumerator over the combined set of keys</returns>
        public IEnumerator<int> getKeyIteratorUnion(SparseVector<T> other)
        {
            HashSet<int> set = new HashSet<int>();
            set.UnionWith(this.data.Keys);
            set.UnionWith(other.data.Keys);
            return set.GetEnumerator();
        }

        /// <summary>
        /// Adds the specified value to the existing value at the specified index
        /// </summary>
        /// <param name="index">Index in the list</param>
        /// <param name="value">Value to add</param>
        /// <param name="ep">Element processor that knows how to add two instances of the parameter type</param>
        /// <returns>The sum of the two values</returns>
        public T addTo(int index, T value, IElementProcessor<T> ep) {
            T total = ep.Add(this[index], value);
            this[index] = total;
            return total;
        }

        /// <summary>
        /// Number of actual elements
        /// </summary>
        /// <returns>Number of actual elements</returns>
        public int getNrOfNonZeroValues()
        {
            return data.Count;
        }
    }
}
