﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using MyMediaProject.RecommenderSystem.Core.Databases;
using System.Linq;
using System.Text;

using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Identifies the store operation.
    /// </summary>
    public enum StoreOperation
    {
        /// <summary>
        /// CatalogItem has been added.
        /// </summary>
        Add,
        /// <summary>
        /// CatalogItem has been modified.
        /// </summary>
        Remove,
        /// <summary>
        /// CatalogItem has been removed.
        /// </summary>
        Update
    }

    /// <summary>
    /// A base class for the catalog update functionality.
    /// </summary>
    /// <remarks>
    /// Every recommender framework contains an item catalog. This class is used to implement the update functionality
    /// for this catalog. It is added to the framework using the <see cref="Recommender&lt;TUserActionType&gt;.RegisterCatalogFactory"/> method and 
    /// removed from the catalog with the <see cref="Recommender&lt;TUserActionType&gt;.UnregisterCatalogFactory"/> method.
    /// </remarks>
    /// <example>
    /// <code>
    /// public class MyClass : CatalogItem
    /// {
    ///   ...
    /// }
    /// 
    /// public class MyCatalogItemFactory : CatalogItemFactory&lt;MyClass&gt;
    /// {
    ///      new Thread(this.UpdateThread).Start();
    /// }
    /// 
    /// private void UpdateThread()
    /// {
    ///     while (true)
    ///     {
    ///         MyClass myClass = new MyClass();
    ///         // fill MyClass
    ///         ...
    ///         base.RecordChange(StoreOperation.Add, myClass);
    /// 
    ///         .... // more add operations
    ///         
    ///         base.SubmitChanges(true);
    ///         
    ///         Thread.Sleep(10000);
    ///     }
    ///     
    /// }
    /// </code></example>
    public class CatalogItemFactory : IDisposable
    {
        private CatalogItemFactoryId catalogFactoryId;
        private string description;
        private PersistentStoreId persistentStoreId;
        private CatalogItemStoreDataContext dataContext;
        private ProgramId programId;
        private bool registered;

        /// <summary>
        /// Initializes a new instance of the <see cref="CatalogItemFactory"/> class.
        /// </summary>
        /// <param name="catalogFactoryId">The catalog factory id.</param>
        /// <param name="description">The description.</param>
        public CatalogItemFactory(CatalogItemFactoryId catalogFactoryId, string description)
        {
            this.catalogFactoryId = catalogFactoryId;
            this.description = description;
        }

        /// <summary>
        /// Initializes the CatalogItemFactory with the given program id.
        /// </summary>
        /// <remarks>
        /// The program id used to initialize this instance must be the same program id used by the Recommender.
        /// </remarks>
        /// <param name="recommenderProgramId">The program id, which is shared with the Recommender system also consuming this store.</param>
        public void Initialize(ProgramId recommenderProgramId)
        {
            Contract.Precondition(this.registered == false, "CatalogItemFactory still registered. Please uninitialize first.");

            this.programId = recommenderProgramId;
            this.dataContext = this.GetDataContext();           

            this.persistentStoreId = Constants.CatalogItemStoreId;

            this.registered = true;
        }

        /// <summary>
        /// Uninitializes this instance.
        /// </summary>
        public void Uninitialize()
        {
            Contract.Precondition(this.registered, "CatalogItemProvider is not initialized.");

            this.DisposeOfDataContext();

            this.registered = false;
        }

        /// <summary>
        /// Submits the changes.
        /// </summary>
        public void SubmitChanges()
        {
            int newChangeNumber = this.ChangeNumber + 1;
            IList<CatalogItem> newItems = this.DataContext.GetChangeSet().Inserts.Where(element => element is CatalogItem).Select(item => item as CatalogItem).ToList();
            IList<CatalogItem> updatedItems = this.DataContext.GetChangeSet().Updates.Where(element => element is CatalogItem).Select(item => item as CatalogItem).ToList();
            IList<CatalogItem> deletedItems = this.DataContext.GetChangeSet().Deletes.Where(element => element is CatalogItem).Select(item => item as CatalogItem).ToList(); 
            
            this.DataContext.SubmitChanges();

            this.AddChanges(newItems, StoreChangeType.Added, newChangeNumber);
            this.AddChanges(updatedItems, StoreChangeType.Changed, newChangeNumber);
            this.AddChanges(deletedItems, StoreChangeType.Removed, newChangeNumber);
            
            SendStoreChangedNotification(new StoreChangedEventArgs(this.persistentStoreId, newChangeNumber));

            this.DisposeOfDataContext();
            this.dataContext = this.GetDataContext();
        }

        /// <summary>
        /// Removed all data from the data store.
        /// </summary>
        public void EmptyStore()
        {
            Contract.Precondition(this.registered, "CatalogItemProvider is not initialized.");

            using (CatalogItemStore store = new CatalogItemStore(this.dataContext))
            {
                store.Initialize();
                store.RemoveAll();
            }
        }

        private static void SendStoreChangedNotification(StoreChangedEventArgs storeChangedEventArgs)
        {
            new CatalogStoreChangedEvent().Fire(storeChangedEventArgs);
        }

        private void AddChanges(IEnumerable<CatalogItem> items, StoreChangeType changeType, int newChangeNumber)
        {
            foreach (CatalogItem item in items)
            {
                StoreTracker<ItemStoreChange>.AddChange(
                    this.DataContext.ItemStoreChanges,
                    newChangeNumber,
                    changeType,
                    item.ItemId);
            }
        }

        private int ChangeNumber
        {
            get
            {
                int changeNumber = 0;
                lock (this)
                {
                    changeNumber = StoreTracker<ItemStoreChange>.ChangeNumber(this.DataContext.ItemStoreChanges);
                }
                return changeNumber;
            }
        }

        /// <summary>
        /// Gets the id.
        /// </summary>
        /// <value>The id.</value>
        public CatalogItemFactoryId Id
        {
            get
            {
                Contract.Precondition(this.registered, "CatalogItemProvider is not initialized.");

                return this.catalogFactoryId;
            }
        }

        /// <summary>
        /// Gets the description.
        /// </summary>
        /// <value>The description.</value>
        public string Description
        {
            get
            {
                return this.description;
            }
        }

        /// <summary>
        /// Gets the data context.  The data context should not be cached, and should be accessed directly via this property for every use.
        /// </summary>
        /// <value>The data context.</value>
        protected CatalogItemStoreDataContext DataContext
        {
            get {
                Contract.Precondition(this.registered, "CatalogItemProvider is not initialized.");

                return this.dataContext; 
            }
        }

        private CatalogItemStoreDataContext GetDataContext()
        {
            string dbConnection = Recommender.GetDatabaseConnectionString(
                this.programId);
            CatalogItemStoreDataContext context = new CatalogItemStoreDataContext(dbConnection);

            return context;
        }

        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.DisposeOfDataContext();
            }
        }

        private void DisposeOfDataContext()
        {
            if (null != this.dataContext)
            {
                this.dataContext.Connection.Dispose();
                this.dataContext.Dispose();
                this.dataContext = null;
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

    }
}
