﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Globalization;
using MyMediaProject.RecommenderSystem.Core.Databases;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Reprents an item in the catalog, which is stored in the framework database.
    /// </summary>
    public partial class CatalogItem
    {
        /// <summary>
        /// Returns a <see cref="T:System.String"/> that represents the current <see cref="T:System.Object"/>.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String"/> that represents the current <see cref="T:System.Object"/>.
        /// </returns>
        public override string ToString()
        {
            string toString;
            ItemTitle itemTitle = this.ItemTitles.FirstOrDefault(element => element.Language.Code == CultureInfo.CurrentCulture.Name);
            if (null == itemTitle)
            {
                toString = this.ItemId.ToString(CultureInfo.InvariantCulture);
            }
            else
            {
                toString = itemTitle.Title;
            }
            return toString;
        }
        #region Private
        partial void OnCreated()
        {
            this.Created = DateTime.Now;
            this.Modified = DateTime.Now;
            RecommenderStatistics.IncrimentCatalogItemsCreated();
        }
        #endregion

        #region Keywords
        /// <summary>
        /// Associated the keyword to this item, adding the keyword to the system if it does not already exist.
        /// </summary>
        /// <param name="dataContext">The context.</param>
        /// <param name="word">The word.</param>
        /// <param name="cultureInfo">The culture info.</param>
        /// <param name="level">The level.</param>
        public void AddKeyword(CatalogItemStoreDataContext dataContext, string word, CultureInfo cultureInfo, int level)
        {
            this.AddKeyword(dataContext, dataContext.CreateKeyword(word, cultureInfo), level);
        }
        /// <summary>
        /// Associated the keyword to this item.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="keyword">The keyword.</param>
        /// <param name="level">The level.</param>
        public void AddKeyword(CatalogItemStoreDataContext dataContext, Keyword keyword, int level)
        {
            this.AddKeyword(dataContext, keyword, level, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds the keyword.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="keyword">The keyword.</param>
        /// <param name="level">The level.</param>
        /// <param name="module">The module.</param>
        public void AddKeyword(CatalogItemStoreDataContext dataContext, Keyword keyword, int level, Module module)
        {
            //Only add this keyword to the item if it does not already exist.
            if (!this.ItemKeywords.Any(item =>
                item.Keyword.Word == keyword.Word &&
                item.Keyword.Language.Code == keyword.Language.Code &&
                item.Level == level))
            {
                ItemKeyword itemKeyword = new ItemKeyword();
                itemKeyword.Keyword = keyword;
                itemKeyword.Module = module;
                itemKeyword.Level = level;
                this.ItemKeywords.Add(itemKeyword);
            }
        }
	    #endregion    

        #region Genres
        /// <summary>
        /// Adds the genre.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="name">The name.</param>
        public void AddGenre(CatalogItemStoreDataContext dataContext, string name)
        {
            this.AddGenre(dataContext, dataContext.CreateGenre(name, string.Empty));
        }
        /// <summary>
        /// Adds the genre.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="name">The name.</param>
        /// <param name="uri">The URI.</param>
        public void AddGenre(CatalogItemStoreDataContext dataContext, string name, string uri)
        {
            this.AddGenre(dataContext, dataContext.CreateGenre(name, uri));
        }
        /// <summary>
        /// Adds the genre.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="genre">The genre.</param>
        public void AddGenre(CatalogItemStoreDataContext dataContext, Genre genre)
        {
            this.AddGenre(dataContext, genre, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds the genre.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="genre">The genre.</param>
        /// <param name="module">The module.</param>
        public void AddGenre(CatalogItemStoreDataContext dataContext, Genre genre, Module module)
        {
            //Contract.Precondition(dataContext.GenreExists(genre.Name, genre.Uri),
            //    "Genres must be added to the system first.  Use CatalogItemStoreDataContext.CreateGenre.");
            if (null == this.ItemGenres.FirstOrDefault(item =>
                item.Genre.Name == genre.Name &&
                item.Genre.Uri == genre.Uri))
            {
                ItemGenre itemGenre = new ItemGenre();
                itemGenre.Genre = genre;
                itemGenre.Module = module;
                this.ItemGenres.Add(itemGenre);
            }
        }
        #endregion

        #region Titles
        /// <summary>
        /// Adds the title.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="title">The title.</param>
        /// <param name="cultureInfo">The culture info.</param>
        public void AddTitle(CatalogItemStoreDataContext dataContext, string title, CultureInfo cultureInfo)
        {
            this.AddTitle(dataContext, dataContext.CreateTitle(title, cultureInfo));
        }
        /// <summary>
        /// Adds the title.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="itemTitle">The item title.</param>
        public void AddTitle(CatalogItemStoreDataContext dataContext, ItemTitle itemTitle)
        {
            this.ItemTitles.Add(itemTitle);
        }
        #endregion

        #region Locators
        /// <summary>
        /// Adds the locator.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="url">The URL.</param>
        public void AddLocator(CatalogItemStoreDataContext dataContext, string url)
        {
            this.AddLocator(dataContext, dataContext.CreateLocator(url));
        }
        /// <summary>
        /// Adds the locator.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="locator">The locator.</param>
        public void AddLocator(CatalogItemStoreDataContext dataContext, Locator locator)
        {
            //Contract.Precondition(dataContext.LocatorExists(locator.Url), 
            //    "Locator must be added to the system before using; see CatalogItemStoreDataContext.CreateLocator");
            if (null == this.Locators.FirstOrDefault(item => item.Url == locator.Url))
            {
                this.Locators.Add(locator);
            }
        }
        #endregion

        #region Descriptions
        /// <summary>
        /// Adds the description.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="description">The description.</param>
        /// <param name="cultureInfo">The culture info.</param>
        public void AddDescription(CatalogItemStoreDataContext dataContext, string description, CultureInfo cultureInfo)
        {
            this.AddDescription(dataContext, dataContext.CreateDescription(description, cultureInfo));
        }

        /// <summary>
        /// Adds the description.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="itemDescription">The item description.</param>
        public void AddDescription(CatalogItemStoreDataContext dataContext, ItemDescription itemDescription)
        {
            this.ItemDescriptions.Add(itemDescription);
        }
        #endregion

        #region CaptionLanguages
        /// <summary>
        /// Adds the caption language.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="cultureInfo">The culture info.</param>
        public void AddCaptionLanguage(CatalogItemStoreDataContext dataContext, CultureInfo cultureInfo)
        {
            this.AddCaptionLanguage(dataContext, dataContext.CreateCaptionLanguage(cultureInfo));
        }
        /// <summary>
        /// Adds the caption language.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="captionLanguage">The caption language.</param>
        public void AddCaptionLanguage(CatalogItemStoreDataContext dataContext, CaptionLanguage captionLanguage)
        {
            //Contract.Precondition(dataContext.CaptionLanguageExists(new CultureInfo(captionLanguage.Language.Code)),
            //    "CaptionLanguage must exist before adding to an item; see CatalogItemStoreDataContext.CreateCaptionLanguage.");
            if (!this.CaptionLanguages.Any(item => item.Language.Code == captionLanguage.Language.Code))
            {
                this.CaptionLanguages.Add(captionLanguage);
            }
        }
        #endregion

        #region ClusterMember
        #endregion

        #region ItemGroup
        #endregion

        #region ItemGroupMember
        /// <summary>
        /// Adds to group.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="group">The group.</param>
        public void AddToGroup(CatalogItemStoreDataContext dataContext, CatalogItem group)
        {
            this.AddToGroup(dataContext, group, false, 0, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds to group.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="group">The group.</param>
        /// <param name="sequenceNumber">The sequence number.</param>
        public void AddToGroup(CatalogItemStoreDataContext dataContext, CatalogItem group, int sequenceNumber)
        {
            this.AddToGroup(dataContext, group, true, sequenceNumber, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds to group.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="group">The group.</param>
        /// <param name="module">The module.</param>
        public void AddToGroup(CatalogItemStoreDataContext dataContext, CatalogItem group, Module module)
        {
            this.AddToGroup(dataContext, group, false, 0, module);
        }
        /// <summary>
        /// Adds to group.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="group">The group.</param>
        /// <param name="sequenceNumber">The sequence number.</param>
        /// <param name="module">The module.</param>
        public void AddToGroup(CatalogItemStoreDataContext dataContext, CatalogItem group, int sequenceNumber, Module module)
        {
            this.AddToGroup(dataContext, group, true, sequenceNumber, module);
        }
        private void AddToGroup(CatalogItemStoreDataContext dataContext, CatalogItem group, bool isOrderedGrouping, int sequenceNumber, Module module)
        {
            //Contract.Precondition(dataContext.ItemExists(group),
            //    "The CatalogItem representing the group does not exist in the system; see CatalogItemStoreDataContext.CreateItem.");
            Contract.Precondition(this != group);
            Contract.Precondition(sequenceNumber >= 0, "Group sequence numbers must be non-negative");
            Contract.Precondition(group.IsOrdered && isOrderedGrouping || group.IsOrdered && !isOrderedGrouping || !group.IsOrdered && !isOrderedGrouping,
                "A sequence number can only be applied to an ordered group.");
            if (!this.ChildRelationships.Any(groupRelation => groupRelation.Parent == group))
            {
                ItemGroupMember itemGroupMember = new ItemGroupMember();
                itemGroupMember.Module = module;
                itemGroupMember.SequenceNumber = sequenceNumber;
                group.ParentRelationships.Add(itemGroupMember);
                this.ChildRelationships.Add(itemGroupMember);
            }
        }
        #endregion

        #region Property
        /// <summary>
        /// Adds the property.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="name">The name.</param>
        /// <param name="value">The value.</param>
        public void AddProperty(CatalogItemStoreDataContext dataContext, string name, string value)
        {
            this.AddProperty(dataContext, dataContext.CreateProperty(name, value));
        }
        /// <summary>
        /// Adds the property.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="property">The property.</param>
        public void AddProperty(CatalogItemStoreDataContext dataContext, Property property)
        {
            this.AddProperty(dataContext, property, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds the property.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="property">The property.</param>
        /// <param name="module">The module.</param>
        public void AddProperty(CatalogItemStoreDataContext dataContext, Property property, Module module)
        {
            //Contract.Precondition(dataContext.PropertyExists(property.Name, property.Value),
            //    "Property must be added before use; see CatalogItemStoreDataContext.CreateProperty");
            if (!this.ItemProperties.Any(item => item.Property.Name == property.Name && item.Property.Value == property.Value))
            {
                ItemProperty itemProperty = new ItemProperty();
                itemProperty.Property = property;
                itemProperty.Module = module;
                this.ItemProperties.Add(itemProperty);
            }
        }
        #endregion

        #region Relationship
        /// <summary>
        /// Adds the relationship.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="name">The name.</param>
        /// <param name="relationshipUri">The URI.</param>
        /// <param name="relatedItemUrl">The related item URL.</param>
        public void AddRelationship(CatalogItemStoreDataContext dataContext, string name, string relationshipUri, string relatedItemUrl)
        {
            this.AddRelationship(dataContext, dataContext.CreateRelationship(name, relationshipUri), relatedItemUrl);
        }
        /// <summary>
        /// Adds the relationship.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="relationship">The relationship.</param>
        /// <param name="relatedItemUrl">The related item URL.</param>
        public void AddRelationship(CatalogItemStoreDataContext dataContext, Relationship relationship, string relatedItemUrl)
        {
            this.AddRelationship(dataContext, relationship, relatedItemUrl, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds the relationship.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="relationship">The relationship.</param>
        /// <param name="relatedItemUrl">The related item URL.</param>
        /// <param name="module">The module.</param>
        public void AddRelationship(CatalogItemStoreDataContext dataContext, Relationship relationship, string relatedItemUrl, Module module)
        {
            //Contract.Precondition(dataContext.RelationshipExists(relationship.Name, relationship.Uri),
            //    "Relationship must be added before use; see CatalogItemStoreDataContext.CreateRelationship.");
            if (!this.ItemRelatedMaterials.Any(item => item.Relationship.Name == relationship.Name && item.Relationship.Uri == relationship.Uri &&
                item.Url == relatedItemUrl))
            {
                ItemRelatedMaterial itemRelatedMaterial = new ItemRelatedMaterial();
                itemRelatedMaterial.Relationship = relationship;
                itemRelatedMaterial.Url = relatedItemUrl;
                itemRelatedMaterial.Module = module;
                this.ItemRelatedMaterials.Add(itemRelatedMaterial);
            }
        }
        #endregion

        #region Credit
        /// <summary>
        /// Adds the credit.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="personCredit">The person credit.</param>
        /// <param name="role">The role.</param>
        public void AddCredit(CatalogItemStoreDataContext dataContext, Credit personCredit, Role role)
        {
            this.AddCredit(dataContext, personCredit, null, role, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds the credit.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="personCredit">The person credit.</param>
        /// <param name="characterCredit">The character credit.</param>
        /// <param name="role">The role.</param>
        public void AddCredit(CatalogItemStoreDataContext dataContext, Credit personCredit, Credit characterCredit, Role role)
        {
            this.AddCredit(dataContext, personCredit, characterCredit, role, dataContext.DefaultModule);
        }
        /// <summary>
        /// Adds the character credit.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="personCredit">The person credit.</param>
        /// <param name="role">The role.</param>
        /// <param name="module">The module.</param>
        public void AddCredit(CatalogItemStoreDataContext dataContext, Credit personCredit, Role role, Module module)
        {
            this.AddCredit(dataContext, personCredit, null, role, module);
        }
        /// <summary>
        /// Adds the person credit.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="personCredit">The person credit.</param>
        /// <param name="characterCredit">The character credit.</param>
        /// <param name="role">The role.</param>
        /// <param name="module">The module.</param>
        public void AddCredit(CatalogItemStoreDataContext dataContext, Credit personCredit, Credit characterCredit, Role role, Module module)
        {
            ItemCredit itemCredit = new ItemCredit()
            {
                PersonCredit = personCredit,
                CharacterCredit = characterCredit,
                Role = role,
                Module = module
            };
            this.ItemCredits.Add(itemCredit);
        }
        #endregion

        #region PromotionalInfo
        /// <summary>
        /// Adds the promotional info.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="promotionalInfo">The promotional info.</param>
        public void AddPromotionalInfo(CatalogItemStoreDataContext dataContext, ItemPromotionalInfo promotionalInfo)
        {
            if (!this.ItemPromotionalInfos.Any(item => item.PromotionalText == promotionalInfo.PromotionalText))
            {
                this.ItemPromotionalInfos.Add(promotionalInfo);
            }
        }
        #endregion

        #region AudioLanguage
        /// <summary>
        /// Adds the audio language to the item, with the defaul module.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="audioCultureInfo">The audio culture info.</param>
        /// <param name="isSupplemental">if set to <c>true</c> [is supplemental].</param>
        public void AddAudioLanguage(CatalogItemStoreDataContext dataContext, CultureInfo audioCultureInfo, bool isSupplemental)
        {
            this.AddAudioLanguage(dataContext, audioCultureInfo, isSupplemental, dataContext.DefaultModule);
        }

        /// <summary>
        /// Adds the audio language.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="audioCultureInfo">The audio culture info.</param>
        /// <param name="isSupplemental">if set to <c>true</c> [is supplemental].</param>
        /// <param name="module">The module.</param>
        public void AddAudioLanguage(CatalogItemStoreDataContext dataContext, CultureInfo audioCultureInfo, bool isSupplemental, Module module)
        {
            if (!this.ItemAudioLanguages.Any(item => item.Language.Code == audioCultureInfo.Name))
            {
                ItemAudioLanguage audioLanguage = new ItemAudioLanguage()
                {
                    Created = DateTime.Now,
                    Modified = DateTime.Now,
                    Language = dataContext.CreateLanguage(audioCultureInfo),
                    Module = module,
                    IsSupplemental = isSupplemental
                };
                this.ItemAudioLanguages.Add(audioLanguage);
            }
        }
        #endregion
    }
}
