﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Globalization;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;

namespace MyMediaProject.RecommenderSystem.Core.Databases
{
    /// <summary>
    /// Core database interface class, which manages the class/database relationship.
    /// </summary>
    public sealed class CatalogItemStoreDataContext : MyMediaDataContext
    {
        private Service defaultService;
        private Genre defaultGenre;
        private string defaultServiceName = "<service name not set>";
        private string defaultGenreName = "<genre not set>";

        private Dictionary<string, Language> languagesCache = new Dictionary<string, Language>();
        private Dictionary<string, Keyword> keywordsCache = new Dictionary<string, Keyword>();
        private Dictionary<string, Genre> genresCache = new Dictionary<string, Genre>();
        private Dictionary<string, CaptionLanguage> captionLanguagesCache = new Dictionary<string, CaptionLanguage>();
        private Dictionary<string, Relationship> relationshipsCache = new Dictionary<string, Relationship>();
        private Dictionary<string, Role> rolesCache = new Dictionary<string, Role>();
        private Dictionary<string, Service> servicesCache = new Dictionary<string, Service>();

        /// <summary>
        /// Initializes a new instance of the <see cref="CatalogItemStoreDataContext"/> class.
        /// </summary>
        /// <param name="connection">The connection.</param>
        public CatalogItemStoreDataContext(string connection)
            : base(connection)
        {
        }

        /// <summary>
        /// Called after a set of changes have been submitted.
        /// </summary>
        protected override void OnSubmitted()
        {
            base.OnSubmitted();
            this.languagesCache.Clear();
            this.genresCache.Clear();
            this.keywordsCache.Clear();
            this.captionLanguagesCache.Clear();
            this.relationshipsCache.Clear();
            this.rolesCache.Clear();
            this.servicesCache.Clear();
        }        

        internal Service DefaultService
        {
            get
            {
                lock (this)
                {
                    if (null == this.defaultService)
                    {
                        this.defaultService = this.Services.FirstOrDefault(service => service.Name == this.defaultServiceName);
                        if (null == this.defaultService)
                        {
                            this.defaultService = this.CreateService(this.defaultServiceName);
                            this.Services.InsertOnSubmit(this.defaultService);
                        }
                    }                    
                }
                Contract.Postcondition(null != this.defaultService);
                return this.defaultService;
            }
        }

        internal Genre DefaultGenre
        {
            get
            {
                lock (this)
                {
                    if (null == this.defaultGenre)
                    {
                        this.defaultGenre = this.Genres.FirstOrDefault(service => service.Name == this.defaultGenreName);
                        if (null == this.defaultGenre)
                        {
                            this.defaultGenre = this.CreateGenre(this.defaultGenreName);
                            this.Genres.InsertOnSubmit(this.defaultGenre);
                        }
                    }
                }
                Contract.Postcondition(null != this.defaultGenre);
                return this.defaultGenre;
            }
        }

        internal new void SubmitChanges()
        {
            base.SubmitChanges();
        }

        #region CatalogItem
        /// <summary>
        /// Creates a new item.
        /// </summary>
        /// <returns></returns>
        public CatalogItem CreateItem()
        {
            CatalogItem item = new CatalogItem();
            item.Module = this.DefaultModule;
            this.CatalogItems.InsertOnSubmit(item);
            return item;
        }
        /// <summary>
        /// Determines if the item exists in the system.
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns></returns>
        public bool ItemExists(CatalogItem item)
        {
            if (null == item)
                throw new ArgumentNullException("item");

            if (0 != item.ItemId)
            {
                return this.CatalogItems.Any(element => element == item);
            }
            else
            {
                return this.GetChangeSet().Inserts.Any(element => element == item);
            }
        }
        #endregion

        #region Keyword
        /// <summary>
        /// Creates the keyword if it does not already exist in the system.  Otherwise, returns the existing keyword.
        /// </summary>
        /// <param name="word">The word.</param>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns></returns>
        public Keyword CreateKeyword(string word, CultureInfo cultureInfo)
        {
            if (null == word)
                throw new ArgumentNullException("word");
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            Keyword keyword = this.GetKeyword(word, cultureInfo);
            if (null == keyword)
            {
                keyword = new Keyword();
                keyword.Word = word;
                keyword.Language = this.CreateLanguage(cultureInfo);
                keyword.Module = this.DefaultModule;
                this.Keywords.InsertOnSubmit(keyword);
                this.keywordsCache.Add(word + cultureInfo.Name, keyword);
            }
            return keyword;
        }

        /// <summary>
        /// Determines whether the specified keyword exists in the system.
        /// </summary>
        /// <param name="word">The word.</param>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns>
        /// 	<c>true</c> if the specified word contains keyword; otherwise, <c>false</c>.
        /// </returns>
        public bool KeywordExists(string word, CultureInfo cultureInfo)
        {
            if (null == word)
                throw new ArgumentNullException("word");
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            return null != this.GetKeyword(word, cultureInfo);
        }

        private Keyword GetKeyword(string word, CultureInfo cultureInfo)
        {
            Keyword keyword = null;
            if (this.keywordsCache.ContainsKey(word + cultureInfo.Name))
            {
                keyword = this.keywordsCache[word + cultureInfo.Name];
            }
            else
            {
                keyword = this.Keywords.SingleOrDefault(c => c.Word == word && c.Language.Code == cultureInfo.Name);
                if (null != keyword)
                {
                    this.keywordsCache.Add(word + cultureInfo.Name, keyword);
                }
            }
            return keyword;
        } 
        #endregion

        #region Language
        /// <summary>
        /// Creates a Language class based on the given culture info, if it does not yet exist in the system.
        /// Otherwise, returns the existing class.
        /// </summary>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns></returns>
        public Language CreateLanguage(CultureInfo cultureInfo)
        {
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            Language language = this.GetLanguage(cultureInfo);
            if (null == language)
            {
                language = new Language();
                language.Code = cultureInfo.Name;
                language.Module = this.DefaultModule;
                this.Languages.InsertOnSubmit(language);
                this.languagesCache.Add(cultureInfo.Name, language);
            }
            return language;
        }

        private Language GetLanguage(CultureInfo cultureInfo)
        {
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            Language language = null;
            if (this.languagesCache.ContainsKey(cultureInfo.Name))
            {
                language = this.languagesCache[cultureInfo.Name];
            }
            else
            {
                language = this.Languages.SingleOrDefault(item => item.Code == cultureInfo.Name);
                if (null != language)
                {
                    this.languagesCache.Add(language.Code, language);
                }
            }
            return language;
        } 
        #endregion

        #region Genre
        /// <summary>
        /// Creates the genre if it doesn't exist, or returns the existing genre object.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public Genre CreateGenre(string name)
        {
            if (null == name)
                throw new ArgumentNullException("name");

            return this.CreateGenre(name, string.Empty);
        }
        /// <summary>
        /// Creates the genre if it doesn't exist, or returns the existing genre object.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="uri">The URI.</param>
        /// <returns></returns>
        public Genre CreateGenre(string name, string uri)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == uri)
                throw new ArgumentNullException("uri");

            Genre genre = this.GetGenre(name, uri);
            if (null == genre)
            {
                genre = new Genre();
                genre.Name = name;
                genre.Uri = uri;
                genre.Module = this.DefaultModule;
                this.Genres.InsertOnSubmit(genre);
                this.genresCache.Add(name + uri, genre);
            }
            return genre;
        }

        /// <summary>
        /// Checks if the Genre exists in the system.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="uri">The URI.</param>
        /// <returns></returns>
        public bool GenreExists(string name, string uri)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == uri)
                throw new ArgumentNullException("uri");

            return null != this.GetGenre(name, uri);
        }

        private Genre GetGenre(string name, string uri)
        {
            Genre genre = null;
            if (this.genresCache.ContainsKey(name + uri))
            {
                genre = this.genresCache[name + uri];
            }
            else
            {
                genre = this.Genres.SingleOrDefault(item => item.Name == name && item.Uri == uri);
                if (null != genre)
                {
                    this.genresCache.Add(name + uri, genre);
                }
            }
            return genre;
        }
        #endregion

        #region Title
        /// <summary>
        /// Creates the title.
        /// </summary>
        /// <param name="title">The title.</param>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns></returns>
        public ItemTitle CreateTitle(string title, CultureInfo cultureInfo)
        {
            if (null == title)
                throw new ArgumentNullException("title");
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            ItemTitle itemTitle;
            itemTitle = new ItemTitle();
            itemTitle.Title = title;
            itemTitle.Language = this.CreateLanguage(cultureInfo);
            itemTitle.Module = this.DefaultModule;
            this.ItemTitles.InsertOnSubmit(itemTitle);
            return itemTitle;
        }
        #endregion

        #region Locator

        /// <summary>
        /// Creates the locator.
        /// </summary>
        /// <param name="url">The URL.</param>
        /// <returns></returns>
        public Locator CreateLocator(string url)
        {
            if (null == url)
                throw new ArgumentNullException("url");

            Locator locator = this.GetLocator(url);
            if (null == locator)
            {
                locator = new Locator();
                locator.Url = url;
                locator.Module = this.DefaultModule;
                locator.Service = this.DefaultService;
                this.Locators.InsertOnSubmit(locator);
            }
            return locator;
        }

        /// <summary>
        /// Locators the exists.
        /// </summary>
        /// <param name="url">The URL.</param>
        /// <returns></returns>
        public bool LocatorExists(string url)
        {
            if (null == url)
                throw new ArgumentNullException("url");

            return null != this.GetLocator(url);
        }

        /// <summary>
        /// Gets the locator.
        /// </summary>
        /// <param name="url">The URL.</param>
        /// <returns></returns>
        private Locator GetLocator(string url)
        {
            Locator locator;
            //= this.GetChangeSet().Inserts.SingleOrDefault(
            //    item => item is Locator &&
            //        (item as Locator).Url == url) as Locator;
            //if (null == locator)
            {
                locator = this.Locators.SingleOrDefault(item => item.Url == url);
            }
            return locator;
        }
        #endregion

        #region ItemDescription
        /// <summary>
        /// Creates the description.
        /// </summary>
        /// <param name="description">The description.</param>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns></returns>
        public ItemDescription CreateDescription(string description, CultureInfo cultureInfo)
        {
            if (null == description)
                throw new ArgumentNullException("description");
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            ItemDescription itemDescription;
            itemDescription = new ItemDescription();
            itemDescription.Description = description;
            itemDescription.Language = this.CreateLanguage(cultureInfo);
            itemDescription.Module = this.DefaultModule;
            this.ItemDescriptions.InsertOnSubmit(itemDescription);
            return itemDescription;
        }
        #endregion

        #region CaptionLanguage
        /// <summary>
        /// Creates the caption language if it does not already exist in the system.
        /// </summary>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns></returns>
        public CaptionLanguage CreateCaptionLanguage(CultureInfo cultureInfo)
        {
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            CaptionLanguage captionLanguage = this.GetCaptionLanguage(cultureInfo);
            if (null == captionLanguage)
            {
                captionLanguage = new CaptionLanguage();
                captionLanguage.Language = this.CreateLanguage(cultureInfo);
                captionLanguage.Module = this.DefaultModule;
                this.CaptionLanguages.InsertOnSubmit(captionLanguage);
                this.captionLanguagesCache.Add(cultureInfo.Name, captionLanguage);
            }
            return captionLanguage;
        }
        /// <summary>
        /// Captions the language exists.
        /// </summary>
        /// <param name="cultureInfo">The culture info.</param>
        /// <returns></returns>
        public bool CaptionLanguageExists(CultureInfo cultureInfo)
        {
            if (null == cultureInfo)
                throw new ArgumentNullException("cultureInfo");

            return null != this.GetCaptionLanguage(cultureInfo);
        }
        private CaptionLanguage GetCaptionLanguage(CultureInfo cultureInfo)
        {
            CaptionLanguage captionLanguage;
            if (this.captionLanguagesCache.ContainsKey(cultureInfo.Name))
            {
                captionLanguage = this.captionLanguagesCache[cultureInfo.Name];
            }
            else
            {
                captionLanguage = this.CaptionLanguages.SingleOrDefault(item => item.Language.Code == cultureInfo.Name);
                if (null != captionLanguage)
                {
                    this.captionLanguagesCache.Add(cultureInfo.Name, captionLanguage);
                }
            }
            return captionLanguage;
        }
        #endregion

        #region ItemGroupMember
        #endregion

        #region Relationship
        /// <summary>
        /// Creates the relationship if it does not exist in the system; otherwise returns a reference to the existing relationship.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="uri">The URI.</param>
        /// <returns></returns>
        public Relationship CreateRelationship(string name, string uri)
        {
            if(null == name)
                throw new ArgumentNullException("name");
            if(null == uri)
                throw new ArgumentNullException("uri");

            Relationship relationship = this.GetRelationship(name, uri);
            if (null == relationship)
            {
                relationship = new Relationship();
                relationship.Name = name;
                relationship.Uri = uri;
                relationship.Module = this.DefaultModule;
                this.Relationships.InsertOnSubmit(relationship);
                this.relationshipsCache.Add(name + uri, relationship);
            }
            return relationship;
        }
        /// <summary>
        /// Returns if the Relationship exists.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="uri">The URI.</param>
        /// <returns></returns>
        public bool RelationshipExists(string name, string uri)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == uri)
                throw new ArgumentNullException("uri");

            return null != this.GetRelationship(name, uri);
        }
        private Relationship GetRelationship(string name, string uri)
        {
            Relationship relationship;
            {
                relationship = this.Relationships.SingleOrDefault(item => item.Name == name && item.Uri == uri);
            }
            return relationship;
        }
        #endregion

        #region Roles
        /// <summary>
        /// Creates the role, and puts the role in a pending insertion state in the system database.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="uri">The URI.</param>
        /// <returns></returns>
        public Role CreateRole(string name, string uri)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == uri)
                throw new ArgumentNullException("uri");

            Role role = this.GetRole(name, uri);
            if (null == role)
            {
                role = new Role();
                role.Name = name;
                role.Uri = uri;
                role.Module = this.DefaultModule;
                this.Roles.InsertOnSubmit(role);
                this.rolesCache.Add(name + uri, role);
            }
            return role;
        }

        private bool RoleExists(string name, string uri)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == uri)
                throw new ArgumentNullException("uri");

            return null != this.GetRole(name, uri);
        }

        private Role GetRole(string name, string uri)
        {
            Role role = null;
            if (this.rolesCache.ContainsKey(name + uri))
            {
                role = this.rolesCache[name + uri];
            }
            else
            {
                role = this.Roles.SingleOrDefault(item => item.Name == name && item.Uri == uri);
                if (null != role)
                {
                    this.rolesCache.Add(name + uri, role);
                }
            }
            return role;
        } 
        #endregion

        #region Service
        /// <summary>
        /// Creates the service if it does not exist in the system, putting it into a pending insertion state.
        /// Otherwise, returns an existing service.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public Service CreateService(string name)
        {
            return this.CreateService(name, string.Empty);
        }

        /// <summary>
        /// Creates the service if it does not exist in the system, putting it into a pending insertion state.
        /// Otherwise, returns an existing service.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="url">The URL.</param>
        /// <returns></returns>
        public Service CreateService(string name, string url)
        {
            return this.CreateService(name, url, string.Empty);
        }

        /// <summary>
        /// Creates the service if it does not exist in the system, putting it into a pending insertion state.
        /// Otherwise, returns an existing service.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="url">The URL.</param>
        /// <param name="owner">The owner.</param>
        /// <returns></returns>
        public Service CreateService(string name, string url, string owner)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == url)
                throw new ArgumentNullException("url");
            if (null == owner)
                throw new ArgumentNullException("owner");

            Service service = this.GetService(name, url, owner);
            if (null == service)
            {
                service = new Service();
                service.Name = name;
                service.Url = url;
                service.Owner = owner;
                service.Genre = this.DefaultGenre;
                this.Services.InsertOnSubmit(service);
                this.servicesCache.Add(name + url, service);
            }
            return service;
        }

        private bool ServiceExists(string name, string uri, string owner)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == uri)
                throw new ArgumentNullException("uri");
            if (null == owner)
                throw new ArgumentNullException("owner");

            return null != this.GetService(name, uri, owner);
        }

        private Service GetService(string name, string uri, string owner)
        {
            Service service = null;
            if (this.servicesCache.ContainsKey(name + uri + owner))
            {
                service = this.servicesCache[name + uri + owner];
            }
            else
            {
                service = this.Services.SingleOrDefault(item => item.Name == name && item.Url == uri && item.Owner == owner);
                if (null != service)
                {
                    this.servicesCache.Add(name + uri + owner, service);
                }
            }
            return service;
        } 
        #endregion

        #region Module
        /// <summary>
        /// Creates the module if it does not exist, otherwise returns a reference to the existing one.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="organization">The organization.</param>
        /// <returns></returns>
        public Module CreateModule(string name, string organization)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == organization)
                throw new ArgumentNullException("organization");

            Module module = this.GetModule(name, organization);
            if (null == module)
            {
                module = new Module();
                module.Name = name;
                module.Organization = organization;
                this.Modules.InsertOnSubmit(module);
            }
            return module;
        }
        private Module GetModule(string name, string organization)
        {
            Module module;
            //= this.GetChangeSet().Inserts.SingleOrDefault(item => item is Module &&
            //    (item as Module).Name == name &&
            //    (item as Module).Organization == organization) as Module;
            //if (null == module)
            {
                module = this.Modules.SingleOrDefault(item => item.Name == name && item.Organization == organization);
            }
            return module;
        }

        #endregion
    }
}
