﻿using System;
using System.Collections.Generic;
using System.Data.Linq;
using System.Linq;
using System.Globalization;
using System.Data.Linq.Mapping;
using System.Text;
using System.Reflection;
using System.Data.SqlClient;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Databaset to store all MyMedia user/item/rating information.
    /// </summary>
    public partial class MyMediaDataContext
    {
        private const string DetailedDatabaseCreationError = 
            "There has been an error creating the database necessary for the Recommender framework.\r\n" +
            "Please verify Sql is installed, and the user account has appropriate permissions to create databases.\r\n" + 
            "See inner exception for more details.\r\n";

        private Module defaultModule;
        private string defaultModuleName = "<module not set>";
        private string defaultModuleDescription = "<description not set>";
        private string defaultModuleOrganization = "<organization not set>";
        internal Dictionary<string, Property> propertiesCache = new Dictionary<string, Property>();

        /// <summary>
        /// Creates a database on the server.
        /// </summary>
        public new void CreateDatabase()
        {
            try
            {
                base.CreateDatabase();
                CreateIndicies();
            }
            catch (SqlException e)
            {
                throw new PreconditionViolationException(DetailedDatabaseCreationError, e);
            }
        }

        /// <summary>
        /// Computes the set of modified objects to be inserted, updated, or deleted, and executes the appropriate commands to implement the changes to the database.
        /// </summary>
        public new void SubmitChanges()
        {
            base.SubmitChanges();
            this.OnSubmitted();
        }

        partial void OnCreated()
        {
            RecommenderStatistics.IncrimentActiveDataContext();
        }

        /// <summary>
        /// Sends changes that were made to retrieved objects to the underlying database, and specifies the action to be taken if the submission fails.
        /// </summary>
        /// <param name="failureMode">The action to be taken if the submission fails. Valid arguments are as follows:
        /// <see cref="F:System.Data.Linq.ConflictMode.FailOnFirstConflict"/><see cref="F:System.Data.Linq.ConflictMode.ContinueOnConflict"/></param>
        public override void SubmitChanges(System.Data.Linq.ConflictMode failureMode)
        {
            base.SubmitChanges(failureMode);
            this.OnSubmitted();
        }

        /// <summary>
        /// Called after a set of changes have been submitted.
        /// </summary>
        protected virtual void OnSubmitted()
        {
        }

        private void CreateIndicies()
        {
            string genreIndex = @"CREATE UNIQUE NONCLUSTERED INDEX IX_Genres_Name_Uri ON dbo.Genres
                (
                Name,
                Uri
                ) WITH( STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
                    ON [PRIMARY]";
            this.ExecuteCommand(genreIndex, new object[] { });
            string itemIndex = @"CREATE NONCLUSTERED INDEX IX_CatalogItem_Crid ON dbo.CatalogItems
                (
                Crid
                ) WITH( STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
                    ON [PRIMARY]";
            this.ExecuteCommand(itemIndex, new object[] { });
            string relationshipIndex = @"CREATE UNIQUE NONCLUSTERED INDEX IX_Relationship_Name_Uri ON dbo.Relationships
                (
                Name,
                Uri
                ) WITH( STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
                    ON [PRIMARY]";
            this.ExecuteCommand(relationshipIndex, new object[] { });
            string keywordIndex = @"CREATE UNIQUE NONCLUSTERED INDEX IX_Keyword_Word ON dbo.Keywords
                (
                Word
                ) WITH( STATISTICS_NORECOMPUTE = OFF, IGNORE_DUP_KEY = OFF, ALLOW_ROW_LOCKS = ON, ALLOW_PAGE_LOCKS = ON) 
                    ON [PRIMARY]";
            this.ExecuteCommand(keywordIndex, new object[] { });
            this.ExecuteCommand(MyMediaDataContext.CascadingDeleteProcedureNoWhereClause, new object[] { });
            this.ExecuteCommand(MyMediaDataContext.CascadingDeleteProcedure, new object[] { });
        }

        internal Module DefaultModule
        {
            get
            {
                lock (this)
                {
                    if (null == this.defaultModule)
                    {
                        this.defaultModule = this.Modules.FirstOrDefault(item => item.Name == this.defaultModuleName);
                        if (null == this.defaultModule)
                        {
                            this.defaultModule = new Module();
                            this.defaultModule.Name = this.defaultModuleName;
                            this.defaultModule.Description = this.defaultModuleDescription;
                            this.defaultModule.Organization = this.defaultModuleOrganization;
                            this.Modules.InsertOnSubmit(this.defaultModule);
                        }
                    }
                }
                Contract.Postcondition(null != this.defaultModule);
                return this.defaultModule;
            }
        }


        /// <summary>
        /// Deletes the database closing connections.
        /// </summary>
        public new void DeleteDatabase()
        {
           this.ExecuteQuery<int>(string.Format(CultureInfo.InvariantCulture, MyMediaDataContext.DeleteDatabaseForced, this.Connection.Database));
        }

        internal void EmptyDatabase()
        {
            foreach (MetaTable table in this.Mapping.GetTables())
            {
                int rowsAffected = 0;
                this.EmptyDatabaseTableCascading(table.TableName, rowsAffected);
            }
        }

        [Function(Name = "dbo.spDeleteRowsNoWhereClause")]
        [return: Parameter(DbType="Int")]
        internal int EmptyDatabaseTableCascading([Parameter(Name="cTableName", DbType="nvarchar(50)")] string tableName, [Parameter(Name="iRowsAffected", DbType="int")] int rowsAffected)
        {
            tableName = tableName.Replace("dbo.", "");
            IExecuteResult result = this.ExecuteMethodCall(this, ((MethodInfo)(MethodInfo.GetCurrentMethod())), tableName, rowsAffected);
            return (int)result.ReturnValue;
        }

        #region Property
        /// <summary>
        /// Creates the property if it doesn't exist; otherwise, returns and existing property.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public Property CreateProperty(string name, string value)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == value)
                throw new ArgumentNullException("value");

            Property property = this.GetProperty(name, value);
            if (null == property)
            {
                property = new Property();
                property.Name = name;
                property.Value = value;
                property.Module = this.DefaultModule;
                this.Properties.InsertOnSubmit(property);
                this.propertiesCache.Add(name + value, property);
            }
            return property;
        }
        /// <summary>
        /// Checks if the Property exists in the system.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        internal bool PropertyExists(string name, string value)
        {
            if (null == name)
                throw new ArgumentNullException("name");
            if (null == value)
                throw new ArgumentNullException("value");

            return null != this.GetProperty(name, value);
        }
        private Property GetProperty(string name, string value)
        {
            Property property;
            if (this.propertiesCache.ContainsKey(name + value))
            {
                property = this.propertiesCache[name + value];
            }
            else
            {
                property = this.Properties.SingleOrDefault(item => item.Name == name && item.Value == value);
                if (null != property)
                {
                    this.propertiesCache.Add(name + value, property);
                }
            }
            return property;
        }
        #endregion

        /// <summary>
        /// Releases all resources used by the <see cref="T:System.Data.Linq.DataContext"/>.
        /// </summary>
        public new void Dispose()
        {
            this.Connection.Dispose();  //Workaround for hanging connections to the DB.
            base.Dispose();
        }

        /// <summary>
        /// Releases resources used by the <see cref="T:System.Data.Linq.DataContext"/>.
        /// </summary>
        /// <param name="disposing">true to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
            RecommenderStatistics.DecrimentActiveDataContext();
        }

        private const string CascadingDeleteProcedureNoWhereClause =
            @"
            CREATE Procedure spDeleteRowsNoWhereClause
            /* 
            Recursive row delete procedure. 

            It deletes all rows in the table specified that conform to the criteria selected, 
            while also deleting any child/grandchild records and so on.  This is designed to do the 
            same sort of thing as Access's cascade delete function. It first reads the sysforeignkeys 
            table to find any child tables, then deletes the soon-to-be orphan records from them using 
            recursive calls to this procedure. Once all child records are gone, the rows are deleted 
            from the selected table.   It is designed at this time to be run at the command line. It could 
            also be used in code, but the printed output will not be available.
            */
	            (
	            @cTableName nvarchar(50), /* name of the table where rows are to be deleted */
	            @iRowsAffected int OUTPUT /* number of records affected by the delete */
	            )
            As
            set nocount on
            declare 	@cTab nvarchar(255), /* name of the child table */
	            @cCol nvarchar(255), /* name of the linking field on the child table */
	            @cRefTab nvarchar(255), /* name of the parent table */
	            @cRefCol nvarchar(255), /* name of the linking field in the parent table */
	            @cFKName nvarchar(255), /* name of the foreign key */
	            @cSQL nvarchar(1000), /* query string passed to the sp_ExecuteSQL procedure */
	            @cChildCriteria nvarchar(1000), /* criteria to be used to delete 
                                                       records from the child table */
	            @iChildRows int /* number of rows deleted from the child table */

            /* declare the cursor containing the foreign key constraint information */
            DECLARE cFKey CURSOR LOCAL FOR 
            SELECT SO1.name AS Tab, 
                   SC1.name AS Col, 
                   SO2.name AS RefTab, 
                   SC2.name AS RefCol, 
                   FO.name AS FKName
            FROM dbo.sysforeignkeys FK  
            INNER JOIN dbo.syscolumns SC1 ON FK.fkeyid = SC1.id 
                                          AND FK.fkey = SC1.colid 
            INNER JOIN dbo.syscolumns SC2 ON FK.rkeyid = SC2.id 
                                          AND FK.rkey = SC2.colid 
            INNER JOIN dbo.sysobjects SO1 ON FK.fkeyid = SO1.id 
            INNER JOIN dbo.sysobjects SO2 ON FK.rkeyid = SO2.id 
            INNER JOIN dbo.sysobjects FO ON FK.constid = FO.id
            WHERE SO2.Name = @cTableName

            OPEN cFKey
            FETCH NEXT FROM cFKey INTO @cTab, @cCol, @cRefTab, @cRefCol, @cFKName
            WHILE @@FETCH_STATUS = 0
                 BEGIN
	            /* build the criteria to delete rows from the child table. As it uses the 
                       criteria passed to this procedure, it gets progressively larger with 
                       recursive calls */
	            SET @cChildCriteria = @cCol + ' in (SELECT [' + @cRefCol + '] FROM [' + 
                                          @cRefTab +'])'
	            print 'Deleting records from table ' + @cTab
	            /* call this procedure to delete the child rows */
	            EXEC spDeleteRows @cTab, @cChildCriteria, @iChildRows OUTPUT
	            FETCH NEXT FROM cFKey INTO @cTab, @cCol, @cRefTab, @cRefCol, @cFKName
                 END
            Close cFKey
            DeAllocate cFKey
            /* finally delete the rows from this table and display the rows affected  */
            SET @cSQL = 'DELETE FROM [' + @cTableName + ']'
            print @cSQL
            EXEC sp_ExecuteSQL @cSQL
            print 'Deleted ' + CONVERT(nvarchar, @@ROWCOUNT) + ' records from table ' + @cTableName
            ";

        private const string CascadingDeleteProcedure =
            @"
            CREATE Procedure spDeleteRows
            /* 
            Recursive row delete procedure. 

            It deletes all rows in the table specified that conform to the criteria selected, 
            while also deleting any child/grandchild records and so on.  This is designed to do the 
            same sort of thing as Access's cascade delete function. It first reads the sysforeignkeys 
            table to find any child tables, then deletes the soon-to-be orphan records from them using 
            recursive calls to this procedure. Once all child records are gone, the rows are deleted 
            from the selected table.   It is designed at this time to be run at the command line. It could 
            also be used in code, but the printed output will not be available.
            */
	            (
	            @cTableName nvarchar(50), /* name of the table where rows are to be deleted */
	            @cCriteria nvarchar(1000), /* criteria used to delete the rows required */
	            @iRowsAffected int OUTPUT /* number of records affected by the delete */
	            )
            As
            set nocount on
            declare 	@cTab nvarchar(255), /* name of the child table */
	            @cCol nvarchar(255), /* name of the linking field on the child table */
	            @cRefTab nvarchar(255), /* name of the parent table */
	            @cRefCol nvarchar(255), /* name of the linking field in the parent table */
	            @cFKName nvarchar(255), /* name of the foreign key */
	            @cSQL nvarchar(1000), /* query string passed to the sp_ExecuteSQL procedure */
	            @cChildCriteria nvarchar(1000), /* criteria to be used to delete 
                                                       records from the child table */
	            @iChildRows int /* number of rows deleted from the child table */

            /* declare the cursor containing the foreign key constraint information */
            DECLARE cFKey CURSOR LOCAL FOR 
            SELECT SO1.name AS Tab, 
                   SC1.name AS Col, 
                   SO2.name AS RefTab, 
                   SC2.name AS RefCol, 
                   FO.name AS FKName
            FROM dbo.sysforeignkeys FK  
            INNER JOIN dbo.syscolumns SC1 ON FK.fkeyid = SC1.id 
                                          AND FK.fkey = SC1.colid 
            INNER JOIN dbo.syscolumns SC2 ON FK.rkeyid = SC2.id 
                                          AND FK.rkey = SC2.colid 
            INNER JOIN dbo.sysobjects SO1 ON FK.fkeyid = SO1.id 
            INNER JOIN dbo.sysobjects SO2 ON FK.rkeyid = SO2.id 
            INNER JOIN dbo.sysobjects FO ON FK.constid = FO.id
            WHERE SO2.Name = @cTableName

            OPEN cFKey
            FETCH NEXT FROM cFKey INTO @cTab, @cCol, @cRefTab, @cRefCol, @cFKName
            WHILE @@FETCH_STATUS = 0
                 BEGIN
	            /* build the criteria to delete rows from the child table. As it uses the 
                       criteria passed to this procedure, it gets progressively larger with 
                       recursive calls */
	            SET @cChildCriteria = @cCol + ' in (SELECT [' + @cRefCol + '] FROM [' + 
                                          @cRefTab +'] WHERE ' + @cCriteria + ')'
	            print 'Deleting records from table ' + @cTab
	            /* call this procedure to delete the child rows */
	            EXEC spDeleteRows @cTab, @cChildCriteria, @iChildRows OUTPUT
	            FETCH NEXT FROM cFKey INTO @cTab, @cCol, @cRefTab, @cRefCol, @cFKName
                 END
            Close cFKey
            DeAllocate cFKey
            /* finally delete the rows from this table and display the rows affected  */
            SET @cSQL = 'DELETE FROM [' + @cTableName + '] WHERE ' + @cCriteria
            print @cSQL
            EXEC sp_ExecuteSQL @cSQL
            print 'Deleted ' + CONVERT(nvarchar, @@ROWCOUNT) + ' records from table ' + @cTableName            ";

        private const string DeleteDatabaseForced =
            @"
            EXEC msdb.dbo.sp_delete_database_backuphistory @database_name = N'{0}'
            
            USE [master]
            
            ALTER DATABASE [{0}] SET  SINGLE_USER WITH ROLLBACK IMMEDIATE
            
            USE [master]
            
            /****** Object:  Database [{0}]    Script Date: 05/28/2009 11:46:40 ******/
            DROP DATABASE [{0}]
            
            ";
    }
}
