﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Core.Databases;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Class representing a user in the system.
    /// </summary>
    public partial class User
    {
        partial void OnCreated()
        {
            this.Created = DateTime.Now;
            this.Modified = DateTime.Now;
            this.IsDisabled = false;
            this.ModuleId = 0;
            RecommenderStatistics.IncrimentUsersCreated();
        }
        
        /// <summary>
        /// Human readable representation of the user for tracing purposes
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return this.FirstName + " " + this.LastName + " (Id='" + this.UserId + "')";
        }

        /// <summary>
        /// Adds the property.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="name">The name.</param>
        /// <param name="value">The value.</param>
        public void AddProperty(MyMediaDataContext dataContext, string name, string value)
        {
            this.AddProperty(dataContext, dataContext.CreateProperty(name, value));
        }

        /// <summary>
        /// Adds the property.
        /// </summary>
        /// <param name="dataContext">The data context.</param>
        /// <param name="property">The property.</param>
        public void AddProperty(MyMediaDataContext dataContext, Property property)
        {
            if (!this.UserProperties.Any(item => item.Property.Name == property.Name && item.Property.Value == property.Value))
            {
                UserProperty itemProperty = new UserProperty();
                itemProperty.Property = property;
                this.UserProperties.Add(itemProperty);
            }
        }

        /// <summary>
        /// Adds the property.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        /// <param name="name">The name.</param>
        /// <param name="value">The value.</param>
        public void AddProperty(Recommender recommender, string name, string value)
        {
            this.AddProperty(recommender, recommender.CreateProperty(name, value));
        }

        /// <summary>
        /// Adds the property.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        /// <param name="property">The property.</param>
        public void AddProperty(Recommender recommender, Property property)
        {
            Contract.Precondition(recommender.PropertyExists(property.Name, property.Value), "Property must already exist in the database; use AddProperty(Recommender, string, string) instead");
            if (!this.UserProperties.Any(item => item.PropertyId == property.PropertyId))
            {
                UserProperty itemProperty = new UserProperty();
                itemProperty.PropertyId = property.PropertyId;
                this.UserProperties.Add(itemProperty);
            }
        }

        /// <summary>
        /// Updates the property.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        /// <param name="name">The name.</param>
        /// <param name="oldValue">The old value.</param>
        /// <param name="newValue">The new value.</param>
        public void UpdateProperty(Recommender recommender, string name, string oldValue, string newValue)
        {
            this.UpdateProperty(recommender, recommender.CreateProperty(name, oldValue), recommender.CreateProperty(name, newValue));
        }

        /// <summary>
        /// Updates the property.
        /// </summary>
        /// <param name="recommender">The recommender.</param>
        /// <param name="oldProperty">The old property.</param>
        /// <param name="newProperty">The new property.</param>
        public void UpdateProperty(Recommender recommender, Property oldProperty, Property newProperty)
        {
            //This is a bit of a workaround for lack of a way to refresh the User object after it's been fetched from the database.
            //The intent is to remove the intermediate table entry, UserProperty, as using the EntitySet<>.Remove method
            //causes an invalid state in that table (UserId may not be null).
            
            //Find the UserProperty which will change
            UserProperty propertyToRemove = this.UserProperties.SingleOrDefault(userProperty => userProperty.Property.PropertyId == oldProperty.PropertyId);
            if (default(UserProperty) != propertyToRemove)
            {
                //Remove the UserProperty row from the database.
                recommender.RemoveUserProperty(propertyToRemove.UserPropertyId);

                //Capture the existing, already fetched, UserProperties from the existing EntitySet<>
                UserProperty[] oldSet = this.UserProperties.ToArray();
                //Remove, or more exactly, take all of the UserProperties except the deleted one.
                oldSet = oldSet.Where(userProperty => userProperty.Property.PropertyId != oldProperty.PropertyId).ToArray();
                //Create a fresh entity set
                this._UserProperties = new EntitySet<UserProperty>(new Action<UserProperty>(this.attach_UserProperties), new Action<UserProperty>(this.detach_UserProperties));
                //Assign the new entity set, which has all the UserProperties except the deleted one.
                this._UserProperties.Assign(oldSet);
            }            
            //Add the changed propery.
            this.AddProperty(recommender, newProperty);
        }
    }
}
