﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Represents an expression of interest or enjoyment of a <see cref="CatalogItem"/> by a <see cref="User"/>.
    /// </summary>
    partial class RatingAction
    {
        /// <summary>
        /// Represents the smallest possible value of a Rating. This field is
        /// constant.
        /// </summary>
        public const double MinValue = -1.0;

        /// <summary>
        /// Represents the largest possible value of a Rating. This field is
        /// constant.
        /// </summary>
        public const double MaxValue = 1.0;

        /// <summary>
        /// Verifies the rating value is within range.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public static bool VerifyRatingValueIsWithinRange(double value)
        {
            return value >= RatingAction.MinValue && value <= RatingAction.MaxValue;
        }

        /// <summary>
        /// Scales the rating to the range [<see cref="RatingAction.MinValue"/>,<see cref="RatingAction.MaxValue"/>],
        /// appropriate for <see cref="RatingAction.Rating"/>.
        /// </summary>
        /// <param name="ratingAction">The rating action.</param>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        public static double ScaleRatingToRecommender(RatingAction ratingAction, double lowerRangeBound, double upperRangeBound)
        {
            Contract.Precondition(ratingAction.Rating.HasValue, "RatingAction property RatingValue has not be set.");
            return RatingAction.ScaleRatingToRecommender(ratingAction.Rating.Value, lowerRangeBound, upperRangeBound);
        }

        /// <summary>
        /// Scales the rating to the range [<see cref="RatingAction.MinValue"/>,<see cref="RatingAction.MaxValue"/>],
        /// appropriate for <see cref="RatingAction.Rating"/>.
        /// </summary>
        /// <param name="rating">The rating.</param>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        public static double ScaleRatingToRecommender(double rating, double lowerRangeBound, double upperRangeBound)
        {
            Contract.Precondition(upperRangeBound - lowerRangeBound > 0, "Invalid range.  Range must be > 0.");
            Contract.Precondition(upperRangeBound > lowerRangeBound, "Invalid range.  upperRangeBound must be > lowerRangeBound.");
            Contract.Precondition(rating <= upperRangeBound && rating >= lowerRangeBound, "Invalid value.  Value must be <= upperRangeBound AND >= lowerRangeBound");

            //First normalize to [0,1].  There may be a better way to do this in one shot, but nothing comes to mind.
            double value0to1 = (rating - lowerRangeBound) / (upperRangeBound - lowerRangeBound);

            //Now normalize to [-1,1].
            return (value0to1 * 2) - 1;
        }

        /// <summary>
        /// Scales the rating to the range [lowerBound, upperBound], as used by the application.
        /// </summary>
        /// <param name="ratingAction">The rating action.</param>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        public static double ScaleRatingToApplication(RatingAction ratingAction, double lowerRangeBound, double upperRangeBound)
        {
            Contract.Precondition(ratingAction.Rating.HasValue, "RatingAction property RatingValue has not be set.");
            return RatingAction.ScaleRatingToApplication(ratingAction.Rating.Value, lowerRangeBound, upperRangeBound);
        }

        /// <summary>
        /// Scales the rating to the range [lowerBound, upperBound], as used by the application.
        /// </summary>
        /// <param name="rating">The rating.</param>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        public static double ScaleRatingToApplication(double rating, double lowerRangeBound, double upperRangeBound)
        {
            Contract.Precondition(RatingAction.VerifyRatingValueIsWithinRange(rating), "ArgumentOutOfRange: ratingValue");

            double value0to1 = (rating + 1) / 2;
            double scaledValue = (value0to1 * (upperRangeBound - lowerRangeBound)) + lowerRangeBound;
            return scaledValue;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="RatingAction"/> class.  This class represents a user
        /// assigning a numeric rating to an item, which indicates how much they like that item.
        /// </summary>
        /// <param name="user">The user.</param>
        /// <param name="item">The item.</param>
        /// <param name="rating">The rating score, of the range [<see cref="RatingAction.MinValue"/>,<see cref="RatingAction.MaxValue"/>].</param>
        public RatingAction(User user, CatalogItem item, double rating) : 
            base(user, item)
        {
            this.Rating = rating;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="RatingAction"/> class.  This class represents a user
        /// assigning a numeric rating to an item, which indicates how much they like that item.
        /// </summary>
        /// <param name="user">The user.</param>
        /// <param name="item">The item.</param>
        /// <param name="ratingScore">The rating score, of the range [<see cref="RatingAction.MinValue"/>,<see cref="RatingAction.MaxValue"/>].</param>
        /// <param name="timestamp">The timestamp.</param>
        public RatingAction(User user, CatalogItem item, double ratingScore, DateTime timestamp) :
            base(user, item, timestamp)
        {
            this.Rating = ratingScore;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="RatingAction"/> class.  This class represents a user
        /// assigning a numeric rating to an item, which indicates how much they like that item.
        /// </summary>
        /// <param name="user">The user.</param>
        /// <param name="item">The item.</param>
        /// <param name="ratingScore">The rating score, of the range [<see cref="RatingAction.MinValue"/>,<see cref="RatingAction.MaxValue"/>].</param>
        /// <param name="virtualProfile">The virtual profile.</param>
        public RatingAction(User user, CatalogItem item, double ratingScore, VirtualProfile virtualProfile) :
            base(user, item)
        {
            this.Rating = ratingScore;
            this.VirtualProfile = virtualProfile;
        }

        /// <summary>
        /// Scales the rating to the range [lowerBound, upperBound], as used by the application.
        /// </summary>
        /// <param name="lowerRangeBound">The lower range bound.</param>
        /// <param name="upperRangeBound">The upper range bound.</param>
        /// <returns></returns>
        public double ScaleRatingToApplication(double lowerRangeBound, double upperRangeBound)
        {
            return RatingAction.ScaleRatingToApplication(this, lowerRangeBound, upperRangeBound);
        }

        partial void OnRatingChanging(System.Nullable<double> value)
        {
            Contract.Precondition(value.HasValue, "Rating does not have a value (value is null).  The rating must have a value.");
            Contract.Precondition(
                RatingAction.VerifyRatingValueIsWithinRange(value.Value),
                "{0} out of range. Value must be <= {1} and >= {2}.",
                this.GetType().ToString(), RatingAction.MaxValue, RatingAction.MinValue);
        }

        /// <summary>
        /// Lambda expression to determine if two UserAction derived classes are logically equivalent.
        /// </summary>
        /// <returns></returns>
        public override System.Linq.Expressions.Expression<Func<UserAction, bool>> EquivalenceQuery()
        {
            return action => this.VirtualProfileId.GetValueOrDefault(0) == action.VirtualProfileId.GetValueOrDefault(0);
        }

        /// <summary>
        /// Resolves the equivalent action.
        /// </summary>
        /// <param name="existingAction">The existing action.</param>
        public override void ResolveEquivalentAction(UserAction existingAction)
        {
            RatingAction existingRatingAction = existingAction as RatingAction;
            existingRatingAction.Rating = this.Rating;
            existingRatingAction.Timestamp = this.Timestamp;
        }

        /// <summary>
        /// Gets or sets the rating value.
        /// </summary>
        /// <value>The rating value.</value>
        [Obsolete("Use the property Rating instead.", true)]
        public Nullable<double> RatingValue { get; set; }
    }
}
