﻿using System;
using System.Collections.Generic;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;

namespace MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm
{
    /// <summary>
    /// Interface for recommendation algorithms which are based on a generalized entity relation paradigm.
    /// </summary>
    /// <remarks>
    /// Entity relationships are a generalized way to look at data.  The "nouns" represent the
    /// entities and the "verb" represents the relationship.
    /// <example>
    /// For example, <see cref="User"/> and <see cref="CatalogItem"/> are entities, and a
    /// </example>
    /// </remarks>
    public interface IEntityRelationEngine : IEntityRelationPredictor
    {
        /// <summary>
        /// Gets the id.
        /// </summary>
        /// <value>The id.</value>
        EngineId Id
        {
            get;
        }

        /// <summary>
        /// Gets the description.
        /// </summary>
        /// <value>The description.</value>
        string Description
        {
            get;
        }

        /// <summary>
        /// Gets or sets the entity relation data provider.
        /// </summary>
        /// <value>The entity relation data provider.</value>
        IEntityRelationDataProvider EntityRelationDataProvider
        {
            get;
            set;
        }

        /// <summary>
        /// Trains the recommender's internal model that will later on be queried using the PredictXXX methods.
        /// </summary>
        void Train();

        /// <summary>
        /// Should be replaced with the proper design pattern approach to asynchronicity.
        /// </summary>
        /// <returns></returns>
        bool IsTrainingCompleted();

        /// <summary>
        /// Training completed must be called by any IEngine when finished with the Train() method.
        /// </summary>
        EventHandler<EventArgs> TrainingCompleted
        {
            get;
            set;
        }

        /// <summary>
        /// Optinal method for the algorithm to learn from the discrepancy between actual and predicted value.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="predictedValue">The predicted value.</param>
        /// <param name="actualValue">The actual value.</param>
        void SubmitRelationFeedback(RelationType relationType, int[] key, double predictedValue, double actualValue);

        /// <summary>
        /// Online update methods for relations.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="values">The values.</param>
        void AddRelation(RelationType relationType, int[] key, object[] values);
        /// <summary>
        /// Updates the relation.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        /// <param name="values">The values.</param>
        void UpdateRelation(RelationType relationType, int[] key, object[] values);
        /// <summary>
        /// Removes the relation.
        /// </summary>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="key">The key.</param>
        void RemoveRelation(RelationType relationType, int[] key);

        /// <summary>
        /// Online update methods for entities
        /// </summary>
        /// <param name="entityType"></param>
        /// <param name="entityId"></param>
        /// <param name="values"></param>
        void AddEntity(EntityType entityType, int entityId, object[] values);
        /// <summary>
        /// Updates the entity.
        /// </summary>
        /// <param name="entityType">The entityType.</param>
        /// <param name="entityId">The entityId.</param>
        /// <param name="values">The values.</param>
        void UpdateEntity(EntityType entityType, int entityId, object[] values);
        /// <summary>
        /// Removes the entity.
        /// </summary>
        /// <param name="entityType">The entityType.</param>
        /// <param name="entityId">The entityId.</param>
        void RemoveEntity(EntityType entityType, int entityId);
    }
}
