﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.EntityRelationshipEngine;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm
{
    /// <summary>
    /// Implementation of an <see cref="Engine"/> which predicts the value of a <see cref="RatingAction"/> 
    /// for a <see cref="User"/>, for a specific <see cref="CatalogItem"/>.
    /// </summary>
    public class PredictRatingEngine : EntityRelationEngine
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="PredictRatingEngine"/> class, with the default <see cref="IEntityRelationDataProvider"/>,
        /// <see cref="MyMediaEntityRelationDataProvider"/>.
        /// </summary>
        /// <param name="algorithm">The algorithm.</param>
        public PredictRatingEngine(IEntityRelationRatingPredictor algorithm)
            : base(algorithm)
        {
            algorithm.MaxRatingValue = RatingAction.MaxValue;
            algorithm.MinRatingValue = RatingAction.MinValue;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="PredictRatingEngine"/> class.
        /// </summary>
        /// <param name="algorithm">The algorithm.</param>
        /// <param name="customDataProvider">The custom data provider.</param>
        public PredictRatingEngine(IEntityRelationRatingPredictor algorithm, IEntityRelationDataProvider customDataProvider)
            : base(algorithm, customDataProvider)
        {
            algorithm.MaxRatingValue = RatingAction.MaxValue;
            algorithm.MinRatingValue = RatingAction.MinValue;
        }

        /// <summary>
        /// Calculates a prediction (recommendation) for each catalog item in the query.
        /// </summary>
        /// <param name="query">The query.</param>
        /// <param name="user">The user.</param>
        /// <returns>a list of predictions.</returns>
        /// <remarks>
        /// This forms the heart of the recommendation algorithm. It has to generate a prediction for each catalog item in the query.
        /// </remarks>
        public override IList<Prediction<CatalogItem>> Predict(IEnumerable<CatalogItem> query, User user)
        {
            List<Prediction<CatalogItem>> predictions = new List<Prediction<CatalogItem>>();
            foreach (CatalogItem catalogItem in query)
            {
                double predictedValue = 0;
                try
                {
                    predictedValue = this.EntityRelationAlgorithm.PredictRelation(RelationType.Rated,
                        new int[] { user.UserId, catalogItem.ItemId });
                }
                catch (NullReferenceException e)
                {
                    TraceLog.OnException(e);
                }
                catch (IndexOutOfRangeException e)
                {
                    TraceLog.OnException(e);
                }
                Prediction<CatalogItem> prediction = new Prediction<CatalogItem>(catalogItem, predictedValue);
                predictions.Add(prediction);
            }
            return predictions;
        }
    }
}
