﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// A top-level and base class for framework events.
    /// </summary>
    /// <remarks>
    /// Framework events are a special kind of events where the actual management of subscribers is not done 
    /// by the instance that fires them (like in traditional C# events) but by a static management object. Subscribers 
    /// also do not subscribe to a specific event generator (like in traditional C#) but to a specific event type (that
    /// is derived from RecommenderEvent.
    /// This mechanisms decouples firing instances from subscribing instances and allows greater flexibility at runtime.
    /// Note that the inheritance hierarchy can be used to subscribe to specific events (siblings) or to parents of 
    /// sibling events.
    /// In the example below, a subscriber can subscribe to a DataChangedEvent and receive notification in both cases: when 
    /// DataAddedEvent is being fired and DataDeletedEvent is being fired.
    /// A subscriber can also subscribe to the DataDeletedEvent in case it only gets notified if this specific event is fired.
    /// </remarks>
    /// <example>
    /// <code>
    /// DataChangedEvent : RecommenderEvent
    /// {
    /// }
    /// 
    /// DataAddedEvent : DataChangedEvent
    /// {
    /// }
    /// 
    /// DataDeletedEvent: DataChangedEvent
    /// {
    /// }
    /// </code>
    /// </example>
    public class RecommenderEvent
    {
        /// <summary>
        /// Subscribes a handler to this event
        /// </summary>
        /// <param name="handler">Handler.</param>
        public void Subscribe(RecommenderEventHandler handler)
        {
            EventManager.AddSubscriber(this.GetType(), handler);
        }

        /// <summary>
        /// Unsubscribes a handler from this event.
        /// </summary>
        /// <param name="handler">Handler.</param>
        public void Unsubscribe(RecommenderEventHandler handler)
        {
            EventManager.RemoveSubscriber(this.GetType(), handler);
        }

        /// <summary>
        /// Fires an event of this type.
        /// </summary>
        /// <param name="args">Arguments that are passed to the handler.</param>
        public void Fire(RecommenderEventArgs args)
        {
            if (args.Type == null)
            {
                args.Type = this.GetType();
            }
            EventManager.Fire(this.GetType(), args);
        }
    }
}
