﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Used to allow releasing the data context after a query has been performed
    /// </summary>
    /// <typeparam name="T">Return type of query</typeparam>
    public class ReleasableQueryResult<T>: IReleasableQueryable<T>
    {
        private ICollection<IDisposable> releaseHandles;
        private IQueryable<T> query;

        /// <summary>
        /// Initializes a new instance of the <see cref="ReleasableQueryResult&lt;T&gt;"/> class.
        /// </summary>
        /// <param name="query">The query.</param>
        /// <param name="releaseHandles">The release handles.</param>
        public ReleasableQueryResult(IQueryable<T> query, params IDisposable[] releaseHandles)
        {
            this.releaseHandles = releaseHandles;
            this.query = query;
        }

        
        #region IEnumerable<T> Members

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1"/> that can be used to iterate through the collection.
        /// </returns>
        public IEnumerator<T> GetEnumerator()
        {
           return query.GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
           return query.GetEnumerator();
        }

        #endregion

        #region IQueryable Members

        /// <summary>
        /// Gets the type of the element(s) that are returned when the expression tree associated with this instance of <see cref="T:System.Linq.IQueryable"/> is executed.
        /// </summary>
        /// <value></value>
        /// <returns>
        /// A <see cref="T:System.Type"/> that represents the type of the element(s) that are returned when the expression tree associated with this object is executed.
        /// </returns>
        public Type ElementType
        {
            get { return query.ElementType; }
        }

        /// <summary>
        /// Gets the expression tree that is associated with the instance of <see cref="T:System.Linq.IQueryable"/>.
        /// </summary>
        /// <value></value>
        /// <returns>
        /// The <see cref="T:System.Linq.Expressions.Expression"/> that is associated with this instance of <see cref="T:System.Linq.IQueryable"/>.
        /// </returns>
        public System.Linq.Expressions.Expression Expression
        {
            get { return query.Expression; }
        }

        /// <summary>
        /// Gets the query provider that is associated with this data source.
        /// </summary>
        /// <value></value>
        /// <returns>
        /// The <see cref="T:System.Linq.IQueryProvider"/> that is associated with this data source.
        /// </returns>
        public IQueryProvider Provider
        {
            get { return query.Provider; }
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        private void Dispose(bool disposing)
        {
            if (disposing)
            {
                foreach (IDisposable handle in this.releaseHandles)
                {
                    handle.Dispose();
                }
            }
        }

        #endregion

        /// <summary>
        /// Replaces the query with the specified new query.
        /// </summary>
        /// <param name="newQuery">The new query.</param>
        public void Change(IQueryable<T> newQuery)
        {
            this.query = newQuery;
        }
    }
}
