﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data.Common;
using System.Data.Linq.Mapping;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Binary;
using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// Implements a versioned generic storage solution with notification support.
    /// </summary>
    /// <remarks>
    /// This class can be used to store any kind of information of type User. It keeps all data in memory and uses SQL Express
    /// for back-end storage.
    /// It behaves like a dictionary: keys are of type <see cref="RecommenderId"/> and values are of type User.
    /// Every change to the store is marked with a change number and is fired as an <see cref="RecommenderEvent"/>. Receiving components
    /// can then pull changes from this store based on the change numbers.
    /// Specialized versions of this class might fire specialized events.
    /// Note: TContType must be marked as serializable.
    /// </remarks>
    /// <example>
    /// <code>
    /// public MyChangeNotification : RecommenderEvent
    /// {
    ///   ....
    /// }
    /// 
    /// 
    /// [Serializable]
    /// public class MyClass
    /// {
    ///   ....
    /// }
    /// 
    /// public MyClassStore : UserActionStore&lt;MyClass&gt;
    /// {
    ///     protected void SendStoreChangedNotification(StoreChangedEventArgs storeChangedEventArgs)
    ///     {
    ///         new MyChangeNotification().Fire(storeChangedEventArgs);
    ///     }
    /// }
    /// </code>
    /// </example>
    public sealed class UserStore :
        IEnumerable<User>,
        IQueryable<User>,
        ITrackableDataStore,
        IDisposable
    {
        private PersistentStoreId storeId;
        private object locker = new object();
        private bool isInitialized;
        private string dbConnectionString;
        private MyMediaDataContext dataContext;
        private bool ownsDataContext;

        internal MyMediaDataContext DataContext
        {
            get { return dataContext; }
        }

        #region constructor and initialization

        /// <summary>
        /// Initializes a new instance of the <see cref="CatalogItemStore"/> class.
        /// </summary>
        /// <param name="programId">The program id.</param>
        /// <remarks>
        /// A store is identified by a program and a store id. The store id is used to distinguish multiple stores within an application.
        /// The program id is used to distinguish different program (instances).
        /// </remarks>
        public UserStore(ProgramId programId)
        {
            this.storeId = Constants.CatalogItemStoreId;
            this.dbConnectionString = Recommender.GetDatabaseConnectionString(programId);
            this.dataContext = new MyMediaDataContext(this.dbConnectionString);
            this.ownsDataContext = true;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="UserActionStore"/> class.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <remarks>
        /// A store is identified by a program and a store id. The store id is used to distinguish multiple stores within an application.
        /// The program id is used to distinguish different program (instances).
        /// </remarks>
        internal UserStore(MyMediaDataContext context)
        {
            this.storeId = Constants.UserStoreId;
            this.dataContext = context;
            this.ownsDataContext = false;
       }

        /// <summary>
        /// Initializes this instance.
        /// </summary>
        public void Initialize()
        {
            lock (locker)
            {
                if (!this.dataContext.DatabaseExists())
                {
                    TraceLog.Write(LogEventType.Information, "Recommender store database does not exist, creating it.");
                    this.dataContext.CreateDatabase();
                }

                isInitialized = true;
            }
        }

        /// <summary>
        /// Initializes and clears the store.
        /// </summary>
        public void InitializeAndClear()
        {
            lock (locker)
            {
                if (this.dataContext.DatabaseExists())
                {
                    this.ClearDatabaseContent();
                }
            }
            this.Initialize();
        }

        /// <summary>
        /// Returns whether or not the store exists.
        /// </summary>
        /// <returns></returns>
        public bool StoreExists()
        {
            bool storeExists = false;
            storeExists = this.dataContext.DatabaseExists();
            return storeExists;
        }

        /// <summary>
        /// Deletes the database.
        /// </summary>
        public void DeleteDatabase()
        {
            if (this.dataContext.DatabaseExists())
            {
                this.dataContext.DeleteDatabase();
            }
        }

        #endregion

        #region modification functions

        /// <summary>
        /// Adds the item.
        /// </summary>
        /// <param name="item">The item.</param>
        internal void AddItem(User item)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            InternalAddItems(new User[] { item });
        }

        /// <summary>
        /// Adds the items.
        /// </summary>
        /// <param name="newItems">The new items.</param>
        internal void AddItems(User[] newItems)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(newItems != null, "There has to be at least one item to add.");

            Contract.Precondition(newItems.Length != 0, "There has to be at least one item to add.");

            InternalAddItems(newItems);
        }

        /// <summary>
        /// Removes the item.
        /// </summary>
        /// <param name="itemId">The item id.</param>
        public void RemoveItem(int itemId)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(itemId != 0, "ItemKey can not be zero.");

            InternalRemoveItems(new int[] { itemId });
        }

        /// <summary>
        /// Removes the items.
        /// </summary>
        /// <param name="itemIds">The item ids.</param>
        public void RemoveItems(int[] itemIds)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            Contract.Precondition(itemIds != null, "There has to be at least one item to remove.");

            Contract.Precondition(itemIds.Length != 0, "There has to be at least one item to remove.");

            InternalRemoveItems(itemIds);
        }

        /// <summary>
        /// Removes all items.
        /// </summary>
        public void RemoveAll()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            this.ClearDatabaseContent();
        }
        #endregion

        #region query functions

        /// <summary>
        /// Gets the id.
        /// </summary>
        /// <value>The id.</value>
        public PersistentStoreId Id
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
                return storeId;
            }
        }

        /// <summary>
        /// Determines whether the UserActionStore contains an item with the specified item id.
        /// </summary>
        /// <param name="itemId">The item id.</param>
        /// <returns>
        /// 	<c>true</c> if it contains the specified item id; otherwise, <c>false</c>.
        /// </returns>
        public bool Contains(int itemId)
        {
            bool containsItem = false;
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            containsItem = null != this.dataContext.Users.FirstOrDefault<User>(c => c.UserId == itemId);
            return containsItem;
        }


        /// <summary>
        /// Gets the change number.
        /// </summary>
        /// <remarks>
        /// For an unchanged store, 0 will be returned.
        /// </remarks>
        /// <value>The change number.</value>
        public int ChangeNumber
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

                int changeNumber = 0;
                lock (locker)
                {
                    changeNumber = StoreTracker<UserStoreChange>.ChangeNumber(this.dataContext.UserStoreChanges);
                }
                return changeNumber;
            }
        }

        /// <summary>
        /// Gets the User with the specified item id.
        /// </summary>
        /// <value></value>
        public User this[int itemKey]
        {
            get
            {
                Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
                User userAction;

                userAction = this.dataContext.Users.Single<User>(c => c.UserId == itemKey);
                return userAction;
            }
        }

        /// <summary>
        /// Gets the differences.
        /// </summary>
        /// <param name="startChangeNumber">The start change number.</param>
        /// <param name="endChangeNumber">The end change number.</param>
        /// <returns>a list with all changes.</returns>
        /// <remark>
        /// Gets a list of all changes that make up the differences between the two given revisions starting with and including the
        /// change number specified in startChangeNumberand ending with and including the change number specified in endChangeNumber.
        /// All change numbers are ascending integer numbers, starting with 1 for the first change of the store.
        /// The list of changes does not contain every single change, but a list of all changes that have to be taken into
        /// account to get a picture of the diff between the startChangeNumber and the endChangeNumber.
        /// I.e. if in change number 50 an item with the id "a" gets added, and then gets removed in change number 60, the list of changes between 40 and 70 will not show
        /// this change because it does not make a difference for the state of the store.
        /// </remark>
        public IList<IDataStoreChange> GetDifferences(int startChangeNumber, int endChangeNumber)
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");

            IList<IDataStoreChange> differences;
            lock (locker)
            {
                differences = StoreTracker<UserStoreChange>.GetDifferences(this.dataContext.UserStoreChanges, startChangeNumber, endChangeNumber);
            }

            return differences;
        }
        #endregion

        #region IEnumerable Members

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref="T:System.Collections.IEnumerator"/> object that can be used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
            IEnumerator enumerator;
            enumerator = this.dataContext.Users.GetEnumerator();
            return enumerator;
        }

        #endregion

        #region IEnumerable<User> Members

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.Collections.Generic.IEnumerator`1"/> that can be used to iterate through the collection.
        /// </returns>
        IEnumerator<User> IEnumerable<User>.GetEnumerator()
        {
            Contract.Precondition(this.isInitialized, "Store is not initialized. Please call Initialize first.");
            IEnumerator<User> enumerator;
            enumerator = (from dbUserAction in this.dataContext.Users
                          where dbUserAction is User
                          select dbUserAction as User).GetEnumerator();

            return enumerator;
        }

        /// <summary>
        /// The number of User elements contained in the store.
        /// </summary>
        /// <remarks>
        /// This instance should be used in leiu of the IEnumerable extension count methods,
        /// which will not necessarily execute on the store server.
        /// </remarks>
        /// <returns>
        /// The number of User elements contained in the store.
        /// </returns>
        public int Count()
        {
            return this.dataContext.Users.Count<User>();
        }
        #endregion

        #region private functions

        /// <summary>
        /// Sends the store changed notification.
        /// </summary>
        /// <param name="storeChangedEventArgs">The <see cref="MyMediaProject.RecommenderSystem.Core.StoreChangedEventArgs"/> instance containing the event data.</param>
        private static void SendStoreChangedNotification(StoreChangedEventArgs storeChangedEventArgs)
        {
            new UserStoreChangedEvent().Fire(storeChangedEventArgs);
        }

        private void InternalAddItems(User[] newItems)
        {
            int newChangeNumber;
            lock (locker)
            {
                newChangeNumber = ChangeNumber + 1;

                foreach (User item in newItems)
                {
                    // Add item to the db
                    this.dataContext.Users.InsertOnSubmit(item);
                }
                this.dataContext.SubmitChanges();
                foreach (User item in newItems)
                {
                    Contract.Postcondition(0 != item.UserId);
                    // Add change to the db
                    StoreTracker<UserStoreChange>.AddChange(
                        this.dataContext.UserStoreChanges,
                        newChangeNumber,
                        StoreChangeType.Added,
                        item.UserId);
                }
                this.dataContext.SubmitChanges();
            }

            SendStoreChangedNotification(new StoreChangedEventArgs(storeId, newChangeNumber));
        }

        /// <summary>
        /// Submits the changed items.
        /// </summary>
        public void SubmitChangedItems()
        {
            IEnumerable<User> changedUsers = this.dataContext.GetChangeSet().Updates.Where(change => change is User).Cast<User>();

            int newChangeNumber;
            lock (locker)
            {
                newChangeNumber = ChangeNumber + 1;

                foreach (User item in changedUsers)
                {
                    // Add change to the db
                    StoreTracker<UserStoreChange>.AddChange(
                        this.dataContext.UserStoreChanges,
                        newChangeNumber,
                        StoreChangeType.Changed,
                        item.UserId);
                }

                this.dataContext.SubmitChanges();
            }
            SendStoreChangedNotification(new StoreChangedEventArgs(storeId, newChangeNumber));
        }

        private void InternalRemoveItems(int[] itemIds)
        {
            int newChangeNumber;
            lock (locker)
            {
                newChangeNumber = ChangeNumber + 1;

                foreach (int itemId in itemIds)
                {
                    // Remove item from db
                    this.dataContext.Users.DeleteOnSubmit(this.dataContext.Users.Single<User>(c => c.UserId == itemId));

                    // Add change to the db
                    StoreTracker<UserStoreChange>.AddChange(
                        this.dataContext.UserStoreChanges,
                        newChangeNumber,
                        StoreChangeType.Removed,
                        itemId);
                }

                this.dataContext.SubmitChanges();
            }
            SendStoreChangedNotification(new StoreChangedEventArgs(storeId, newChangeNumber));
        }

        private void ClearDatabaseContent()
        {
            if (this.dataContext.DatabaseExists())
            {
                this.dataContext.DeleteDatabase();
                this.dataContext.CreateDatabase();
            }
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        public void Dispose(bool disposing)
        {
            if (this.ownsDataContext && disposing)
            {
                this.dataContext.Connection.Dispose();
                this.dataContext.Dispose();
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        #region IQueryable Members

        /// <summary>
        /// Gets the type of the element(s) that are returned when the expression tree associated with this instance of <see cref="T:System.Linq.IQueryable"/> is executed.
        /// </summary>
        /// <value></value>
        /// <returns>
        /// A <see cref="T:System.Type"/> that represents the type of the element(s) that are returned when the expression tree associated with this object is executed.
        /// </returns>
        public Type ElementType
        {
            get { return (this.dataContext.Users as IQueryable).ElementType; }
        }

        /// <summary>
        /// Gets the expression tree that is associated with the instance of <see cref="T:System.Linq.IQueryable"/>.
        /// </summary>
        /// <value></value>
        /// <returns>The <see cref="T:System.Linq.Expressions.Expression"/> that is associated with this instance of <see cref="T:System.Linq.IQueryable"/>.</returns>
        public System.Linq.Expressions.Expression Expression
        {
            get { return (this.dataContext.Users as IQueryable).Expression; }
        }

        /// <summary>
        /// Gets the query provider that is associated with this data source.
        /// </summary>
        /// <value></value>
        /// <returns>
        /// The <see cref="T:System.Linq.IQueryProvider"/> that is associated with this data source.
        /// </returns>
        public IQueryProvider Provider
        {
            get { return (this.dataContext.Users as IQueryable).Provider; }
        }

        #endregion
    }
}