﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using System.Globalization;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Databases;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

using EMIC.Design;

namespace MyMediaProject.RecommenderSystem.Core
{
    /// <summary>
    /// The main entry point for a recommender framework instance for a specific catalog and feedback item type.
    /// </summary>
    /// <remarks>
    /// The recommender class is the main control class for an instance of a recommender framework. It is instantiated
    /// for a specific catalog and feedback item type that derive from resp. <see cref="CatalogItem"/> and
    /// <see cref="UserAction"/>. This class manages users, flows, providers and triggers recommendations.
    /// </remarks>
    public class Recommender : Recommender<UserAction>
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="Recommender"/> class.
        /// </summary>
        public Recommender(){}
        /// <summary>
        /// Initializes a new instance of the <see cref="Recommender&lt;TUserActionType&gt;"/> class.
        /// </summary>
        /// <remarks>
        /// The id is used to identify and distinguish stored data (e.g. user/engine settings, catalog/feedback items etc.).
        /// </remarks>
        /// <param name="id">The id.</param>
        public Recommender(ProgramId id)
            : base(id)
        {
        }

        /// <summary>
        /// Gets the database connection string.
        /// </summary>
        /// <param name="programId">The program id.</param>
        /// <returns></returns>
        static public string GetDatabaseConnectionString(ProgramId programId)
        {
            string dbConnectionString = Config.ConfigManager.DBConnectionString;
            if (string.IsNullOrEmpty(dbConnectionString))
            {
                dbConnectionString = string.Format(
                    CultureInfo.InvariantCulture,
                    Config.ConfigManager.DefaultDBConnectionStringFormat,
                    programId.FriendlyName,
                    programId.Value.ToString());
                TraceLog.Write(LogEventType.Information, "DB connection not specified in the configuration; using the default, \"{0}\"",
                    dbConnectionString);
                Config.ConfigManager.DBConnectionString = dbConnectionString;
            }
            return dbConnectionString;
        }
    }

    /// <summary>
    /// The main entry point for a recommender framework instance for a specific catalog and feedback item type.
    /// </summary>
    /// <typeparam name="TUserActionType">The feedback item type.</typeparam>
    /// <remarks>
    /// The recommender class is the main control class for an instance of a recommender framework. It is instantiated
    /// for a specific catalog and feedback item type that derive from resp. <see cref="CatalogItem"/> and
    /// <see cref="UserAction"/>. This class manages users, flows, providers and triggers recommendations.
    /// </remarks>
    public partial class Recommender<TUserActionType> : IDisposable
        where TUserActionType : UserAction
    {
        private List<CatalogItemFactory> registeredCatalogFactories;
        private Dictionary<FlowId, Flow<CatalogItem, TUserActionType>> registeredFlows;
        private ContextBroker contextBroker;

        /// <summary>
        /// Initializes a new instance of the <see cref="Recommender&lt;TUserActionType&gt;"/> class.
        /// </summary>
        internal Recommender() { }

        /// <summary>
        /// Initializes a new instance of the <see cref="Recommender&lt;TUserActionType&gt;"/> class.
        /// </summary>
        /// <remarks>
        /// The id is used to identify and distinguish stored data (e.g. user/engine settings, catalog/feedback items etc.).
        /// </remarks>
        /// <param name="id">The id.</param>
        internal Recommender(ProgramId id)
        {
            this.programId = id;
        }

        /// <summary>
        /// Initializes an instance of the Recommender class.
        /// </summary>
        public virtual void Initialize()
        {
            Contract.Precondition(!this.initialized, "Already initialized.");
            Contract.Precondition(!ActiveRecommenderIds.Contains(this.programId),
                "The framework does not support multiple instances of the same Recommender ProgramID.  Verify you are Disposing any previous instances used in this process.");

            ActiveRecommenderIds.Add(this.programId);

            Config.ConfigManager.DBConnectionString = Recommender.GetDatabaseConnectionString(this.programId);

            using (CatalogItemStoreDataContext dataContext = new CatalogItemStoreDataContext(Config.ConfigManager.DBConnectionString))
            {
                if (!dataContext.DatabaseExists())
                {
                    dataContext.CreateDatabase();
                }
            }

            this.registeredEngines = new List<Engine<CatalogItem, TUserActionType>>();
            this.registeredCatalogFactories = new List<CatalogItemFactory>();

            this.registeredFlows = new Dictionary<FlowId, Flow<CatalogItem, TUserActionType>>();

            this.contextBroker = new ContextBroker();

            this.initialized = true;
        }

        /// <summary>
        /// Initializes from configuration files.
        /// </summary>
        /// <remarks>
        /// This method uses configuration files to set up the recommender using the appropriate
        /// flows, hybridizers, and engines.
        /// </remarks>
        public virtual void InitializeFromConfiguration()
        {
            this.Initialize();
            throw new NotImplementedException("InitializeFromConfiguration not implemented");
        }

        /// <summary>
        /// Gets or sets the program id.  This field can only be set before <see cref="Initialize()"/> or
        /// <see cref="InitializeFromConfiguration()"/> are called.  The constructor, <see cref="Recommender(ProgramId)"/>
        /// will set this value automatically.
        /// </summary>
        /// <value>The program id.</value>
        public ProgramId ProgramId
        {
            get
            {            
                return this.programId;
            }
            set
            {
                Contract.Precondition(false == this.initialized, "The field ProgramId must be set before calling intialize.");
                Contract.Precondition(null != value, "null is an illegal value for the field ProgramId");
                this.programId = value;
            }
        }

        /// <summary>
        /// Resets all settings including those of the Engine (feedback, catalog items etc.).
        /// </summary>
        /// <remarks>
        /// Note: this method deletes all data (feedback, catalog items etc.) from all stores and basically resets the whole system.
        /// </remarks>
        public void ResetAllSettings()
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");

            foreach (Engine<CatalogItem, TUserActionType> engine in this.registeredEngines)
            {
                engine.ResetAllSettings();
            }
            using (CatalogItemStore store = new CatalogItemStore(this.CreateDataContext()))
            {
                store.InitializeAndClear();
            }
        }

        /// <summary>
        /// Deletes all stores, removing all data associated with this recommender application (feedback, catalog items etc.).
        /// This method may be called before the recommender has been initialized (for efficiency), however, 
        /// the system must be re-initialized after using this method.
        /// </summary>
        public void DeleteAllStores()
        {
            using (CatalogItemStoreDataContext dataContext = new CatalogItemStoreDataContext(Config.ConfigManager.DBConnectionString))
            {
                if (dataContext.DatabaseExists())
                {
                    dataContext.DeleteDatabase();
                }
            }
        }

        /// <summary>
        /// Determines if the MyMedia database exists.  Can be used before the framework is initialized.
        /// </summary>
        /// <returns></returns>
        public bool DatabaseExists()
        {
            using (CatalogItemStoreDataContext context = this.CreateDataContext())
            {
                return context.DatabaseExists();
            }
        }

        /// <summary>
        /// Empties all user's settings and their feedback.
        /// </summary>
        public void ResetUserSettings()
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");

            foreach (Engine<CatalogItem, TUserActionType> engine in this.registeredEngines)
            {
                engine.ResetLocalUserSettings();
            }
            this.DeleteAllStores();
        }

        /// <summary>
        /// Reprocesses all internal engine data and/or state.  Rebuilds the engine models.
        /// </summary>
        /// <remarks>
        /// This is used to force all engines back to a "clean" state, and 
        /// reprocess all of the items and feedback.
        /// </remarks>
        public void ReprocessEngineData()
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");

            foreach (Engine<CatalogItem, TUserActionType> engine in this.registeredEngines)
            {
                engine.ResetAllSettings();
            }
        }

        /// <summary>
        /// Processes feedback.
        /// </summary>
        /// <param name="feedbackItems">The feedback items.</param>
        /// <remarks>
        /// Feedback is (permanently) stored by the framework. A copy of the related catalog item is automatically
        /// (and permanently) stored. After storing it, the feedback is forwarded to all currently active engines that can
        /// then process it.
        /// It is possible to have multiple feedback items for a catalog item.
        /// </remarks>
        public void ProcessFeedback(IEnumerable<TUserActionType> feedbackItems)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");

            using (UserActionStore store = new UserActionStore(this.CreateDataContext()))
            {
                store.Initialize();
                store.ProcessFeedback(feedbackItems.Cast<UserAction>());
            }
        }

        /// <summary>
        /// Removes the feedback.
        /// </summary>
        /// <remarks>
        /// Feedback is (permanetly) deleted from the store, and will no longer affect user 
        /// recommendations.
        /// </remarks>
        /// <param name="feedbackItem">The feedback item to be removed.</param>
        public void RemoveFeedback(TUserActionType feedbackItem)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(null != feedbackItem, "Null parameter for feedbackItem not allowed.");

            lock (this.lockObject)
            {
                using (UserActionStore store = new UserActionStore(this.CreateDataContext()))
                {
                    store.Initialize();
                    store.RemoveItem(feedbackItem.Key);
                }
            }
        }

        /// <summary>
        /// Removes the feedback.
        /// </summary>
        /// <remarks>
        /// Feedback is (permanetly) deleted from the store, and will no longer affect user 
        /// recommendations.
        /// </remarks>
        /// <param name="feedbackItems">The feedback items to be removed.</param>
        public void RemoveFeedback(IEnumerable<TUserActionType> feedbackItems)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(null != feedbackItems, "Null parameter for feedbackItems not allowed.");

            lock (this.lockObject)
            {
                int[] feedbackIds = (from fbId in feedbackItems select fbId.Key).ToArray();
                using (UserActionStore store = new UserActionStore(this.CreateDataContext()))
                {
                    store.Initialize();
                    store.RemoveItems(feedbackIds);
                }
            }
        }

        /// <summary>
        /// Registers a catalog factory.
        /// </summary>
        /// <remarks>
        /// A <see cref="CatalogItemFactory"/> is responsible for adding/updating/deleting catalog items. Catalog items are the items
        /// for which a recommender generates predictions.
        /// </remarks>
        /// <param name="factory">The catalog item factory.</param>
        public void RegisterCatalogFactory(CatalogItemFactory factory)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(!this.registeredCatalogFactories.Contains(factory), "CatalogItemFactory already registered.");

            // connect the CatalogItemFactory to us
            factory.Initialize(this.programId);

            this.registeredCatalogFactories.Add(factory);
        }

        /// <summary>
        /// Unregisters the catalog factory.
        /// </summary>
        /// <param name="factory">The catalog item factory.</param>
        public void UnregisterCatalogFactory(CatalogItemFactory factory)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(this.registeredCatalogFactories.Contains(factory), "CatalogItemFactory not registered.");

            // disconnect the CatalogItemFactory
            factory.Uninitialize();

            this.registeredCatalogFactories.Remove(factory);
        }

        /// <summary>
        /// Exposes the catalog items as IEnumerable.
        /// </summary>
        /// <remarks>
        /// This enables iterations and LINQ queries against all catalog items. (Sub)Sets from this (another IEnumerable)
        /// can then be used as input for (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User, int, int)"/>) 
        /// (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User)"/>)
        /// </remarks>
        /// <value>All catalog items as IEnumerable.</value>
        public IReleasableQueryable<CatalogItem> CatalogItemStore
        {
            get
            {
                Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
                CatalogItemStore store = new CatalogItemStore(this.programId);
                store.Initialize();
                ReleasableQueryResult<CatalogItem> result = new ReleasableQueryResult<CatalogItem>(store, store);
                return result;
            }
        }

        /// <summary>
        /// Registers a flow.
        /// </summary>
        /// <remarks>
        /// A recommender generates prefictions for catalog items through a <see cref="Flow&lt;CatalogItem, TUserActionType&gt;"/> object. A flow object represents the
        /// prediction logic and is used when generating predictions (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User, int, int)"/>) 
        /// (<see cref="Recommender&lt;TUserActionType&gt;.Predict(IQueryable&lt;CatalogItem&gt;, FlowId, User)"/>).
        /// </remarks>
        /// <param name="flow">The flow.</param>
        public void RegisterFlow(Flow<CatalogItem, TUserActionType> flow)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(!this.registeredFlows.ContainsKey(flow.Id), "Flow already registered.");
            Contract.Precondition(!flow.Registered, "Flow already registered.");

            IContextSink contextSink;

            // register every prefilter that is a context sink with the context broker
            foreach (IPreFilter<CatalogItem> preFilter in flow.PreFilters)
            {
                contextSink = preFilter as IContextSink;
                if (contextSink != null)
                    contextBroker.AddContextSink(contextSink);
            }

            // register every postfilter that is a context sink with the context broker
            foreach (IPostFilter<CatalogItem> postFilter in flow.PostFilters)
            {
                contextSink = postFilter as IContextSink;
                if (contextSink != null)
                    contextBroker.AddContextSink(contextSink);
            }

            // if the hybridizer is a context sink, register it with the context broker
            contextSink = flow.Hybridizer as IContextSink;
            if (contextSink != null)
                contextBroker.AddContextSink(contextSink);

            foreach (Engine<CatalogItem, TUserActionType> engine in flow.Engines)
            {
                this.RegisterEngine(engine);
            }

            this.registeredFlows.Add(flow.Id, flow);
            flow.Registered = true;
        }

        /// <summary>
        /// Unregisters the flow.
        /// </summary>
        /// <param name="flow">The flow.</param>
        public void UnregisterFlow(Flow<CatalogItem, TUserActionType> flow)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(this.registeredFlows.ContainsKey(flow.Id), "Flow not registered.");
            Contract.Precondition(flow.Registered, "Flow not registered.");

            IContextSink contextSink;

            // unregister every prefilter that is a context sink from the context broker
            foreach (IPreFilter<CatalogItem> preFilter in flow.PreFilters)
            {
                contextSink = preFilter as IContextSink;
                if (contextSink != null)
                    contextBroker.RemoveContextSink(contextSink);

            }

            // unregister every postfilter that is a context sink from the context broker
            foreach (IPostFilter<CatalogItem> postFilter in flow.PostFilters)
            {
                contextSink = postFilter as IContextSink;
                if (contextSink != null)
                    contextBroker.RemoveContextSink(contextSink);

            }

            // if the hybridizer is a context sink, unregister it from the context broker
            contextSink = flow.Hybridizer as IContextSink;
            if (contextSink != null)
                contextBroker.RemoveContextSink(contextSink);

            foreach (Engine<CatalogItem, TUserActionType> engine in flow.Engines)
            {
                this.UnregisterEngine(engine);
            }

            this.registeredFlows.Remove(flow.Id);
            flow.Registered = false;
        }

        /// <summary>
        /// Adds users.
        /// </summary>
        /// <remarks>
        /// A user has to be added to the system before predictions can be generated for this user.
        /// </remarks>
        /// <param name="userIds">a list of user id's.</param>
        public void AddUsers(User[] userIds)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(userIds != null, "userId can not be null");

            lock (this.lockObject)
            {
                using (UserStore store = new UserStore(this.CreateDataContext()))
                {
                    store.Initialize();
                    foreach (User userId in userIds)
                    {
                        if (!store.Contains(userId.UserId))
                        {
                            store.AddItem(userId);
                        }
                        else
                        {
                            TraceLog.Write(LogEventType.Warning, "Attempted duplicate user addition: {0}.", userId);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Updates the user.
        /// </summary>
        /// <remarks>
        /// A user has to be added to the system before being updated.
        /// </remarks>
        [Obsolete("This method has been replaced by GetUpdateableUsers, please use this instead.", true)]
        public void UpdateUser()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Gets a collection of users which can be modified and updated.
        /// </summary>
        public UserUpdate GetUpdateableUsers()
        {
            return new UserUpdate(new UserStore(this.CreateDataContext()));
        }

        /// <summary>
        /// Removes a user.
        /// </summary>
        /// <remarks>
        /// Removing a user involves deleting all settings for this user.
        /// </remarks>
        /// <param name="userId">The user id.</param>
        public void RemoveUser(User userId)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(userId != null, "userId can not be null");


            lock (this.lockObject)
            {
                using (UserStore store = new UserStore(this.CreateDataContext()))
                {
                    store.Initialize();
                    Contract.Precondition(store.Contains(userId), string.Format(CultureInfo.InvariantCulture, "Unknown user: {0}", userId));
                    store.RemoveItem(userId.UserId);
                }
            }
        }

        /// <summary>
        /// Gets the users.
        /// </summary>
        /// <value>The users.</value>
        public IReleasableQueryable<User> Users
        {
            get
            {
                UserStore store = new UserStore(this.programId);
                store.Initialize();
                //store.DataContext.ObjectTrackingEnabled = false;
                return new ReleasableQueryResult<User>(store, store);
            }
        }
        
        /// <summary>
        /// Gets a user's feedback.
        /// </summary>
        /// <remarks>
        /// Returns a list of all feedback and original catalog items for a specific user.
        /// </remarks>
        public IReleasableQueryable<TUserActionType> GetUserFeedback(User userId)
        {
            Contract.Precondition(userId != null, "UserId cannot be null.");
            UserActionStore store = new UserActionStore(this.programId);
            IQueryable<TUserActionType> userFeedback = store.Where(feedback => feedback.UserId == userId.UserId && false == feedback.TestAction).Cast<TUserActionType>();
            ReleasableQueryResult<TUserActionType> releasableQuery = new ReleasableQueryResult<TUserActionType>(userFeedback, store);
            return releasableQuery;
        }

        /// <summary>
        /// Adds another context source to the recommender. 
        /// Context status changes of the added context sources will be propagated to all
        /// flow elements, that implement the IContextSink interface.
        /// </summary>
        /// <param name="contextSource">The context source to add.</param>
        public void AddContextSource(ContextSource contextSource)
        {
            contextBroker.AddContextSource(contextSource);
        }

        /// <summary>
        /// Removes the context source from the recommender.
        /// </summary>
        /// <param name="contextSource">The context source to remove.</param>
        public void RemoveContextSource(ContextSource contextSource)
        {
            contextBroker.RemoveContextSource(contextSource);
        }

        /// <summary>
        /// Generates predictions.
        /// </summary>
        /// <remarks>
        /// Predictions are generated using a <see cref="Flow&lt;CatalogItem, TUserActionType&gt;"/> class that defines the prediction logic over a set of items
        /// for a specific user.
        /// For performance reasons, it is advantageous to restrict the query set as much as possible prior to calling predict.
        /// Predictions against the entire catalog set, especially for large catalogs, may have performance implications.
        /// </remarks>
        /// <param name="query">The set of catalog items to generate predictions for.</param>
        /// <param name="flowId">The flow id.</param>
        /// <param name="userId">The user id.</param>
        /// <returns>A list of predictions.</returns>
        public IList<Prediction<CatalogItem>> Predict(IQueryable<CatalogItem> query, FlowId flowId, User userId)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(query != null, "query can not be null");
            Contract.Precondition(userId != null, "userId can not be null");
            Contract.Precondition(flowId != null, "flowId can not be null");
            Contract.Precondition(this.registeredFlows.ContainsKey(flowId), string.Format(CultureInfo.InvariantCulture, "Unknown flow: {0}", flowId));

            Flow<CatalogItem, TUserActionType> flow = this.registeredFlows[flowId];

            return flow.Process(query, userId); //this.userSettingsStore[userId.UserRecommenderId]);
        }

        /// <summary>
        /// Gets the name of the database used by the recommender.
        /// </summary>
        /// <value>The name of the database.</value>
        public string DatabaseName
        {
            get
            {
                using (System.Data.Common.DbConnection connection = new System.Data.SqlClient.SqlConnection(
                    Recommender.GetDatabaseConnectionString(this.programId)))
                {
                    return connection.Database;
                }
            }
        }

        /// <summary>
        /// Gets the database instance used by the recommender.
        /// </summary>
        /// <value>The database instance.</value>
        public string DatabaseSqlInstance
        {
            get
            {
                using (System.Data.Common.DbConnection connection = new System.Data.SqlClient.SqlConnection(
                    Recommender.GetDatabaseConnectionString(this.programId)))
                {
                    return connection.DataSource;
                }
            }
        }

        /// <summary>
        /// Generates predictions, returning the top items relevant for the user.
        /// </summary>
        /// <param name="query">The set of catalog items to generate predictions for.</param>
        /// <param name="flowId">The flow id.</param>
        /// <param name="userId">The user id.</param>
        /// <param name="skip">The number of items to skip in the ordering.</param>
        /// <param name="take">The take (i.e. "TopN").</param>
        /// <returns>A list of predictions.</returns>
        /// <remarks>
        /// Predictions are generated using a <see cref="Flow&lt;CatalogItem, TUserActionType&gt;"/> class that defines the prediction logic over a set of items
        /// for a specific user.
        /// This is the entry point for a "TopN" recommendation pattern.
        /// For performance reasons, it is advantageous to restrict the query set as much as possible prior to calling predict.
        /// Predictions against the entire catalog set, especially for large catalogs, may have performance implications.
        /// </remarks>
        public IList<Prediction<CatalogItem>> Predict(IQueryable<CatalogItem> query, FlowId flowId, User userId, int skip, int take)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(query != null, "query can not be null");
            Contract.Precondition(userId != null, "userId can not be null");
            Contract.Precondition(flowId != null, "flowId can not be null");
            Contract.Precondition(this.registeredFlows.ContainsKey(flowId), string.Format(CultureInfo.InvariantCulture, "Unknown flow: {0}", flowId));

            Flow<CatalogItem, TUserActionType> flow = this.registeredFlows[flowId];

            return flow.Process(query, userId, skip, take);
        }

        /// <summary>
        /// Given an identifier for the relation type (e.g. "rating") and the proper key, the predicted value
        /// for this relation is returned.
        /// </summary>
        /// <param name="flowId">The flow id.</param>
        /// <param name="entityType">Type of the entity.</param>
        /// <param name="keys">The keys.</param>
        /// <param name="count">The count.</param>
        /// <returns></returns>
        public int[] PredictEntity(FlowId flowId, EntityType entityType, int[] keys, int count)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(entityType != EntityType.Undefined, "entityType can not be null");
            Contract.Precondition(keys != null, "keys can not be null");
            Contract.Precondition(flowId != null, "flowId can not be null");
            Contract.Precondition(this.registeredFlows.ContainsKey(flowId), string.Format(CultureInfo.InvariantCulture, "Unknown flow: {0}", flowId));

            Flow<CatalogItem, TUserActionType> flow = this.registeredFlows[flowId];

            return flow.PredictEntity(entityType, keys, count);
        }

        /// <summary>
        /// This is the TopN method.
        /// </summary>
        /// <param name="flowId">The flow id.</param>
        /// <param name="relationType">Type of the relation.</param>
        /// <param name="keys">Combination of other relation's keys. E.g. for predicting top-n items, keys[0] contains the user_id for which top-n ratings should be predicted.</param>
        /// <returns>An ordered array of keys to that entity.</returns>
        public double PredictRelation(FlowId flowId, RelationType relationType, int[] keys)
        {
            Contract.Precondition(this.initialized, "Not initialized. Please call Initialize() first.");
            Contract.Precondition(relationType != RelationType.Undefined, "relationId can not be undefined.");
            Contract.Precondition(keys != null, "keys can not be null");
            Contract.Precondition(flowId != null, "flowId can not be null");
            Contract.Precondition(this.registeredFlows.ContainsKey(flowId), string.Format(CultureInfo.InvariantCulture, "Unknown flow: {0}", flowId));

            Flow<CatalogItem, TUserActionType> flow = this.registeredFlows[flowId];

            return flow.PredictRelation(relationType, keys);
        }

        internal Property CreateProperty(string name, string value)
        {
            using (CatalogItemStoreDataContext context = this.CreateDataContext())
            {
                Property property = context.CreateProperty(name, value);
                context.SubmitChanges();
            }
            using (CatalogItemStoreDataContext context = this.CreateDataContext())
            {
                context.ObjectTrackingEnabled = false;
                Property property = context.CreateProperty(name, value);
                return property;
            }
        }

        internal void RemoveUserProperty(int propertId)
        {
            using (CatalogItemStoreDataContext context = this.CreateDataContext())
            {
                context.UserProperties.DeleteOnSubmit(context.UserProperties.Single(userProp => userProp.UserPropertyId == propertId));
                context.SubmitChanges();
            }
        }

        internal bool PropertyExists(string name, string value)
        {
            using (CatalogItemStoreDataContext context = this.CreateDataContext())
            {
                return context.PropertyExists(name, value);
            }
        }

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        private void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (null != this.registeredFlows)
                {
                    foreach (Flow<CatalogItem, TUserActionType> flow in this.registeredFlows.Values.ToArray())
                    {
                        this.UnregisterFlow(flow);
                    }
                }
                ActiveRecommenderIds.Remove(this.programId);
            }
        }

        #region IDisposable Members

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        /// <summary>
        /// Releases unmanaged resources and performs other cleanup operations before the
        /// <see cref="Recommender&lt;TUserActionType&gt;"/> is reclaimed by garbage collection.
        /// </summary>
        ~Recommender()
        {
            this.Dispose(false);
        }
    }
}
