﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Data.Linq;
using System.Text;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Framework;

namespace MyMediaProject.RecommenderSystem.Framework.Metrics
{
    /// <summary>
    /// Maintains the functionality to produce a numeric representation of the utility of
    /// a recommender algorithm(s), as definied by the test sets and utility metrics used.
    /// </summary>
    /// <typeparam name="TQualifierUserActionType">The type of the qualifier user action type.</typeparam>
    /// <typeparam name="TMetricUserActionType">The type of the metric user action type.</typeparam>
    public class RecommenderMetrics<TQualifierUserActionType, TMetricUserActionType> : IDisposable
        where TQualifierUserActionType : UserAction where TMetricUserActionType : UserAction
    {
        private Recommender recommender;
        private TestSetQualifier<TQualifierUserActionType> testSetQualifier;
        private MetricBase<TMetricUserActionType>[] metrics;

        private bool metricsPartitioned = false;
        private bool testSetChanged = false;

        /// <summary>
        /// Initializes a new instance of the <see cref="RecommenderMetrics&lt;TQualifierUserActionType, TMetricUserActionType&gt;"/> class.
        /// </summary>
        /// <param name="programId">The program id of the recommender, used to identify the appropriate stores.</param>
        /// <param name="recommender">The recommender to use.</param>
        /// <param name="testSetQualifier">The test set qualifier class, for partitioning training and test data.</param>
        /// <param name="metrics">The metrics class for rating the recommendations.</param>
        public RecommenderMetrics(
            ProgramId programId,
            Recommender recommender, 
            TestSetQualifier<TQualifierUserActionType> testSetQualifier, 
            MetricBase<TMetricUserActionType>[] metrics)
        {
            Contract.Precondition(null != programId);
            Contract.Precondition(null != recommender);
            Contract.Precondition(null != testSetQualifier);
            Contract.Precondition(null != metrics);
            Contract.Precondition(0 < metrics.Count());

            this.recommender = recommender;
            this.testSetQualifier = testSetQualifier;
            this.metrics = metrics;
        }

        /// <summary>
        /// Removes all test feedback.
        /// </summary>
        public void RemoveAllTestFeedback()
        {
            this.recommender.ResetTestSet();
        }

        /// <summary>
        /// Gets the results for each metric.
        /// </summary>
        /// <param name="flowId">The flow id to compute metrics for.</param>
        /// <returns>A dictionary of metrics related to the final result.</returns>
        public ReadOnlyDictionary<RecommenderId, double> ProduceMetrics(FlowId flowId)
        {
            Dictionary<RecommenderId, double> metricResults =
                new Dictionary<RecommenderId, double>();

            this.PartitionTestSet();
            if(this.testSetChanged)
            {
                this.recommender.ReprocessEngineData();
                this.testSetChanged = false;
            }

            this.ComputeMetricsForEachUser(flowId);

            this.ComputeFinalMetrics(metricResults);

            return new ReadOnlyDictionary<RecommenderId, double>(metricResults);
        }

        private void ComputeFinalMetrics(Dictionary<RecommenderId, double> metricResults)
        {
            foreach (MetricBase<TMetricUserActionType> metric in this.metrics)
            {
                double result = metric.FinalMetric();
                metricResults.Add(metric.Id, result);
            }
        }

        /// <summary>
        /// Partitions the test set.
        /// </summary>
        public void PartitionTestSet()
        {
            int reportedPercentComplete = -1;
            int currentPercentComplete = 0;
            int count = 0;

            if (false == this.metricsPartitioned)
            {
                this.recommender.ResetTestSet();
                using (IReleasableQueryable<User> users = this.recommender.Users)
                {
                    TraceLog.Write(LogEventType.Information,
                        "Partitioning user training set for {0} users.",
                        users.Count());

                    foreach (User userId in users)
                    {
                        using (IReleasableQueryable<UserAction> allUserFeedbacks = this.recommender.GetUserFeedback(userId))
                        {
                            IQueryable<TQualifierUserActionType> userFeedbacks = allUserFeedbacks.Where(userFeedback => userFeedback is TQualifierUserActionType).Cast<TQualifierUserActionType>();
                            Collection<TQualifierUserActionType> testSet = new Collection<TQualifierUserActionType>();
                            this.testSetQualifier.ExtractTestSet(
                                userId,
                                userFeedbacks,
                                testSet);

                            if (0 != testSet.Count())
                            {
                                this.recommender.SeparateForTestSet(testSet.Cast<UserAction>());
                                this.testSetChanged = true;
                            }

                            count++;
                            currentPercentComplete = (count * 100) / users.Count();
                            if (currentPercentComplete > reportedPercentComplete)
                            {
                                reportedPercentComplete = currentPercentComplete;
                                TraceLog.Write(LogEventType.Verbose,
                                    "Partitioning user training set: {0} % complete.",
                                    reportedPercentComplete);
                            }
                        }
                    }
                }
            }

            TraceLog.Write(LogEventType.Information,
                "Partitioning user training set complete.");
            
            this.metricsPartitioned = true;
        }

        private void ComputeMetricsForEachUser(FlowId flowId)
        {
            int reportedPercentComplete = -1;
            int currentPercentComplete = 0;
            int count = 0;

            using (IReleasableQueryable<User> users = this.recommender.Users)
            {
                TraceLog.Write(LogEventType.Information,
                    "Computing metrics for {0} users.",
                    users.Count());

                foreach (User user in users)
                {
                    IList<Prediction<CatalogItem>> predictions = this.recommender.Predict(
                        this.recommender.CatalogItemStore,
                        flowId,
                        user);
                    using(IReleasableQueryable<UserAction> trainingFeedbacks = this.recommender.GetUserFeedback(user))
                    using(IReleasableQueryable<UserAction> testFeedbacks = this.recommender.GetUserTestFeedback(user))
                    {
                        IList<TMetricUserActionType> userTestFeedbacks = 
                            testFeedbacks.Where(testFeedback => testFeedback is TMetricUserActionType).Cast<TMetricUserActionType>().ToList();
                        IList<TMetricUserActionType> userTrainingFeedbacks =
                            trainingFeedbacks.Where(testFeedback => testFeedback is TMetricUserActionType).Cast<TMetricUserActionType>().ToList();
                        foreach (MetricBase<TMetricUserActionType> metric in this.metrics)
                        {
                            metric.ComputeMetricsForUser(
                                user,
                                userTestFeedbacks.AsQueryable(),
                                userTrainingFeedbacks.AsQueryable(),
                                predictions.ToArray<Prediction<CatalogItem>>());
                        }

                        count++;
                        currentPercentComplete = (count * 100) / users.Count();
                        if (currentPercentComplete > reportedPercentComplete)
                        {
                            reportedPercentComplete = currentPercentComplete;
                            TraceLog.Write(LogEventType.Verbose,
                                "Computing metrics: {0} % complete.",
                                reportedPercentComplete);
                        }
                    }
                }
            }
            TraceLog.Write(LogEventType.Information,
                "Computing metrics complete.");

        }

        #region IDisposable Members

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
            }
        }
    }
}