﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using MyMediaProject.RecommenderSystem.Framework.EntityRelationAlgorithm;

namespace MyMediaProject.RecommenderSystem.Algorithms
{
    /// <summary>
    /// Lightweight implementation of the data interface. 
    /// The implementation is for experimental purposes only.
    /// It works with CSV-files; no database like SQL is required.
    /// </summary>
    /// <author>Steffen Rendle, University of Hildesheim</author>
    class WP2Backend : IEntityRelationDataProvider
    {
        private static WP2Backend instance = null;

        public static WP2Backend GetInstance()
        {
            if (instance == null)
            {
                instance = new WP2Backend();
            }
            return instance;
        }
        
        private const EntityType ENTITY_ID_ITEM = EntityType.CatalogItem;
        private const EntityType ENTITY_ID_USER = EntityType.User;
        private const RelationType RELATION_ID_RATING = RelationType.Rated;
        private const RelationType RELATION_ID_VIRTUAL_PROFILE = RelationType.VirtualProfileRated;
        
        protected DataSetMemoryBased items;
        protected DataSetMemoryBased users;
        protected DataSetMemoryBased ratings;
        //protected DataSetMemoryBased virtual_profile;
        protected Dictionary<EntityType, IEntityRelationDataReader> entity_map = new Dictionary<EntityType, IEntityRelationDataReader>();
        protected Dictionary<RelationType , IEntityRelationDataReader> relation_map = new Dictionary<RelationType, IEntityRelationDataReader>();

        protected Dictionary<RelationType, HashSet<IEntityRelationEngine>> engine_relation_update = new Dictionary<RelationType, HashSet<IEntityRelationEngine>>();
        protected Dictionary<EntityType, HashSet<IEntityRelationEngine>> engine_entity_update = new Dictionary<EntityType, HashSet<IEntityRelationEngine>>();

        public WP2Backend()
        {
            ratings = new DataSetMemoryBased(
                new DataSetMemoryBased.DataType[] { DataSetMemoryBased.DataType.dt_int, DataSetMemoryBased.DataType.dt_int, DataSetMemoryBased.DataType.dt_double },
                "..\\..\\..\\Data\\ml1m-train.txt"
            );
            ratings.Open();
            /*virtual_profile = new DataSetMemoryBased(
                new DataSetMemoryBased.DataType[] { DataSetMemoryBased.DataType.dt_int, DataSetMemoryBased.DataType.dt_double, DataSetMemoryBased.DataType.dt_int, DataSetMemoryBased.DataType.dt_int},
                "..\\..\\data\\virtual_profile.txt"
            );
            virtual_profile.Open();
            */
            // find max_user_id, max_item_id:
            int max_item_id = 0;
            for (int i = 0; i < ratings.RowCount; i++)
            {
                max_user_id = Math.Max(max_user_id, ((DataSetMemoryBased.DataArrayInteger)ratings.data.columns[0]).data[i]);
                max_item_id = Math.Max(max_item_id, ((DataSetMemoryBased.DataArrayInteger)ratings.data.columns[1]).data[i]);
            }

            System.Console.WriteLine("max item id: " + max_item_id);
            System.Console.WriteLine("max user id: " + max_user_id);
            System.Console.WriteLine("#ratings:    " + ratings.RowCount);

            items = new DataSetMemoryBased(new DataSetMemoryBased.DataType[] { DataSetMemoryBased.DataType.dt_int }, max_item_id + 1);
            users = new DataSetMemoryBased(new DataSetMemoryBased.DataType[] { DataSetMemoryBased.DataType.dt_int }, max_user_id + 1);
            
            // fill data storage
            for (int i = 0; i <= max_item_id; i++)
            {
                ((DataSetMemoryBased.DataArrayInteger) items.data.columns[0]).data[i] = i;
            }
            for (int i = 0; i <= max_user_id; i++)
            {
                ((DataSetMemoryBased.DataArrayInteger) users.data.columns[0]).data[i] = i;
            }
            
            entity_map.Add(ENTITY_ID_ITEM, items);
            entity_map.Add(ENTITY_ID_USER, users);
            relation_map.Add(RELATION_ID_RATING, ratings);
            //relation_map.Add(RELATION_ID_VIRTUAL_PROFILE, virtual_profile);
            relation_map.Add(RelationType.Viewed, ratings);
        }


        // for the demo:
        public int max_user_id = -1;


        public void AddRelation(RelationType relation_id, int[] key, object[] values)
        {
            HashSet<IEntityRelationEngine> engines = null;
            engine_relation_update.TryGetValue(relation_id, out engines);
            if (engines != null)
            {
                foreach (IEntityRelationEngine engine in engines)
                {
                    engine.AddRelation(relation_id, key, values);
                }
            }
        }

        public void UpdateRelation(RelationType relation_id, int[] key, object[] values)
        {
            HashSet<IEntityRelationEngine> engines = null;
            engine_relation_update.TryGetValue(relation_id, out engines);
            if (engines != null)
            {
                foreach (IEntityRelationEngine engine in engines)
                {
                    engine.UpdateRelation(relation_id, key, values);
                }
            }
        }

        public void RemoveRelation(RelationType relation_id, int[] key)
        {
            HashSet<IEntityRelationEngine> engines = null;
            engine_relation_update.TryGetValue(relation_id, out engines);
            if (engines != null)
            {
                foreach (IEntityRelationEngine engine in engines)
                {
                    engine.RemoveRelation(relation_id, key);
                }
            }
        }

        public IEntityRelationDataReader GetRelation(RelationType relation_id)
        {
            return GetRelation(relation_id, new int[] { });
        }

        public IEntityRelationDataReader GetRelation(RelationType relation_id, int[] attributes)
        {
            if (relation_map.ContainsKey(relation_id))
            {
                IEntityRelationDataReader result;
                relation_map.TryGetValue(relation_id, out result);
                result.Open();
                return result;
            }
            else
            {
                throw new Exception("unknown relation_id");
            }
        }

        public void RegisterForRelationUpdates(RelationType relation_id, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engine_list = null;
            engine_relation_update.TryGetValue(relation_id, out engine_list);
            if (engine_list == null)
            {
                engine_list = new HashSet<IEntityRelationEngine>();
                engine_relation_update.Add(relation_id, engine_list);
            }
            engine_list.Add(engine);
        }

        public void RegisterForEntityUpdates(EntityType entity_id, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engine_list = null;
            engine_entity_update.TryGetValue(entity_id, out engine_list);
            if (engine_list == null)
            {
                engine_list = new HashSet<IEntityRelationEngine>();
                engine_entity_update.Add(entity_id, engine_list);
            }
            engine_list.Add(engine);
        }

        public void UnregisterForRelationUpdates(RelationType relation_id, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engine_list = null;
            engine_relation_update.TryGetValue(relation_id, out engine_list);
            if (engine_list != null)
            {
                engine_list.Remove(engine);
            }         
        }

        public void UnregisterForEntityUpdates(EntityType entity_id, IEntityRelationEngine engine)
        {
            HashSet<IEntityRelationEngine> engine_list = null;
            engine_entity_update.TryGetValue(entity_id, out engine_list);
            if (engine_list != null)
            {
                engine_list.Remove(engine);
            }
        }

        public IEntityRelationDataReader GetEntity(EntityType entityType)
        {
            return GetEntity(entityType, new int[] {});
        }

        public IEntityRelationDataReader GetEntity(EntityType entityType, int[] attributes)
        {
            if (entity_map.ContainsKey(entityType))
            {
                IEntityRelationDataReader result;
                entity_map.TryGetValue(entityType, out result);
                result.Open();
                return result;
            }
            else
            {
                throw new Exception("unknown relation_id");
            }
        }
    }
}
