﻿using System;
using System.IO;
using System.Configuration;
using System.Globalization;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using MyMediaProject.RecommenderSystem.Core.Config;
using System.Reflection;

using MyMediaProject.RecommenderSystem.Core;

namespace ConfigGui
{
    /// <summary>
    /// Graphical User Interface allowing for viewing and changing program's configuration.
    /// </summary>
    public partial class FrmConfigGui : Form
    {
        private const String APP_SETTING_NAME_PATH_TO_CONF_DIR = "PathToConfigDir";
        private const String APP_SETTING_NAME_PREVIOUS_PATH_TO_CONF_DIR = "PathToConfigDir_previous";
        
        private const String UTILITY_APPLICATION_EXE_FILE_NAME = "ConfigGui.exe";

        private const String BREAK_LINE_CHARS = "\n";
        private const int GRID_COLUMNS_MAX_INDEX = 4; // [number of columns in the grid] - 1
        private const int GRID_COLUMN_INDEX_NAME = 0;
        private const int GRID_COLUMN_INDEX_VALUE = 1;
        private const int GRID_COLUMN_INDEX_DEFAULT_VALUE = 2;
        private const int GRID_COLUMN_INDEX_DESCRIPTION = 3;

        private const int GRID_COLUMN_INDEX_RESET_BUTTON = 4;

        private List<ConfigModule> sortedModules;
        private bool[] isAnyValueChanged;

        private DataGridView[] grids;

        /// <summary>
        /// Constructor
        /// </summary>
        public FrmConfigGui()
        {
            InitializeComponent();
        }


        private static System.Configuration.Configuration appConfigObject;
        private static System.Configuration.Configuration getApplicationConfigurationObject()
        {
            if (appConfigObject != null)
            {
                return appConfigObject;
            }
            appConfigObject = ConfigurationManager.OpenExeConfiguration(UTILITY_APPLICATION_EXE_FILE_NAME);
            return appConfigObject;
        }

        /// <summary>
        /// Stores local application setting of the ConfigGui application to the application specific storage (does not affect MyMedia framework)
        /// </summary>
        /// <param name="settingName"></param>
        /// <param name="settingValue"></param>
        private static void SaveApplicationSetting(string settingName, string settingValue)
        {
            // Get the configuration file.
            System.Configuration.Configuration appConfig = getApplicationConfigurationObject();

            // if the setting exists already, remove it first (otherwise its value would be appended)
            if (appConfig.AppSettings.Settings.AllKeys.Contains(settingName))
            {
                RemoveApplicationSetting(settingName);
            }

            appConfig.AppSettings.Settings.Add(settingName, settingValue);
            _save_and_reload_application_config();
        }

        private static void RemoveApplicationSetting(string settingName)
        {
            System.Configuration.Configuration appConfig = getApplicationConfigurationObject();
            if (!appConfig.AppSettings.Settings.AllKeys.Contains(settingName))
            {
                return;
            }

            appConfig.AppSettings.Settings.Remove(settingName);
            _save_and_reload_application_config();
        }

        private static void _save_and_reload_application_config(){
            // Save the configuration file.
            getApplicationConfigurationObject().Save(ConfigurationSaveMode.Minimal);
            // Force a reload of the changed section.
            ConfigurationManager.RefreshSection("appSettings");
        }

        private static String GetApplicationSetting(string settingName)
        {
            return ConfigurationManager.AppSettings[settingName];
        }


        private void frmConfigGui_Load(object sender, EventArgs e)
        {
            String PathToConfigDir = GetApplicationSetting(APP_SETTING_NAME_PATH_TO_CONF_DIR);
            
            if (PathToConfigDir == null || PathToConfigDir.Length == 0 || !Directory.Exists(PathToConfigDir))
            {
                FolderSpecificationControl folderControl = new FolderSpecificationControl();

                try
                {
                    PathToConfigDir = folderControl.AskConfigurationFolderPath(false);
                }
                catch (OperationCanceledException)
                {
                    PathToConfigDir = null;
                }

                if (String.IsNullOrEmpty(PathToConfigDir))
                {
                    // try to use the previous value of the setting, if it exists
                    PathToConfigDir = GetApplicationSetting(APP_SETTING_NAME_PREVIOUS_PATH_TO_CONF_DIR);
                }

                // if the folder is chosen, or if there is a previous value of that setting
                if (PathToConfigDir != null && PathToConfigDir.Length > 0)
                {
                    SaveApplicationSetting(APP_SETTING_NAME_PATH_TO_CONF_DIR, PathToConfigDir);
                }
                else
                {
                    MessageBox.Show("Error: the path to the configuration folder is not set. It is not an optional parameter. The program will now quit.");
                    Application.Exit();
                    return;
                }
            }

            textBoxCurrentConfDir.Text = PathToConfigDir;
            
            ConfigManager.Initialize(PathToConfigDir, true);

            // to display modules in a user friendly order and manipulate them by index, we need to sort them
            sortedModules = ConfigManager.Modules.Values.ToList<ConfigModule>();
            sortedModules.Sort();
            isAnyValueChanged = new bool[sortedModules.Count];

            grids = new DataGridView[sortedModules.Count];
            System.Windows.Forms.Label[] lblModulesDescriptions = new System.Windows.Forms.Label[sortedModules.Count];

            for (int i = 0; i < sortedModules.Count; i++)
            {
                tabModulesList.TabPages.Add(sortedModules.ElementAt(i).ShortName);
                lblModulesDescriptions[i] = new System.Windows.Forms.Label();
                lblModulesDescriptions[i].Location = new System.Drawing.Point(10, 8);
                lblModulesDescriptions[i].Width = 700; // hardcoded length of description label in px
                lblModulesDescriptions[i].Height = 26; // hardcoded length of description label in px
                lblModulesDescriptions[i].Text = sortedModules.ElementAt(i).Description;
                tabModulesList.TabPages[i].Controls.Add(lblModulesDescriptions[i]);
            }
            for (int i = 0; i < sortedModules.Count; i++)
            {
                InstantiatePropertiesGrid(i);
                isAnyValueChanged[i] = false;
            }

            this.btnApply.Enabled = false;
        }

        private void InstantiatePropertiesGrid(int iTabIndex){
            this.tabModulesList.TabPages[iTabIndex].ResumeLayout(false);

            grids[iTabIndex] = new System.Windows.Forms.DataGridView();

            this.PropertyName = new System.Windows.Forms.DataGridViewTextBoxColumn();
            this.PropertyValue = new System.Windows.Forms.DataGridViewTextBoxColumn();
            this.DefaultValue = new System.Windows.Forms.DataGridViewTextBoxColumn();
            this.Description = new System.Windows.Forms.DataGridViewTextBoxColumn();

            this.ResetToDefault = new System.Windows.Forms.DataGridViewButtonColumn();

            grids[iTabIndex].ColumnHeadersHeightSizeMode = System.Windows.Forms.DataGridViewColumnHeadersHeightSizeMode.AutoSize;
            grids[iTabIndex].Columns.AddRange(new System.Windows.Forms.DataGridViewColumn[] {
                                this.PropertyName,
                                this.PropertyValue,
                                this.DefaultValue,
                                this.Description,
                                this.ResetToDefault});
            grids[iTabIndex].Location = new System.Drawing.Point(0, 42);
            grids[iTabIndex].Name = "dataGridView" + iTabIndex.ToString(new NumberFormatInfo());
            grids[iTabIndex].Size = new System.Drawing.Size(747, 326); // 751, 326 // 747, 229
            grids[iTabIndex].Columns[GRID_COLUMN_INDEX_VALUE].Width *= 2;
            grids[iTabIndex].Columns[GRID_COLUMN_INDEX_DESCRIPTION].Width *= 2;

            grids[iTabIndex].TabIndex = 0;
            // 
            // PropertyName
            // 
            this.PropertyName.HeaderText = "Property name";
            this.PropertyName.Name = "PropertyName";
            this.PropertyName.ReadOnly = true;
            // 
            // PropertyValue
            // 
            this.PropertyValue.HeaderText = "Property Value";
            this.PropertyValue.Name = "PropertyValue";
            // 
            // DefaultValue
            // 
            this.DefaultValue.HeaderText = "Default Value";
            this.DefaultValue.Name = "DefaultValue";
            this.DefaultValue.ReadOnly = true;
            // 
            // description
            // 
            this.Description.HeaderText = "Description";
            this.Description.Name = "description";
            this.Description.ReadOnly = true;
            // 
            // ResetToDefault
            // 
            this.ResetToDefault.HeaderText = "Reset to default";
            this.ResetToDefault.Name = "ResetToDefault";
            this.ResetToDefault.Text = "Reset to default";
            this.ResetToDefault.UseColumnTextForButtonValue = true;

            this.ResumeLayout(false);
            this.tabModulesList.TabPages[iTabIndex].Controls.Add(grids[iTabIndex]);

            Object[] gridRow = new Object[GRID_COLUMNS_MAX_INDEX];

            int iRow = 0;
            foreach (ConfigProperty property in sortedModules.ElementAt(iTabIndex).Properties.Values)
            {
                gridRow[GRID_COLUMN_INDEX_NAME] = property.Name;
                gridRow[GRID_COLUMN_INDEX_VALUE] = property.Value;
                gridRow[GRID_COLUMN_INDEX_DEFAULT_VALUE] = property.DefaultValue;
                gridRow[GRID_COLUMN_INDEX_DESCRIPTION] = property.Description;
                grids[iTabIndex].Rows.Add(gridRow);

                if (property.IsChanged)
                {
                    grids[iTabIndex].Rows[iRow].Cells[GRID_COLUMN_INDEX_VALUE].Style.ForeColor = Color.Red;
                }
                iRow++;
            }

            grids[iTabIndex].CellContentClick += new System.Windows.Forms.DataGridViewCellEventHandler(this.dataGridView1_CellContentClick);
            grids[iTabIndex].AllowUserToAddRows = false;

            // the update-event must be the registered AFTER grid is initialized.
            grids[iTabIndex].CellValueChanged += new System.Windows.Forms.DataGridViewCellEventHandler(this.dataGridView1_CellValueChanged);
        }

        private void btnCancel_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void dataGridView1_CellContentClick(object sender, DataGridViewCellEventArgs e)
        {
            if (e.RowIndex < 0)
            {
                return;
            }

            int tabActiveIndex = this.tabModulesList.SelectedIndex;
            if (e.ColumnIndex == GRID_COLUMN_INDEX_RESET_BUTTON)
            {
                ConfigModule module = sortedModules.ElementAt(tabActiveIndex);
                String propertyName = grids[tabActiveIndex].Rows[e.RowIndex].Cells[GRID_COLUMN_INDEX_NAME].Value.ToString();
                module.ResetPropertyToDefault(propertyName);
                AddStatusTextLine("Property '" + propertyName + "' from the '" + module.Name + "' module has been reset.");
                NoteModuleDataModification(tabActiveIndex);
                RefreshModuleGridView(tabActiveIndex);
                return;
            }
        }

        /// <summary>
        /// Enables 'Apply' button, and schedules the module's data for potential future update.
        /// </summary>
        /// <param name="moduleIndex"></param>
        private void NoteModuleDataModification(int moduleIndex)
        {
            isAnyValueChanged[moduleIndex] = true;
            this.btnApply.Enabled = true;
        }

        private void dataGridView1_CellValueChanged(object sender, DataGridViewCellEventArgs e)
        {
            NoteModuleDataModification(this.tabModulesList.SelectedIndex);
        }

        /// <summary>
        /// Redraws the list of module's properties according to the module's state in the memory
        /// </summary>
        /// <param name="moduleIndex">The </param>
        private void RefreshModuleGridView(int moduleIndex)
        {
            // re-fill the rows
            grids[moduleIndex].Rows.Clear();
            Object[] gridRow = new Object[GRID_COLUMNS_MAX_INDEX];
            ConfigModule module = sortedModules.ElementAt(moduleIndex);

            foreach (ConfigProperty property in module.Properties.Values)
            {
                gridRow[GRID_COLUMN_INDEX_NAME] = property.Name;
                gridRow[GRID_COLUMN_INDEX_VALUE] = property.Value;
                gridRow[GRID_COLUMN_INDEX_DEFAULT_VALUE] = property.DefaultValue;
                gridRow[GRID_COLUMN_INDEX_DESCRIPTION] = property.Description;
                grids[moduleIndex].Rows.Add(gridRow);
            }
        }

        /// <summary>
        /// Saves configurations of all changed modules.
        /// </summary>
        private void SaveAllChangedConfigModules()
        {
            for( int i = 0; i < sortedModules.Count; i++){
                if (isAnyValueChanged[i])
                {
                    SaveConfigForModule(i);
                }
            }
            AddStatusTextLine("Configuration of all modules has been saved.");
        }

        /// <summary>
        /// Saves configuration of the specified module
        /// </summary>
        /// <param name="moduleIndex"></param>
        private void SaveConfigForModule(int moduleIndex)
        {
            foreach (DataGridViewRow row in grids[moduleIndex].Rows){
                if (row.Cells.Count == 0 || row.Cells[GRID_COLUMN_INDEX_NAME] == null 
                    || row.Cells[GRID_COLUMN_INDEX_NAME].Value == null 
                    || row.Cells[GRID_COLUMN_INDEX_NAME].Value.ToString().Trim().Length == 0)
                {
                    break;
                }
                String propertyName = row.Cells[GRID_COLUMN_INDEX_NAME].Value.ToString(); // name can't be null
                String propertyValue = GetCellValueOrEmptyString(row, GRID_COLUMN_INDEX_VALUE);
                String propertyDescrition = GetCellValueOrEmptyString(row, GRID_COLUMN_INDEX_DESCRIPTION);

                ConfigProperty updatedProperty = new ConfigProperty(propertyName, propertyValue, propertyDescrition);
                sortedModules.ElementAt(moduleIndex).AddOrUpdateProperty(updatedProperty);
            }

            ConfigManager.SaveModuleConfiguration(sortedModules.ElementAt(moduleIndex).Name);
            isAnyValueChanged[moduleIndex] = false;
            AddStatusTextLine("Configuration of the '" + sortedModules.ElementAt(moduleIndex).ShortName + "' module saved.");
        }

        /// <summary>
        /// Utility method to escape ToString() call on null object (which could be the value of the cell)
        /// </summary>
        /// <param name="row"></param>
        /// <param name="rowIndex"></param>
        /// <returns></returns>
        private static String GetCellValueOrEmptyString(DataGridViewRow row, int rowIndex)
        {
            String result = "";
            if (row.Cells[rowIndex].Value != null)
            {
                result = row.Cells[rowIndex].Value.ToString();
            }

            return result;
        }

        private void btnAccept_Click(object sender, EventArgs e)
        {
            SaveAllChangedConfigModules();

            Application.Exit();
        }

        private void btnApply_Click(object sender, EventArgs e)
        {
            SaveAllChangedConfigModules();
        }

        private void dataGridView1_CellBeginEdit(object sender, DataGridViewCellCancelEventArgs e)
        {
            NoteModuleDataModification(this.tabModulesList.SelectedIndex);
        }

        /// <summary>
        /// Adds a new row into the status box
        /// </summary>
        /// <param name="msg"></param>
        private void AddStatusTextLine(String msg)
        {
            richTextBoxStatus.Text = msg + BREAK_LINE_CHARS + richTextBoxStatus.Text;
        }

        /// <summary>
        /// Saves empty string as the value of the APP_SETTING_NAME_PATH_TO_CONF_DIR
        /// (which is needed to ask the user to specify its location, because the value is empty) 
        /// and restarts the application.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnChangeConfigDir_Click(object sender, EventArgs e)
        {
            SaveApplicationSetting(APP_SETTING_NAME_PREVIOUS_PATH_TO_CONF_DIR, GetApplicationSetting(APP_SETTING_NAME_PATH_TO_CONF_DIR));
            RemoveApplicationSetting(APP_SETTING_NAME_PATH_TO_CONF_DIR);
            Application.Restart();

            // activate the window
            this.Activate();
            this.BringToFront();
            this.Show();
            this.TopMost = true;
            btnChangeConfigDir.Focus();
        }

        private void CreateDbButton_Click(object sender, EventArgs e)
        {
            String dbConnectionString = null;
            try{
                dbConnectionString = ConfigManager.DBConnectionString;
            }catch(System.Collections.Generic.KeyNotFoundException){
                MessageBox.Show("The property '" + ConfigConstants.PropertyNameDBConnectionString + "' is not configured for the default module ('" + ConfigConstants.DefaultModuleName + "')");
                return;
            }

            System.Data.Linq.DataContext context = new MyMediaDataContext(dbConnectionString);
            if (context.DatabaseExists())
            {
                DialogResult result = MessageBox.Show("The Database '" + context.Connection.Database + "' exists already.\r\nDo you want to re-create it (All the entries will be lost)?", "Warning", MessageBoxButtons.YesNo);
                if (DialogResult.Yes.Equals(result))
                {
                    context.DeleteDatabase();
                    context.CreateDatabase();
                    MessageBox.Show("The database '" + context.Connection.Database + "' has been re-created!");
                }
                return;
            }
            else
            {
                context.CreateDatabase();
                MessageBox.Show("The database '" + context.Connection.Database + "' has been created!");
            }
        }
    }
}
