using System;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Framework;
using System.Data.Linq;
using System.Data;
using System.Collections.Generic;
using System.Linq;
using System.Globalization;
using System.Xml;
using MyMediaProject.RecommenderSystem.Core.Databases;
using DatabaseInterface;

namespace mymedia
{
    /// <summary>
    /// Class for reading descriptions from CatalogItems from the MyMedia recommender framework.
    /// Descriptions are passed to the UIMAWrapper
    /// </summary>
    public class KeywordExtractor
    {
        private String name;
        private String organization;
        private String dependencyName;
        private String dependencyOrganization;
        private String descriptor;
        private Boolean processAll = false;
        private DateTime since = DateTime.MinValue;
        private List<AnnotationType> annotations;
        private UIMAWrapper wrapper;
        MyMediaDatabaseInterface mminterface;

        /// <summary>
        /// Constructor, initializes the MyMediaDatabaseInterface.
        /// </summary>
        public KeywordExtractor()
        {
            mminterface = new MyMediaDatabaseInterface();
            annotations = new List<AnnotationType>();
        }

        /// <summary>
        /// load a configuration.
        /// </summary>
        /// <param name="config">The full path to a configuration file.</param>
        /// <returns></returns>
        private void readConfiguration(String config) 
        {
            String fileName = config;
            if (config == null) {
                // read from standard loc (current directory)
                fileName = System.Environment.CurrentDirectory + "\\config.xml";                
            }

            XmlTextReader reader = new XmlTextReader(fileName);
            Boolean inDependencies = false;
            AnnotationType currentAnnType = null;
            
            
            while(reader.Read())
            {
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("name")) 
                {
                    if (inDependencies)
                    {
                        reader.Read();
                        this.dependencyName = reader.Value;
                    }
                    else
                    {
                        reader.Read();
                        this.name = reader.Value;
                    }

                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("organization"))
                {
                    if (inDependencies)
                    {
                        reader.Read();
                        this.dependencyOrganization = reader.Value;
                    }
                    else
                    {
                        reader.Read();
                        this.organization = reader.Value;
                    }

                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("dependency"))
                {
                    inDependencies = true;
                }
                if (reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("dependency"))
                {
                    inDependencies = false;
                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("annotation"))
                {
                    currentAnnType = new AnnotationType();
                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("type"))
                {
                    reader.Read();
                    currentAnnType.setType(reader.Value);
                }

                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("field"))
                {
                    reader.Read();
                    currentAnnType.setField(reader.Value);
                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("keywordType"))
                {
                    reader.Read();
                    currentAnnType.setKeywordType(reader.Value);
                }
                if (reader.NodeType == XmlNodeType.EndElement && reader.Name.Equals("annotation"))
                {
                    annotations.Add(currentAnnType);
                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("descriptor"))
                {
                    reader.Read();
                    this.descriptor = reader.Value;
                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("all"))
                {
                    processAll = true;
                }
                if (reader.NodeType == XmlNodeType.Element && reader.Name.Equals("since"))
                {
                    reader.Read();
                    try
                    {
                        since = DateTime.ParseExact(reader.Value, "yyyy'-'MM'-'dd HH':'mm", null);
                    }
                    catch (ArgumentNullException e)
                    {
                        Console.WriteLine("DateTime argument in since tag is empty");
                        throw e;
                    }
                    catch (ArgumentException e)
                    {
                        Console.WriteLine("DateTime argument in since tag is invalid");
                        throw e;
                    }
                    catch (FormatException e)
                    {
                        Console.WriteLine("DateTime argument in since tag is invalid: use yyyy-MM-dd HH:mm");
                        throw e;
                    }
                }
            }

            wrapper = new UIMAWrapper(descriptor);
            foreach (AnnotationType annotation in annotations)
            {
                wrapper.addAnnotationType(annotation);
            }
            
        }

        /// <summary>
        /// Run the keywordExtraction process. This function retrieves descriptions from the database, passes them
        /// to UIMA and stores the resulting keywords in the database.
        /// </summary>
        public void run()
        {
            // do we have a dependency, get all items processed by that module and not yet by this one
            Module module = mminterface.CreateModule(name, organization);
            Module dependencyModule = null;
            if (dependencyName != null) 
            {
                dependencyModule = mminterface.CreateModule(dependencyName, dependencyOrganization);                
            }

            List<int> items = null;
            if (processAll)
            {
                items = mminterface.getItems();
            } 
            else if (dependencyModule != null)
            {
                if (since == DateTime.MinValue)
                {
                    items = mminterface.getProcessedBy(dependencyModule.ModuleId, module.ModuleId);
                }
                else
                {
                    items = mminterface.getProcessedBy(dependencyModule.ModuleId, module.ModuleId, since);
                }
            }
            else
            {
                if (since == DateTime.MinValue)
                {
                    items = mminterface.getProcessedBy(-1, module.ModuleId);
                }
                else
                {
                    items = mminterface.getProcessedBy(-1, module.ModuleId, since);
                }                
            }

            //System.Console.WriteLine("found " + items.Count + " items to process");

            foreach (int index in items)
            {
                CatalogItem i = mminterface.getItem(index);
                EntitySet<ItemDescription> descriptions = i.ItemDescriptions;
                //System.Console.WriteLine("Descriptions " + descriptions.Count);
                foreach (ItemDescription description in descriptions)
                {
                    String desc = description.Description;
                    // desc holds the text of the description
                    //System.Console.WriteLine("Description " + desc);
                    LinkedList<AnnotationWrapper> result = wrapper.tag(desc);
                    foreach (AnnotationWrapper a in result)
                    {
                        String keyword = a.getText();
                        mminterface.addKeyword(i.ItemId, keyword, module);
                        System.Console.WriteLine("Keyword " + keyword);
                    }


                    // Call mminterface.addKeyword(i.itemId, NEW_KEYWORD, module_id);
                }
                mminterface.markEnriched(i.ItemId, module);
                System.Console.WriteLine("Done processing: " + i.ItemTitles[0].Title);
            }

            // End code for Tag Extraction
            wrapper.finish();

        }

        /// <summary>
        /// Start the program with the following arguments UIMALauncher.exe CONFIG
        /// </summary>
        /// <param name="args"></param>
        public static void Main(string[] args)
        {
            KeywordExtractor p = new KeywordExtractor();

            try
            {
                if (args.Length > 0)
                {
                    p.readConfiguration(args[0]);
                }
                else
                {
                    p.readConfiguration(null);
                }
            }
            catch (XmlException e)
            {
                System.Console.WriteLine("Can't find configuration file, you can use the standard config.xml in the base directory: " + e.Message);
                return;
            }
            catch (Exception e)
            {
		System.Console.WriteLine(e.Message);
                return;
            }

            try
            {
                p.run();
            }
            catch (Exception e)
            {
                System.Console.WriteLine("The following error occured: " + e.Message);
                System.Console.WriteLine("The following error occured: " + e.StackTrace);
            }
        }
    }
}
