package nl.telin.mymedia.db.util;

import java.util.Iterator;
import java.util.List;

import nl.telin.mymedia.db.InfGainVector;
import nl.telin.mymedia.db.Lemma;
import nl.telin.mymedia.db.Stats;

import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.cfg.AnnotationConfiguration;

/**
 * @author Wout.Slakhorst
 * Utility class for accessing the database though Hibernate
 */
public class HibernateUtil {
	
	private static long timeInFind = 0;
	private static long timeInSave = 0;
	
	private static final SessionFactory sessionFactory;
	static {
		try {
			sessionFactory = new AnnotationConfiguration()
					.addAnnotatedClass(Lemma.class)
					.addAnnotatedClass(Stats.class)
					.addAnnotatedClass(InfGainVector.class)
					.configure().buildSessionFactory();
		} catch (Throwable e) {
			throw new ExceptionInInitializerError(e);
		}
	}

	/**
	 * open a hibernate session
	 * @return a new session
	 * @throws HibernateException
	 */
	public static Session getSession() throws HibernateException {
		return sessionFactory.openSession();
	}

	/**
	 * close database
	 * @throws HibernateException
	 */
	public static void shutdown() throws HibernateException {
		sessionFactory.close();
	}
	
	/**
	 * return all lemma's from the database
	 * @return a list of Lemma
	 */
	public static List<Lemma> getAllLemma() {
		Session s = getSession();
		//Transaction newTransaction = s.beginTransaction();
		List lemmas = s.createQuery("from Lemma p").list();

		//newTransaction.commit();
		s.close();
		return lemmas;
	}

	/**
	 * return lemma from the database with given string
	 * @param lemma the lemma string
	 * @return Found lemma
	 */
	public static Lemma findLemma(String lemma) {
		Session s = getSession();
		//Transaction newTransaction = s.beginTransaction();
		Lemma lq = (Lemma) s.createQuery("from Lemma lq where lemma = :lemma")
				.setString("lemma", lemma).uniqueResult();

		//newTransaction.commit();
		s.close();
		return lq;
	}
	
	/**
	 * save a lemma to the database
	 * @param lf The Lemma
	 * @return The saved Lemma with id
	 */
	public static Lemma saveLemma(Lemma lf) {
		Session s = getSession();
		Transaction newTransaction = s.beginTransaction();
		s.saveOrUpdate(lf);
		newTransaction.commit();
		s.close();
		return lf;
	}
	
	/**
	 * @return the number of documents used to create the database. (Training size)
	 */
	public static Stats findDocCount() {
		Session s = getSession();
		//Transaction newTransaction = s.beginTransaction();
		Stats stats = (Stats) s.createQuery("from Stats stats where name = 'documents'").uniqueResult();

		//newTransaction.commit();
		s.close();
		return stats;
	}
	
	/**
	 * Updste the document count.
	 * @param stats Stats object with doccount, stats.name = documents
	 */
	public static void updateDocCount(Stats stats) {
		Session s = getSession();
		Transaction newTransaction = s.beginTransaction();
		s.saveOrUpdate(stats);

		newTransaction.commit();
		s.close();		
	}
	
	/**
	 * Get all docFreq objects
	 * @param an open hibernate session
	 * @return an iterator
	 */
	public static Iterator getAllDocFreq(Session s) {
		Iterator iter = s.createQuery("from DocFreq").iterate();
		return iter;
	}
	
	/**
	 * remove all Lemma
	 */
	public static void removeAll() {
		Session s = getSession();
		Transaction newTransaction = s.beginTransaction();
		s.createSQLQuery("delete from lemma").executeUpdate();
		newTransaction.commit();
		s.close();
	}
	
	/**
	 * remove all InfGain
	 */
	public static void removeAllInfGain() {
		Session s = getSession();
		Transaction newTransaction = s.beginTransaction();
		s.createSQLQuery("delete from InfGainVector").executeUpdate();
		newTransaction.commit();
		s.close();
	}
	
	/**
	 * save InfGainVector to database
	 * @param vector the vector
	 * @return the vector with id. 
	 */
	public static InfGainVector saveCODistr(InfGainVector vector) {
		Session s = getSession();
		Transaction newTransaction = s.beginTransaction();
		s.save(vector);
		newTransaction.commit();
		s.close();	
		return vector;
	}

	/**
	 * get the number of lemma's in the database
	 * @return the count
	 */
	public static int getLemmaCount() {
		Session s = getSession();
		Transaction newTransaction = s.beginTransaction();
		int count = (Integer) s.createSQLQuery("select count(*) from lemma").uniqueResult();
		newTransaction.commit();
		s.close();
		return count;
	}

	/**
	 * remove common lemma's from the database
	 * They have a count > doccount/2
	 */
	public static void filterCommonLemma() {
		// get doc count
		Integer docs = HibernateUtil.findDocCount().getValue();
		try {
			Session s = getSession();
			Transaction newTransaction = s.beginTransaction();
			s.createSQLQuery("delete from lemma where docs > ?")
					.setInteger(0, docs/2)
					.executeUpdate();
			newTransaction.commit();
			s.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * remove rare lemma's from the database
	 * They have a count < doccount/1000
	 */
	public static void filterSpecificLemma() {
		// get doc count
		Integer docs = HibernateUtil.findDocCount().getValue();
		try {
			Session s = getSession();
			Transaction newTransaction = s.beginTransaction();
			s.createSQLQuery("delete from lemma where docs < ?")
					.setInteger(0, docs/1000)
					.executeUpdate();
			newTransaction.commit();
			s.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * get infgainvector from a lemma
	 * @param Lemma
	 * @return the vector
	 */
	public static InfGainVector findInfGainVector(Lemma l) {
		Session s = getSession();
		//Transaction newTransaction = s.beginTransaction();
		InfGainVector lq = (InfGainVector) s.createQuery("from InfGainVector v where lemma_id = :lemma")
				.setInteger("lemma", l.getId()).uniqueResult();

		//newTransaction.commit();
		s.close();
		return lq;
	}
	
	/**
	 * get infgainvector from an id
	 * @param id
	 * @return the vector
	 */
	public static InfGainVector findInfGainVector(Integer id) {
		Session s = getSession();
		//Transaction newTransaction = s.beginTransaction();
		InfGainVector lq = (InfGainVector) s.createQuery("from InfGainVector v where id = :id")
				.setInteger("id", id).uniqueResult();

		//newTransaction.commit();
		s.close();
		return lq;
	}
}
