package nl.telin.mymedia.processor;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import com.hp.hpl.jena.rdf.model.InfModel;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Selector;
import com.hp.hpl.jena.rdf.model.SimpleSelector;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;

/**
 * @author Wout.Slakhorst
 * @author Christian.Wartena
 * The CarrotProcessor computes the top n interesting ontology concepts found in a 
 * text based on the Carrot algorithm
 */
public class CarrotProcessor {

	private String defaultNS;
	private InfModel ontologyInf;
	private Set<String> terms;
	Map<String, Double> weights;
	private Map<String, Set<String>> distance1;
	private Map<String, Set<String>> distance2;

	public CarrotProcessor() {

	}

	/**
	 * load the ontology model
	 * @param ontology path to ontology file
	 * @param defaultNS default namespace
	 * @throws FileNotFoundException
	 */
	public void loadModel(String ontology, String defaultNS)
			throws FileNotFoundException {
		this.defaultNS = defaultNS;
		try {
			InputStream stream = new FileInputStream(ontology);

			// load jena
			Model model = ModelFactory.createDefaultModel();
			model.read(stream, "");
			ontologyInf = ModelFactory.createRDFSModel(model);
		} catch (FileNotFoundException e) {
			throw e;
		}
	}
	
	/**
	 * Load the subjects that are to be ranked.
	 * @param terms subjects
	 */
	public void setTerms(Set<String> terms, Map<String, Double> weights) {
		this.terms = terms;
		this.weights = weights;
	}
	
	/**
	 * @return a map of concept to scores. The higher the score the better
	 */
	public Map<String, Double> recomputeWeights() {
		final double avlinks = 3.0;
		final double sd = 8.0;
		findAllDistance1Subjects();
		findAllDistance2Subjects();
		//int nrOfRelations = distance1.size();
		//int nrOfWords = terms.size();
		
		for(String subject:terms) {
			Distance l1r = distance1Relations(subject);
			Distance l2r = distance2Relations(subject);
			
			double tf = 1.0 + Math.log(weights.get(subject));
			double q = 1.0/(2 * avlinks); // counted from ontology
			
			/*
			double pr = 1 + l1r.getOntRel() + q * l2r.getOntRel();
			double rr = 1 + l1r.getTextRel() + q * l2r.getTextRel();
			double rrf = rr / ( 1 + Math.log(1 + pr));
			*/
			double pr1 = l1r.getOntRel();
			double rr1 = l1r.getTextRel();
			//double rrf1 = rr1 / Math.log(pr1);
			// double pr2 = 1 + l2r.getOntRel();
			double rr2 = l2r.getTextRel();
			//double rrf2 = rr2 / Math.log(pr2);
			//double rrf = rrf1 + q * rrf2;
			double rr = 1 + q * (rr1 + q  * rr2);
			double alpha = avlinks/sd;
			double cw = 1 + pr1 * Math.exp(-1 * alpha * pr1/avlinks);
			
			double weight = tf * rr * cw;
			weights.put(subject, weight);
		}
		
		return weights;
	}

	private Distance distance1Relations(String subject) {
		int textualRel = 0;
		int ontologyRel = 0;
		Set<String> relatedTerms = distance1.get(subject);
		
		for(String term:relatedTerms) {
			if(terms.contains(term)) {
				textualRel++;
			}
			ontologyRel++;
		}
		
		return new Distance(textualRel, ontologyRel);
	}

	private Distance distance2Relations(String subject) {
		int textualRel = 0;
		int ontologyRel = 0;
		Set<String> relatedTerms = distance2.get(subject);
		
		for(String term:relatedTerms) {
			if(terms.contains(term)) {
				textualRel++;
			}
			ontologyRel++;
		}
		
		return new Distance(textualRel, ontologyRel);
	}
	
	
	
	private void findAllDistance1Subjects() {
		distance1 = new HashMap<String, Set<String>>();
		for(String subject:terms) {
			distance1.put(subject, getDistance1Subjects(subject));
		}
	}
	
	private void findAllDistance2Subjects() {
		distance2 = new HashMap<String, Set<String>>();
		for(String subject:terms) {
			distance2.put(subject, getDistance2Subjects(subject));
		}
	}
	
	/**
	 * returns a set of subjects, each with the namespace
	 * @param subject with namespace
	 * @return
	 */
	private  Set<String> getDistance1Subjects(String subject) {

		Set<String> result = new HashSet<String>();
		
		// related
		Selector query = new SimpleSelector(ontologyInf.createResource(subject), ontologyInf.createProperty(
				"http://www.w3.org/2004/02/skos/core#", "related"),
				(RDFNode) null);
		StmtIterator iter = ontologyInf.listStatements(query);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			result.add(stmt.getObject().toString());
		}
		
		// narrower
		query = new SimpleSelector(ontologyInf.createResource(subject), ontologyInf.createProperty(
				"http://www.w3.org/2004/02/skos/core#", "narrower"),
				(RDFNode) null);
		iter = ontologyInf.listStatements(query);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			result.add(stmt.getObject().toString());
		}
		
		// broader
		query = new SimpleSelector(ontologyInf.createResource(subject), ontologyInf.createProperty(
				"http://www.w3.org/2004/02/skos/core#", "broader"),
				(RDFNode) null);
		iter = ontologyInf.listStatements(query);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			result.add(stmt.getObject().toString());
		}
		
		result.remove(subject);
		
		return result;
	}
	
	/**
	 * get all level 2 subjects. For each subject the level 1 subjects are
	 * retrieved after which for each of those the level 1 subjects are retrieved.
	 * @param subject
	 * @return
	 */
	private Set<String> getDistance2Subjects(String subject) {

		Set<String> result = new HashSet<String>();
		
		Set<String> l1 = distance1.get(subject);
		for (String s:l1) {
			result.addAll(getDistance1Subjects(s));
		}
		result.remove(subject);
		result.removeAll(l1); //CW 2008/12/23
		return result;
	}
	
	public String getPrefLabel(String subject) {
		// bla
		Selector query = new SimpleSelector(ontologyInf.createResource(subject), ontologyInf.createProperty(
				"http://www.w3.org/2004/02/skos/core#", "prefLabel"),
				(RDFNode) null);
		StmtIterator iter = ontologyInf.listStatements(query);
		if (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			return stmt.getObject().toString();
		}
		return null;
	}
}
