package nl.telin.mymedia.processor.token;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import nl.telin.mymedia.processor.AbstractProcessor;
import nl.telin.mymedia.processor.ProcessorException;
import nl.telin.mymedia.processor.ProcessorInitializeException;

/**
 * @author Wout.Slakhorst
 * The RegExpTokenizer uses regular expressions to create word and sentence tokens.
 * This UIMA processor gets its word expressions from a file and its sentence expression
 * from a parameter
 * 
 * parameters:
 * PARAM_WORDEXPRFILE - path to wordExprFile
 * PARAM_SENTENCEEXPR - the sentence expression
 */
public class RegExpTokenizer extends AbstractProcessor<String, List<List<TextToken>>> {

	public static final String PARAM_WORDEXPRFILE = "PARAM_WORDEXPRFILE";
	public static final String PARAM_SENTENCEEXPR = "PARAM_SENTENCEEXPR";

	private Map<String, String> expressions = null;
	private String sentenceExpr = null;

	@Override
	/**
	 * @inheritDoc
	 * initialize expressions from file
	 */
	public void initialize(Map<String, String> params)
			throws ProcessorInitializeException {
		
		expressions = new HashMap<String, String>();

		String path = params.get(PARAM_WORDEXPRFILE);
		sentenceExpr = params.get(PARAM_SENTENCEEXPR);

		if (path == null || sentenceExpr == null) {
			throw new ProcessorInitializeException(
					"Missing initialize parameters: " + PARAM_WORDEXPRFILE
							+ ", " + PARAM_SENTENCEEXPR);
		}

		parseExpressions(path);		

	}

	/**
	 * @inheritDoc
	 * add TokenAnnotations and SentenceAnnotations to CAS
	 */
	public List<List<TextToken>> process(String text) throws ProcessorException {
		
		List<TextToken> sentences = new ArrayList<TextToken>();
		List<List<TextToken>> subj = new ArrayList<List<TextToken>>();
		//List<TextToken> words = new ArrayList<TextToken>();
		//List<TextToken> puncs = new ArrayList<TextToken>();
		
		// sentence annotator
		Pattern pattern = Pattern.compile(sentenceExpr);
		Matcher matcher = pattern.matcher(text);
		try {
			while (matcher.find()) {
				int begin = matcher.start();
				int end = matcher.end();
				sentences.add(new TextToken(text.substring(begin, end), begin, end, "SENT", text));
			}
		} catch (Error e) {
			// Stackoverflow occurs occasionally
			//System.out.println(text);
			throw new ProcessorException(e);
		}
		
		LinkedList<TextToken> queue = new LinkedList<TextToken>();
		
		// place all words on stack
		for(String s : expressions.keySet()) {
			String type = expressions.get(s);
			pattern = Pattern.compile(s);
		
			// place all words on stack
			matcher = pattern.matcher(text);
			while (matcher.find()) {
				int begin = matcher.start();
				int end = matcher.end();
				String t = text.substring(begin, end);
				queue.add(new TextToken(t, begin, end, type, text));							
			}
		}
		
		// sort words
		Collections.sort(queue, new TextTokenComparator());
		
		// divide over sentences
		for (TextToken tt : sentences ) {
			if (queue.isEmpty()) {
				break;
			}
			List<TextToken> words = new ArrayList<TextToken>();
			
			int end = tt.getEnd();
			TextToken token = queue.peek();
			
			while (token != null && token.getEnd() <= end) {
				words.add(queue.poll());
				token = queue.peek();
			}			
			subj.add(words);			
		}		
		
		return subj;
	}

	private void parseExpressions(String path) throws ProcessorInitializeException {
		try {
			BufferedReader reader = new BufferedReader(new FileReader(path));
			String line = null;
			while ((line = reader.readLine()) != null) {
				String[] splitted = line.split(" ");
				if (splitted.length != 2) {
					throw new ProcessorInitializeException("Error parsing regular expression rule: " + line);
				} else {
					expressions.put(splitted[0], splitted[1]);
				}
			}
		} catch (IOException e) {
			throw new ProcessorInitializeException(e);
		}

	}

}
