package nl.telin.mymedia.uima.annotator;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.io.IOException;
import java.net.URL;

import org.apache.uima.TokenAnnotation;
import org.apache.uima.SentenceAnnotation;
import org.apache.uima.analysis_engine.ResultSpecification;
import org.apache.uima.analysis_engine.annotator.AnnotatorConfigurationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorContext;
import org.apache.uima.analysis_engine.annotator.AnnotatorInitializationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorProcessException;
import org.apache.uima.analysis_engine.annotator.JTextAnnotator_ImplBase;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.jcas.JCas;
import org.apache.uima.resource.ResourceProcessException;
import org.apache.uima.util.ProcessTrace;

/**
 * This class is a wrapper for HepTag, Mark Hepple's POS tagger.
 * @author Wout.Slakhorst
 * @author Christian.Wartena
 */
public class HeppleTaggerAnnotator extends JTextAnnotator_ImplBase {

	protected hepple.postag.POSTagger tagger;
	 
	private static final String TOKEN_ANNOTATION_NAME = "org.apache.uima.TokenAnnotation";
	private static final String SENTENCE_ANNOTATION_NAME = "org.apache.uima.SentenceAnnotation";

	private Type tokenAnnotation;
	private Type sentenceAnnotation;

	
	// arguments
	private String sEncoding;
	private URL urlLexicon;
	private URL urlRules;
	

	@Override
	/**
	 * @inheritDoc
	 * config parameters:
	 * encoding - ?
	 * lexiconURL - path to lexicon file
	 * rulesURL - path to rules file
	 */
	public void initialize(AnnotatorContext context)
	throws AnnotatorInitializationException,
	AnnotatorConfigurationException {
		// TODO Auto-generated method stub
		super.initialize(context);
		try{
			sEncoding = (String) context.getConfigParameterValue("encoding");
			String sLexicon = (String) context.getConfigParameterValue("lexiconURL");
			String sRules = (String) context.getConfigParameterValue("rulesURL");
			
			urlLexicon = new URL(sLexicon);
			urlRules = new URL(sRules);
			if(sEncoding == null || sEncoding.isEmpty()) {
				tagger = new hepple.postag.POSTagger(urlLexicon,urlRules);
			} else {
				tagger = new hepple.postag.POSTagger(urlLexicon,urlRules,sEncoding);
			}
		}catch(Exception e){
			throw new AnnotatorInitializationException(e);
		}
	}
	
	

	/* (non-Javadoc)
	 * @see org.apache.uima.analysis_engine.annotator.Annotator_ImplBase#typeSystemInit(org.apache.uima.cas.TypeSystem)
	 */
	@Override
	/**
	 * @inheritDoc
	 */
	public void typeSystemInit(TypeSystem typeSystem)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {
		// TODO Auto-generated method stub
		super.typeSystemInit(typeSystem);
		// initialize cas token type
		this.tokenAnnotation = typeSystem.getType(TOKEN_ANNOTATION_NAME);
		this.sentenceAnnotation = typeSystem.getType(SENTENCE_ANNOTATION_NAME);
	}



	/**
	 * @inheritDoc
	 */
	public void process(JCas jCas, ResultSpecification arg1)
	throws AnnotatorProcessException {

		//prepare the input for HepTag
		List sentenceForTagger = new ArrayList();
		List sentencesForTagger = new ArrayList(1);
		sentencesForTagger.add(sentenceForTagger);
		List tokensInCurrentSentence = new ArrayList();

		FSIterator sentenceIterator = jCas.getCas().getAnnotationIndex(this.sentenceAnnotation).iterator();
		FSIterator tokenIterator = jCas.getCas().getAnnotationIndex(this.tokenAnnotation).iterator();
		
		if (tokenIterator.hasNext()) {
			TokenAnnotation currentToken = (TokenAnnotation)tokenIterator.next();
			
			while (sentenceIterator.hasNext()) {
				// get token content
				SentenceAnnotation sentence = (SentenceAnnotation) sentenceIterator.next();
				tokensInCurrentSentence.clear();
				sentenceForTagger.clear();
				
				while(currentToken != null &&
						currentToken.getEnd() < sentence.getEnd()){
					tokensInCurrentSentence.add(currentToken);
					sentenceForTagger.add(currentToken.getCoveredText()); //TODO of staat die tekst al in een feature?
					currentToken = (TokenAnnotation)(tokenIterator.hasNext() ?
							tokenIterator.next() : null);
				}
				//run the POS tagger
				List taggerList = tagger.runTagger(sentencesForTagger);
				if(taggerList != null && taggerList.size() > 0){
					List taggerResults = (List) taggerList.get(0);
					//add the results
					//make sure no malfunction occurred
					if(taggerResults.size() != tokensInCurrentSentence.size()) {
						continue; //eigenlijk is nu alles mis!
					}
						
					Iterator resIter = taggerResults.iterator();
					Iterator tokIter = tokensInCurrentSentence.iterator();
					while(resIter.hasNext()){
						TokenAnnotation annot = (TokenAnnotation) tokIter.next();
						String[] htResult = ((String[])resIter.next());
						annot.setPos(htResult[1]);
						annot.setLemma(htResult[2]);
					}
				}
			}
		}
	}
	
	//@Override
	/**
	 * @inheritDoc
	 */
	public void destroy() {
		System.out.println("Finalizing HeppleTaggerAnnotator");
	}
}
