package nl.telin.mymedia.uima.annotator;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import nl.telin.mymedia.analysis.infgain.SparseVector;
import nl.telin.mymedia.db.InfGainVector;
import nl.telin.mymedia.db.Lemma;
import nl.telin.mymedia.db.util.HibernateUtil;
import nl.telin.mymedia.uima.annotation.InfGainAnnotation;

import org.apache.uima.TokenAnnotation;
import org.apache.uima.analysis_engine.ResultSpecification;
import org.apache.uima.analysis_engine.annotator.AnnotatorConfigurationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorContext;
import org.apache.uima.analysis_engine.annotator.AnnotatorContextException;
import org.apache.uima.analysis_engine.annotator.AnnotatorInitializationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorProcessException;
import org.apache.uima.analysis_engine.annotator.JTextAnnotator_ImplBase;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.cas.text.AnnotationIndex;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.util.Logger;

/**
 * @author Wout.Slakhorst
 * @author Christian.Wartena
 * Creates annotations based on the Information gain algorithm
 * 
 */
public class InfGainAnnotator extends JTextAnnotator_ImplBase {

	private Logger logger;

	private static final String TT_ANNOTATION_NAME = "org.apache.uima.TokenAnnotation";
	private static final String PT_ANNOTATION_NAME = "nl.telin.mymedia.uima.annotation.PersonAnnotation";
	private static final String LT_ANNOTATION_NAME = "nl.telin.mymedia.uima.annotation.LocationAnnotation";

	private Type ttAnnotation;
	private Type ptAnnotation;
	private Type ltAnnotation;

	private static final int MAX_RESULTS = 5;
	
	private static Set<String> semPos;
	
	static {
		semPos = new HashSet<String>();
		semPos.add("NN");
		semPos.add("NNP");
		semPos.add("VBG");
		semPos.add("VBD");
		semPos.add("VBN");
		semPos.add("VBP");
		semPos.add("VB");
		semPos.add("VBZ");
		semPos.add("FW");
		semPos.add("NNS");
		semPos.add("NNPS");
		semPos.add("NP");
		semPos.add("NPS");		
	}
	
	private static Set<String> keywordPos;
	
	static {
		keywordPos = new HashSet<String>();
		keywordPos.add("NN");
		keywordPos.add("NNP");
		keywordPos.add("FW");
		keywordPos.add("NNS");
		keywordPos.add("NNPS");
		keywordPos.add("NP");
		keywordPos.add("NPS");		
	}

	private static Set<String> auxList;
	
	static {
		auxList = new HashSet<String>();
		auxList.add("do");
		auxList.add("be");
		auxList.add("have");
		auxList.add("will");
		auxList.add("shall");
		auxList.add("can");
		auxList.add("may");
		auxList.add("go");
		auxList.add("make");
		auxList.add("let");
		auxList.add("try");
		auxList.add("get");
	}
	
	// @Override
	/**
	 * @inheritDoc
	 */
	public void initialize(AnnotatorContext context)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {
		super.initialize(context);

		try {
			logger = getContext().getLogger();
		} catch (AnnotatorContextException e) {
			throw new AnnotatorConfigurationException(e);
		}
	}

	// @Override
	/**
	 * @inheritDoc
	 */
	public void typeSystemInit(TypeSystem aTypeSystem)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {

		// initialize cas token type
		this.ttAnnotation = aTypeSystem.getType(TT_ANNOTATION_NAME);
		this.ptAnnotation = aTypeSystem.getType(PT_ANNOTATION_NAME);
		this.ltAnnotation = aTypeSystem.getType(LT_ANNOTATION_NAME);

	}

	// @Override
	/**
	 * @inheritDoc
	 */
	public void process(JCas jCas, ResultSpecification arg1)
			throws AnnotatorProcessException {

		String input = jCas.getDocumentText();
		int start = 0;
		int end = input.length() - 1;
		
		int corpusSize = HibernateUtil.findDocCount().getValue();
		//int nrOfOccurrences =  HibernateUtil.getLemmaCount();
		
		
		SparseVector<Float> generalDistr = (SparseVector<Float>) HibernateUtil.findInfGainVector(-1).getVector();
		
		FSIterator ttIterator = jCas.getCas().getAnnotationIndex(
				this.ttAnnotation).iterator();
		

		SparseVector<Float> docDistr = new SparseVector<Float>();
		int nrOfWords = 0;
		HashSet<Lemma> lemset = new HashSet<Lemma>();
		
		while (ttIterator.hasNext()) {
			TokenAnnotation annot = (TokenAnnotation) ttIterator.next();
			if ("WORD".equals(annot.getTokentype())
					&& semanticLemma(annot)) {
				String lemStr = annot.getLemma().toLowerCase();
				if(lemStr == null || lemStr.isEmpty()) {
					lemStr = annot.getCoveredText().toLowerCase();
				}
				Lemma l = HibernateUtil.findLemma(lemStr);
				if (l != null) {
					int lvIndex = l.getVectorId();
					if(kwLemma(annot,jCas)) {
						lemset.add(l);
					}
					docDistr.addTo(lvIndex, 1);
					nrOfWords++;
				}
			}
		}
		docDistr.divideElementsBy(nrOfWords);
	
		

		SortedSet<Keyword> keywords = new TreeSet<Keyword>();
		Iterator<Lemma> lemIt = lemset.iterator();
		while(lemIt.hasNext()) {
			Lemma l = lemIt.next();
			if (l.getDocs() > corpusSize/1000) {
				int lvIndex = l.getVectorId();
				InfGainVector lv = HibernateUtil.findInfGainVector(lvIndex);
				if (lv != null) {
					SparseVector<Float> lvCoocDistr = (SparseVector<Float>) lv.getVector();
					float dist = Statistics.JSDivergence(docDistr, lvCoocDistr) - 0.5f * Statistics.JSDivergence(lvCoocDistr,generalDistr);
					if(keywords.size() < MAX_RESULTS) {
						keywords.add(new Keyword(l, dist));
					} else if (dist < keywords.last().fit) {
						keywords.add(new Keyword(l, dist));
						keywords.remove(keywords.last());
					}
				}
			} else {
				int t = 1;
				t++;
			}
		}

		//System.out.println("---------------------");
		//System.out.println(jCas.getDocumentText());
		//float dqqd = distanceJS(docDistr,generalDistr,0);
		//System.out.println(dqqd);
		//for (Keyword k : keywords) {
//			System.out.println(k.getLem().getLemma() + "\t" + k.getFit());
		//}
		
		// add to uima
		LinkedList<Keyword> list = new LinkedList<Keyword>();
		list.addAll(keywords);
		
		for (int i = 0; i < MAX_RESULTS && i < list.size(); i++) {
			InfGainAnnotation annot = new InfGainAnnotation(jCas, start, end);
			annot.setLemma(list.get(i).getLem().getLemma());
			annot.setScore(list.get(i).getFit());
			annot.setRank(i + 1);
			annot.addToIndexes();
		}
	}

	private boolean coveredByOneOf(TokenAnnotation annot, AnnotationIndex annotIndex) {
		FSIterator it = annotIndex.iterator();
		Annotation cov;
		while(it.hasNext()) {
			cov = (Annotation) it.next();
			if(cov.getBegin() <= annot.getBegin() && cov.getEnd() >= annot.getEnd()) {
				return true;
			}
		}
		return false;
	}
	
	private boolean semanticLemma(TokenAnnotation currentToken) {
		String POS = currentToken.getPos();
		return (semPos.contains(POS) && !auxVerb(currentToken));
	}

	private boolean kwLemma(TokenAnnotation annot, JCas jCas) {
		String POS = annot.getPos();
		AnnotationIndex persIndex = jCas.getCas().getAnnotationIndex(this.ptAnnotation);
		AnnotationIndex locIndex = jCas.getCas().getAnnotationIndex(this.ltAnnotation);
		
		return (keywordPos.contains(POS) 
				&& !auxVerb(annot)
				&& !coveredByOneOf(annot,persIndex) 
				&& !coveredByOneOf(annot,locIndex));
	}
	
	private boolean auxVerb(TokenAnnotation token) {
		boolean rc = false;
		String lemma = token.getLemma();
		if(auxList.contains(lemma)) {
			rc = true;
		}
		return rc;
	}
	
	

	private class Keyword implements Comparable<Keyword> {

		private Lemma lem;
		private Float fit;

		public Keyword(Lemma lem, Float fit) {
			super();
			this.lem = lem;
			this.fit = fit;
		}

		public int compareTo(Keyword o) {

			return fit.compareTo(o.getFit());
		}

		/**
		 * @return the lem
		 */
		public Lemma getLem() {
			return lem;
		}

		/**
		 * @return the fit
		 */
		public Float getFit() {
			return fit;
		}

	}
}
