package nl.telin.mymedia.uima.annotator;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import nl.telin.mymedia.db.Lemma;
import nl.telin.mymedia.db.Stats;
import nl.telin.mymedia.db.util.HibernateUtil;
import nl.telin.mymedia.uima.annotation.TfidfAnnotation;

import org.apache.uima.TokenAnnotation;
import org.apache.uima.analysis_engine.ResultSpecification;
import org.apache.uima.analysis_engine.annotator.AnnotatorConfigurationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorContext;
import org.apache.uima.analysis_engine.annotator.AnnotatorContextException;
import org.apache.uima.analysis_engine.annotator.AnnotatorInitializationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorProcessException;
import org.apache.uima.analysis_engine.annotator.JTextAnnotator_ImplBase;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.jcas.JCas;
import org.apache.uima.util.Logger;

/**
 * @author Wout.Slakhorst
 * Annotator that uses tfidf to get the top n words from a text.
 * This annotator requires the TfIdfTrainer to train first. They can' be used in 
 * the same pipeline.
 */
public class TfidfAnnotator extends JTextAnnotator_ImplBase {

	private Logger logger;
		
	private static final String TT_ANNOTATION_NAME = "org.apache.uima.TokenAnnotation";
	
	private Type ttAnnotation;
	
	@Override
	/**
	 * @inheritDoc
	 */
	public void initialize(AnnotatorContext context)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {
		super.initialize(context);

		try {
			logger = getContext().getLogger();			
		} catch (AnnotatorContextException e) {
			throw new AnnotatorConfigurationException(e);
		}
	}
	
	@Override
	/**
	 * @inheritDoc
	 */
	public void typeSystemInit(TypeSystem aTypeSystem)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {

		// initialize cas token type
		this.ttAnnotation = aTypeSystem.getType(TT_ANNOTATION_NAME);		

	}
	
	//@Override
	/**
	 * @inheritDoc
	 * create TfidfAnnotations for this cas
	 */
	public void process(JCas jCas, ResultSpecification arg1)
			throws AnnotatorProcessException {
		
		String input = jCas.getDocumentText();
		int start = 0;
		int end = input.length() - 1;
		int totalCount = 0; 
		
		Map<String, Integer> count = new HashMap<String, Integer>();
		Map<String, Double> score = new HashMap<String, Double>();
		
		FSIterator ttIterator = jCas.getCas().getAnnotationIndex(
				this.ttAnnotation).iterator();

		while (ttIterator.hasNext()) {
			TokenAnnotation annot = (TokenAnnotation) ttIterator.next();
			if ("WORD".equals(annot.getTokentype())) { 
				String lemma = annot.getLemma();
				totalCount++;
				
				// if no lemma exists use original word
				if (lemma == null || "".equals(lemma)) {
					lemma = annot.getCoveredText();
				}
				
				lemma = lemma.toLowerCase();
				
				// add to map
				Integer c = count.get(lemma);
				if (c == null) {
					c = new Integer(0);
				}
				c++;
				count.put(lemma, c);
			}
			
		}
		
		// Iterate over lemma's and score them
		for (Map.Entry<String, Integer> entry : count.entrySet()) {
			Integer c = entry.getValue();
			String lemma = entry.getKey();
			Double s = score(lemma, c, totalCount);
			score.put(lemma, s);
		}
		
		// sort the scores in the map;
		List<Double> l = new ArrayList<Double>();
		l.addAll(score.values());
		Collections.sort(l);
		Collections.reverse(l);
		
		if (l.size() > 0) {
			int size = (int) Math.round(Math.log(l.size()));
			double minScore = l.get(size);
			
			// if score is in scoreSet, create annotation
			for (Map.Entry<String, Double> entry : score.entrySet()) {
				Double d = entry.getValue();
				String lemma = entry.getKey();
				if (d > minScore) {
					TfidfAnnotation newAnnot = new TfidfAnnotation(jCas, start, end);
					newAnnot.setLemma(lemma);
					newAnnot.setScore(d);
					newAnnot.addToIndexes();
				}
			}
		}
	}
	
	private double score(String lemma, int freq, int totalCount) {
		Lemma lf = HibernateUtil.findLemma(lemma);
		Stats docCount = HibernateUtil.findDocCount();
		
		double tf = ((double) freq) / totalCount;
		
		Integer lc = 1;
		lc = lf == null ? Integer.MAX_VALUE : lf.getFreq(); // non existing words are not important
		double idf = Math.log(docCount.getValue() / lc);
		return idf * tf;
	}
}

