package nl.telin.mymedia.uima.annotator;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.LinkedList;
import java.util.Queue;

import org.apache.uima.TokenAnnotation;
import org.apache.uima.analysis_engine.ResultSpecification;
import org.apache.uima.analysis_engine.annotator.AnnotatorConfigurationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorContext;
import org.apache.uima.analysis_engine.annotator.AnnotatorContextException;
import org.apache.uima.analysis_engine.annotator.AnnotatorInitializationException;
import org.apache.uima.analysis_engine.annotator.AnnotatorProcessException;
import org.apache.uima.analysis_engine.annotator.JTextAnnotator_ImplBase;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.jcas.JCas;
import org.apache.uima.util.Logger;

/**
 * TreeTagger annotator for UIMA
 * Tis class calls the external c treetagger.
 * @author Wout.Slakhorst
 *
 */
public class TreeTaggerAnnotator extends JTextAnnotator_ImplBase {

	private static final String TOKEN_ANNOTATION_NAME = "org.apache.uima.TokenAnnotation";

	private Type tokenAnnotation;
	private Queue<TokenAnnotation> tokenQueue;
	
	// arguments
	private String ttPath;
	private String ttLib;
	private String ttArg;
	
	private Logger logger;

	
	
	@Override
	/**
	 * @inheritDoc
	 * sets the correct paths from parameters:
	 * TreeTaggerPath
	 * TreeTaggerLib
	 * TreeTaggerOptions
	 */
	public void initialize(AnnotatorContext context)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {
		super.initialize(context);
		try {
			logger = getContext().getLogger();
			
			ttPath = (String) getContext().getConfigParameterValue("TreeTaggerPath");
			ttLib = (String) getContext().getConfigParameterValue("TreeTaggerLib");
			ttArg = (String) getContext().getConfigParameterValue("TreeTaggerOptions");
		} catch (AnnotatorContextException e) {
			throw new AnnotatorConfigurationException(e);
		}
	}

	@Override
	/**
	 * @inheritDoc
	 */
	public void typeSystemInit(TypeSystem aTypeSystem)
			throws AnnotatorInitializationException,
			AnnotatorConfigurationException {
	
		// initialize cas token type
		this.tokenAnnotation = aTypeSystem.getType(TOKEN_ANNOTATION_NAME);
		this.tokenQueue = new LinkedList<TokenAnnotation>();			
		
	}

	//@Override
	/**
	 * @inheritDoc
	 * process a single text, treetagger is started and stopped for each text
	 */
	public void process(JCas jCas, ResultSpecification arg1)
			throws AnnotatorProcessException {

		try {
			// iterate over all token annotations and add stem if available
			FSIterator tokenIterator = jCas.getCas().getAnnotationIndex(
					this.tokenAnnotation).iterator();
			
			Runtime runtime = Runtime.getRuntime();
			Process p = runtime.exec(ttPath + " " + ttLib + " " + ttArg);
	
			BufferedReader input = new BufferedReader(new InputStreamReader(p
					.getInputStream()));
			
			BufferedReader error = new BufferedReader(new InputStreamReader(p
					.getErrorStream()));
			
			PrintWriter output = new PrintWriter(new BufferedOutputStream(p.getOutputStream()));
						
			new Thread(new AnnotateTaggerOutput(input, jCas)).start();
			new Thread(new AnnotateDebugOutput(error)).start();
			
			while (tokenIterator.hasNext()) {
				// get token content
				TokenAnnotation annot = (TokenAnnotation) tokenIterator.next();
				String span = annot.getCoveredText();
	
				// add to Queue
				tokenQueue.add(annot);
				
				output.write(span);
				output.write("\n");
				output.flush();
			}
			output.close();
			p.waitFor();			
		} catch (IOException e) {
			throw new AnnotatorProcessException(e);
		} catch (InterruptedException e) {
			throw new AnnotatorProcessException(e);
		}
	}

	/**
	 * Class for reading TreeTagger output
	 * @author Wout.Slakhorst
	 *
	 */
	private class AnnotateTaggerOutput implements Runnable {

		BufferedReader reader;
		JCas jCas;

		public AnnotateTaggerOutput(BufferedReader reader, JCas jCas) {
			this.reader = reader;
			this.jCas = jCas;
		}

		//@Override
		public void run() {
			try {
				String line;
				while ((line = reader.readLine()) != null) {
					String[] info = line.split("\t");

					String token = null;
					TokenAnnotation queuedAnnot = null;
					while (!(info[0].equals(token))) {
						if (tokenQueue.peek() != null) {
							queuedAnnot = tokenQueue.poll();
							token = queuedAnnot.getCoveredText();
						} else {
							break;
						}
					}
					
					if (queuedAnnot != null) {
						queuedAnnot.setLemma(info[2]);
						//TODO setStem
						/*
						TreeTaggerAnnotation newAnnot = new TreeTaggerAnnotation(jCas,
								queuedAnnot.getBegin(), queuedAnnot.getEnd());
						newAnnot.setToken(info[0]);
						newAnnot.setPOS(info[1]);
						newAnnot.setStem(info[2]);
	
						newAnnot.addToIndexes();
						*/
					}
				}
			} catch (IOException e) {
				throw new RuntimeException(e);
				// exit prog
			}
		}
	}
	
	/**
	 * Class for reading TreeTagger errors
	 * @author Wout.Slakhorst
	 *
	 */
	private class AnnotateDebugOutput implements Runnable {

		BufferedReader reader;

		public AnnotateDebugOutput(BufferedReader reader) {
			this.reader = reader;
		}

		//@Override
		public void run() {
			try {
				String line;
				while ((line = reader.readLine()) != null) {
					System.err.println("[TREETAGGER] " + line);
				}
			} catch (IOException e) {
				throw new RuntimeException(e);
				// exit prog
			}
		}
	}

}
