package nl.telin.mymedia.wordnet;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.ontology.impl.OntModelImpl;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Selector;
import com.hp.hpl.jena.rdf.model.SimpleSelector;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;

/**
 * Enriches an ontology with adjective forms from WordNet
 * 
 * After generating the owl file add these lines to the owl:
 * 
 * <rdf:Description rdf:about="http://www.w3.org/2004/02/skos/core#altLabelAdj">
 *  <rdfs:subPropertyOf rdf:resource="http://www.w3.org/2004/02/skos/core#altLabel"/>
 *  <rdf:type rdf:resource="http://www.w3.org/2002/07/owl#DatatypeProperty"/>
 * </rdf:Description>
 * 
 * @author Wout.Slakhorst
 * 
 */
public class WordnetEnricher {

	private Model model, outputModel;
	private Map<Integer, Synset> nouns;
	private Map<Integer, Synset> adjs;
	private Map<String, Integer> lemmaMap;
	
	public WordnetEnricher() {
		
	}
	
	public void enrich(String args[]) {
		String nounFile = args[0];
		String adjFile = args[1];
		String ontFile = args[2];
		String outFile = args[3];
		
		try {
			// load wordnet
			WordnetReader nounReader = new WordnetReader(new File(nounFile));
			WordnetReader adjReader = new WordnetReader(new File(adjFile));
			nouns = nounReader.read();
			adjs = adjReader.read();
			lemmaMap = createLemmaMap(nouns);
			
			// load jena
			InputStream stream = new FileInputStream(ontFile);
			model = ModelFactory.createDefaultModel();
			model.read(stream, "");
			
			
			
			outputModel = new OntModelImpl(OntModelSpec.RDFS_MEM, model);
						
			enhanceAll();
			
			OutputStream out = new FileOutputStream(outFile);
			outputModel.write(out);
			outputModel.close();
			model.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void enhanceAll() {
		// find all prefLabel
		Selector query = new SimpleSelector(null, model.createProperty(
				"http://www.w3.org/2004/02/skos/core#", "prefLabel"),
				(RDFNode) null);
		enhanceOntology(query);

		// find all altLabel
		query = new SimpleSelector(null, model.createProperty(
				"http://www.w3.org/2004/02/skos/core#", "altLabel"),
				(RDFNode) null);
		enhanceOntology(query);
		
		// find all altLabelSg
		query = new SimpleSelector(null, model.createProperty(
				"http://www.w3.org/2004/02/skos/core#", "altLabelSg"),
				(RDFNode) null);
		enhanceOntology(query);
	}
	
	private void enhanceOntology(Selector query) {
		StmtIterator iter = model.listStatements(query);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			
			// split words
			String[] split = stmt.getObject().toString().split("\\(");
			String left = split[0];
			String word = left.toLowerCase();
			
			//search for word in lemmaMap
			Integer id = lemmaMap.get(word);
			if (id != null) {
				if (word.startsWith("elect")) {
					System.out.println(word);
				}
				Synset set = nouns.get(id);
				List<Synlink> links = set.getLinks();
				if (links.size() != 0) {
					for (Synlink link : links) {
						// if the synlink has a link from the current word to another word
						int fromWord = link.getFromWord();
						int toWord = link.getToWord();
						if (fromWord == 0) {
							// synset to synset
							// ignored
						} else {
							fromWord -= 1; // wordnet is not zero based
							toWord -= 1;
							if (word.equals(set.getLemmas().get(fromWord))) {
								String altLabel = adjs.get(link.getTargetByteOffset()).getLemmas().get(toWord);
								if (!(altLabel.equals(word))) {
									outputModel.add(outputModel.createStatement(stmt.getSubject(), outputModel
											.createProperty("http://www.w3.org/2004/02/skos/core#",
													"altLabelAdj"), altLabel));
									//System.out.println("Found word: " + word + " = " + altLabel);									
								}
							}
						}
					}
				}
			}
		}
	}

	private Map<String, Integer> createLemmaMap(Map<Integer, Synset> nouns) {
		Map<String, Integer> lemmaMap = new HashMap<String, Integer>();
		for (Map.Entry<Integer, Synset> entry : nouns.entrySet()) {
			for (String lemma : entry.getValue().getLemmas()) {
				lemmaMap.put(lemma, entry.getKey());				
			}
		}
		return lemmaMap;
	}

	public static void printUsage() {
		System.out.println("Usage: ");
		System.out.println("java WordnetEnricher PATH_TO_WORDNET_DATA_NOUN " +
				"PATH_TO_WORDNET_DATA_ADJ PATH_TO_ONTOLOGY OUTPUT");
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length != 4) {
			printUsage();
			System.exit(1);
		}
		new WordnetEnricher().enrich(args);
	}
}
