package nl.telin.mymedia.wordnet;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Wout.Slakhorst
 * Class for reading WordNet file.
 * Creates SynSet and SynLinks
 * Only looks at the derived adjective forms
 */
public class WordnetReader {
	private Map<Integer, Synset> synsets;
	private File input;
	

	public WordnetReader(File f) {
		input = f;
		synsets = new HashMap<Integer, Synset>();
	}

	public Map<Integer, Synset> read() throws IOException {
		BufferedReader buf = new BufferedReader(new FileReader(input));
		String line = "";
		while ((line = buf.readLine()) != null) {
			if (!line.startsWith(" ")) {
				analyseLine(line);
			}			
		}
		buf.close();
		return synsets;
	}

	public Map<Integer, Synset> getSynsets() {
		return synsets;
	}

	private void analyseLine(String line) {
		// cut of the sentence part
		String dataPart = line;
		List<String> lemmas = new ArrayList<String>();

		if (line.indexOf("|") != -1) {
			dataPart = line.substring(0, line.indexOf("|"));
		}

		// subdivide using space as delimiter
		String[] parts = dataPart.split(" ");
		String wordCountPart = parts[3]; // number of lemma's
		int words = 0;
		try {
			words = Integer.parseInt(wordCountPart, 16);
		} catch (NumberFormatException e) {
			// incorrect format
			System.out.println("Can't process line due to NFE: " + line);
			return;
		}
		
		// store the words temporary
		for (int i = 0; i < words; i++) {
			lemmas.add(parts[4 + 2 * i].replace('_', ' ')); // replace _ with space			
		}

		// only add synlinks from n -> a
		if ("n".equals(parts[2])) {
			// continue with pointers
			int ptrOffset = 4 + 2 * words;
			int ptrs = 0;
			try {
				ptrs = Integer.parseInt(parts[ptrOffset]);
			} catch (NumberFormatException e) {
				// incorrect format
				System.out.println("Can't process line due to NFE: " + line);
				return;
			}
	
			// pointers consist of four parts: [ptr_symbol byteoffset type
			// wordmapping]
			// we're only interested in type='a';
			int currentPtr = ptrOffset + 1;
			for (int i = 0; i < ptrs; i++) {
				String byteOffset = parts[currentPtr + 1];
				String type = parts[currentPtr + 2];
				String wordMapping = parts[currentPtr + 3];
				String link = parts[currentPtr];
	
				// ignore all other synsets
				if ("a".equals(type) && "+".equals(link)) {
					addSynlink(parts[0], lemmas, byteOffset, wordMapping);
				}
	
				currentPtr += 4;
			}
		}
		
		addSynset(parts[0], lemmas);
	}

	private void addSynset(String byteOffset, List<String> lemmas) {
		int id = Integer.parseInt(byteOffset);
		Synset set = synsets.get(id);
		if (set == null) {
			set = new Synset(id, lemmas);
			synsets.put(id, set);
		}
	}
	
	private void addSynlink(String byteOffset, List<String> lemmas,
			String adjByteOffset, String wordMapping) {
		int id = Integer.parseInt(byteOffset);
		Synset set = synsets.get(id);
		if (set == null) {
			set = new Synset(id, lemmas);
		}
		set.addSynlink(new Synlink(adjByteOffset, wordMapping));
		synsets.put(id, set);
	}
}
