﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Linq;
using EMIC.Design;
using ExternalSystemIntegration;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Databases;


namespace FacebookIntegrationComponent
{
    /// <summary>
    /// Integration component interacting with Facebook (external social network)
    /// </summary>
    public partial class FacebookIntegrator : AbstractIntegrationDecorator
    {
        // keywords specified in Facebook's responses.
        /// <summary>
        /// Name of the authentication token parameter used in the communication with Facebook 
        /// </summary>
        public const String CgiParameterNameAuthenticationToken = "auth_token";

        /// <summary>
        /// Name of the session key parameter used in the communication with Facebook 
        /// </summary>
        public const String CgiParameterNameSessionKey = "session_key";

        /// <summary>
        /// Name of the Facebook user identification parameter used in the communication with Facebook 
        /// </summary>
        public const String CgiParameterNameFacebookUserId = "uid";

        /// <summary>
        /// Corresponding configuration property name 
        /// </summary>
        protected internal const String PropertyNameApiKey = "API_KEY";

        /// <summary>
        /// Corresponding configuration property name 
        /// </summary>
        protected internal const String PropertyNameApplicationSecret = "APP_SECRET";

        /// <summary>
        /// Corresponding configuration property name 
        /// </summary>
        protected internal const String PropertyNameFacebookLogOnUrl = "FACEBOOK_LOGIN_URL";

        /// <summary>
        /// Corresponding configuration property name 
        /// </summary>
        protected internal const String PropertyNameFacebookRestApi = "FACEBOOK_REST_URL";

        /// <summary>
        /// Corresponding configuration property name 
        /// </summary>
        protected internal const String PropertyNameFacebookOfflineAccessCodeGenUrl = "FACEBOOK_OFFLINE_ACCESS_CODE_GEN_URL";

        /// <summary>
        /// Smart default to be used if the corresponding setting is missing in the configuration
        /// </summary>
        protected internal const String OfflineAccessCodeGenDefaultUrl = "http://www.facebook.com/code_gen.php";

        /// <summary>
        /// Corresponding configuration property name 
        /// </summary>
        protected internal const String PropertyNameAuthenticationCallbackDefaultUri = "AUTHENTICATION_CALLBACK_DEFAULT_URI";

        // component name as registered in the DB/Configuration(ExtSnkComponent.Name)
        private const String integratorsName = "FacebookIntegrator";

        /// <summary>
        /// The name of the parameter required to be passed to the method 'PerformExternalAuthentication()'
        /// </summary>
        public const String AuthParameterCallback = "next";

        private String UrlToFacebookOfflineAccessCodeGeneration;

        // object that performs lower level communication to Facebook API
        private FacebookRest facebookLib;

        private String ApiKey;

        /// <summary>
        /// Default constructor that does not ensure compoent's deployment
        /// </summary>
        public FacebookIntegrator()
            : this(false)
        {
        }

        /// <summary>
        /// Constructor that makes call to its abstract parent's constructor
        /// </summary>
        /// <param name="isDeploymentToBeEnsured">if true, ensures that the DB entries are in line with the configuration;</param>
        /// <exception cref="ExternalSystemIntegratorException">if the component is not registered properly</exception>
        public FacebookIntegrator(bool isDeploymentToBeEnsured)
            : this(isDeploymentToBeEnsured, null)
        {
        }

        /// <summary>
        /// Constructor that makes call to its abstract parent's constructor
        /// </summary>
        /// <param name="isDeploymentToBeEnsured">if true, ensures that the DB entries are in line with the configuration;</param>
        /// <param name="context">Initialized data context to be used (or null)</param>
        /// <exception cref="ExternalSystemIntegratorException">if the component is not registered properly</exception>
        public FacebookIntegrator(bool isDeploymentToBeEnsured, MyMediaDataContext context)
            : base(integratorsName, isDeploymentToBeEnsured, context)
        {
            TraceLog.Write(LogEventType.Verbose, "Facebook's integration component initialized.");
            this.ApiKey = GetProperty(PropertyNameApiKey);
            this.UrlToFacebookOfflineAccessCodeGeneration = GetProperty(PropertyNameFacebookOfflineAccessCodeGenUrl);
            if (this.UrlToFacebookOfflineAccessCodeGeneration == null || this.UrlToFacebookOfflineAccessCodeGeneration.Length == 0){
                this.UrlToFacebookOfflineAccessCodeGeneration = OfflineAccessCodeGenDefaultUrl;
            }
            String AppSecret = GetProperty(PropertyNameApplicationSecret);
            String AuthenticationCallbackDefaultUri = GetProperty(PropertyNameAuthenticationCallbackDefaultUri);
            String FacebookLoginUrl = GetProperty(PropertyNameFacebookLogOnUrl);
            String FacebookRestUrl = GetProperty(PropertyNameFacebookRestApi);
            

            if (ApiKey == null || AppSecret == null)
            {
                throw new ExternalSystemIntegratorException("Facebook's constructor could not load all required properties.");
            }

            facebookLib = new FacebookRest(ApiKey, AppSecret, FacebookLoginUrl, FacebookRestUrl, AuthenticationCallbackDefaultUri);
        }


        /// <summary>
        /// Returns an ExternalSystemIntegratorResult that holds a redirection request(WebBrowserRedirectRequest)
        /// The client application should then call a method specified in the returned 'WebBrowserRedirectRequest.MethodNameToPassCallback'.
        /// </summary>
        /// <param name="dictionaryParameters">Optional parameter called 'next' that is a callback URL
        /// to which Facebook would redirect user's WEB browser after successful authentication or authentication failure can be set. 
        /// Default value from the configuration would be used if it is not provided.
        /// The callback URL is relative to the URL where Facebook application is registered.
        /// </param>
        /// <returns>ExternalSystemIntegratorResult holding user's WEB browser's redirection request</returns>
        public override ExternalSystemIntegratorResult PerformExternalAuthentication(Dictionary<String, Object> dictionaryParameters)
        {
            String callbackUri = null;
            if (dictionaryParameters != null && dictionaryParameters.Keys.Contains<String>(AuthParameterCallback))
            {
                callbackUri = (String)dictionaryParameters[AuthParameterCallback];
            }

            String url2authentication;
            if (!String.IsNullOrEmpty(callbackUri))
            {
                url2authentication = facebookLib.ComposeAuthenticationUrl(callbackUri);
            }
            else
            {
                // default value from configuration will be used.
                TraceLog.Write(LogEventType.Information, "PerformExternalAuthentication(): Using default call back URI on authentication.");
                url2authentication = facebookLib.ComposeAuthenticationUrl();
            }

            WebBrowserRedirectRequest redirect = new WebBrowserRedirectRequest(new Uri(url2authentication), "ProcessAuthenticationCallback");

            return new ExternalSystemIntegratorResult(redirect);
        }

        /// <summary>
        /// Method that needs to be called after calling 'PerformExternalAuthentication(..)' to process 
        /// external authentication's result
        /// </summary>
        /// <param name="cgiParameters">CGI parameters sent in Facebook's callback request</param>
        /// <returns>SESSION_KEY value that is associated with the authenticated user's session(or ExternalSystemIntegratorException on failure)</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public ExternalSystemIntegratorResult ProcessAuthenticationCallback(Dictionary<String, String> cgiParameters)
        {
            if (String.IsNullOrEmpty(cgiParameters[CgiParameterNameAuthenticationToken]))
            {
                throw new ExternalSystemIntegratorException("The authentication token(param '" + CgiParameterNameAuthenticationToken + "') is not provided.");
            }

            // and now 'session_key' parameter must be retrieved from Facebook
            Dictionary<String, String> result = facebookLib.RetrieveSessionKeyAndUserId(cgiParameters[CgiParameterNameAuthenticationToken]);
            if (result == null || !result.ContainsKey(FacebookIntegrator.CgiParameterNameSessionKey))
            {
                throw new ExternalSystemIntegratorException("session_key could not be retrieved from Facebook");
            }

            return new ExternalSystemIntegratorResult(result);
        }

        /// <summary>
        /// Sends specified message to given list of users
        /// </summary>
        /// <param name="extUsers">list of users (initialized with ExternalUserIdentifier) the message is going to be sent to</param>
        /// <param name="message">the message to be sent</param>
        /// <param name="sessionKey"></param>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public void SendMessageToExternalUsers(ICollection<ExternalSystemUser> extUsers, String message, String sessionKey)
        {
            String serialized_ext_users_ids = ExternalSystemUser.SerializeIds(extUsers, ",");
            facebookLib.SendMessageToAUser(serialized_ext_users_ids, message, sessionKey);
        }

        /// <summary>
        /// Sends specified message to given list of users
        /// </summary>
        /// <param name="commaSeparatedFacebookUsersIds">Comma separated IDs of Facebook users</param>
        /// <param name="message">the message to be sent</param>
        /// <param name="sessionKey"></param>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public void SendMessageToExternalUsers(String commaSeparatedFacebookUsersIds, String message, String sessionKey)
        {
            facebookLib.SendMessageToAUser(commaSeparatedFacebookUsersIds, message, sessionKey);
        }

        /// <summary>
        /// Retrieves notifications sending limits for the Facebook application per time various periods.
        /// </summary>
        /// <returns>Facebook response with notifications sending limits</returns>
        public String GetFacebookNotificationLimits()
        {
            String result = "";
            String[] limitationNames = new String[] { "notifications_per_day", "announcement_notifications_per_week", "requests_per_day", "emails_per_day", "email_disable_message_location" };

            foreach (String operationName in limitationNames){
                Dictionary<String, String> cgiParameters = new Dictionary<String, String>();
                cgiParameters.Add("integration_point_name", operationName);
                String value = facebookLib.CallFbMethod("Admin.getAllocation", cgiParameters, "//fb:Admin_getAllocation_response", null, null);

                if (!String.IsNullOrEmpty(result)){
                    result += ",\r\n ";
                }
                result += "'" + operationName + "' -> '" + value +"'";
            }
            TraceLog.Write(LogEventType.Information, "Facebook App allocation values: {0}", new object[]{result});

            return result;
        }


        /// <summary>
        /// Retrieves basic external user's profile(by asking external system)
        /// </summary>
        /// <param name="externalUserIdentifier">The external user id.</param>
        /// <param name="sessionKey"></param>
        /// <returns>ExternalSystemUser with basic profile properties set based on the latest external data</returns>
        public ExternalSystemUser GetUserProfile(String externalUserIdentifier, String sessionKey)
        {
            ExternalSystemUser[] extUsers = facebookLib.GetUserProfile(this.ExternalSystemId, new String[] { externalUserIdentifier }, sessionKey);
            if (extUsers != null && extUsers.Count() > 0)
            {
                return extUsers[0];
            }
            return null;
        }

        /// <summary>
        /// Retrieves list of Facebook friends of an already authenticated user.
        /// </summary>
        /// <param name="sessionKey"></param>
        /// <returns>List of ExternalSystemUser objects representing authenticated 
        /// user's friends at Facebook</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public ICollection<ExternalSystemUser> GetListOfFriends(String sessionKey)
        {
            List<ExternalSystemUser> result;

            String[] friends_ids = facebookLib.RequestListOfFriends(sessionKey);
            TraceLog.Write(LogEventType.Verbose, "get_list_of_friends() : " + friends_ids.ToString());
            ExternalSystemUser[] friends = facebookLib.GetUserProfile(ExternalSystemId, friends_ids, sessionKey);
            result = friends.ToList<ExternalSystemUser>();

            //TraceLog.Write(LogEventType.Verbose, "get_list_of_friends() asks for friends names : " + ExternalSystemUser.SerializeList(result, ", "));
            return result;
        }

        /// <summary>
        /// Returns list of the groups the specified user is a member of
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        public Dictionary<String, String> GetListOfUserGroups(String externalSystemUserId, String sessionKey)
        {
            return facebookLib.GetListOfUserGroups(externalSystemUserId, sessionKey);
        }


        /// <summary>
        /// Asks Facebook whether the Facebook application (with the APP_KEY registered for this component) is installed by the user
        /// </summary>
        /// <param name="sessionKey"></param>
        /// <returns>true, if application is installed, false otherwise</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public virtual ExternalSystemIntegratorResult IsApplicationAdded(String sessionKey)
        {
            return new ExternalSystemIntegratorResult(facebookLib.IsAppAdded(sessionKey));
        }
        
        /// <summary>
        /// Returns URL where user needs to be redirected to generate authentication code required for obtaining permanent session key
        /// </summary>
        /// <returns></returns>
        public Uri GetUrlForGeneratingOfflineAccessCode()
        {
            return new Uri(this.UrlToFacebookOfflineAccessCodeGeneration + "?v=1.0&api_key=" + this.ApiKey);
        }

        /// <summary>
        /// Retrieves permanent session key by offline access code generated by Facebook 
        /// </summary>
        /// <param name="offlineAccessCode">Facebook generated offline access code that user provided to the application.</param>
        /// <returns>String dictionary, holding permanent session key and external system identity</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception on error</exception>
        public virtual Dictionary<String, String> GetPermanentSessionKeyByOfflineAccessCode(String offlineAccessCode)
        {
            Dictionary<String, String> dResult = facebookLib.RetrieveSessionKeyAndUserId(offlineAccessCode);
            if (dResult.ContainsKey(FacebookIntegrator.CgiParameterNameSessionKey))
            {
                return dResult;
            }

            throw new ExternalSystemIntegratorException("Failed to retrieve permanent session key by the specified offline access code(" + offlineAccessCode + ").");
        }

    }
}
