﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ExternalSystemIntegration;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;


namespace FacebookIntegrationComponent
{
    /// <summary>
    /// Integration component interacting with Facebook (external social network)
    /// </summary>
    public partial class FacebookIntegrator
    {
        /// <summary>
        /// Postfix of the name of the configuration property specifying the Facebook template bundle ID 
        /// that is used to publish a story to user's newsfeed (about user liking certain content item)
        /// </summary>
        protected internal const String PropertyNamePostfixTemplateBundleId = "_template_bundle_id";

        /// <summary>
        /// Name of the configuration property specifying the newsfeed text about user liking certain content item
        /// </summary>
        protected internal const String PropertyNameNewsfeedUserLikedItemMessage = "news_feed_text_user_liked_an_item";

        /// <summary>
        /// Complete name of the configuration property specifying the Facebook template bundle ID 
        /// that is used to publish a story to user's newsfeed (about user liking certain content item)
        /// </summary>
        protected internal const String PropertyNameNewsfeedUserLikedItemTemplateBundleId = PropertyNameNewsfeedUserLikedItemMessage + PropertyNamePostfixTemplateBundleId;

        /// <summary>
        /// All possible names of the bundles that are used by the application. (Needed to clear all bundles when requested.)
        /// </summary>
        protected internal String[] AllTemplateBundleNames = new String[] {PropertyNameNewsfeedUserLikedItemTemplateBundleId};

        /// <summary>
        /// Publishes a story to logged in user's feed
        /// </summary>
        /// <param name="contentItemString">Name of the content item (can include HTML, e.g. a link to the item).</param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public bool PublishUserFavoriteInTheNewsfeed(String contentItemString, String sessionKey)
        {
            String bundleId = InitializeUserLikedItemNewsfeedTemplateBundle();

            return facebookLib.PublishUserFavoriteInTheNewsfeed(bundleId, contentItemString, sessionKey);
        }

        /// <summary>
        /// Publishes a story(specified by 'templateBundleId') to logged in user's feed
        /// </summary>
        /// <param name="templateBundleId"></param>
        /// <param name="storyParameters">String Dictionary holding named values of the parameters that should be substituted in the template bundle</param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public bool PublishStoryInUsersNewsfeed(String templateBundleId, Dictionary<String, String> storyParameters, String sessionKey)
        {
            return facebookLib.PublishUserStoryInTheNewsfeed(templateBundleId, storyParameters, sessionKey);
        }

        /// <summary>
        /// Publishes a story(specified by 'templateBundleId') to logged in user's feed
        /// </summary>
        /// <param name="templateBundleId"></param>
        /// <param name="storyJsonEncodedParameters">String Dictionary holding named values of the parameters that should be substituted in the template bundle</param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        public bool PublishUserStoryInTheNewsfeed(String templateBundleId, String storyJsonEncodedParameters, String sessionKey)
        {
            return facebookLib.PublishUserStoryInTheNewsfeed(templateBundleId, storyJsonEncodedParameters, sessionKey);
        }

        /// <summary>
        /// Initializes template bundle ID (registers at Facebook if not registered yet) 
        /// </summary>
        /// <returns></returns>
        public String InitializeUserLikedItemNewsfeedTemplateBundle()
        {
            String templateBundleId = GetProperty(PropertyNameNewsfeedUserLikedItemTemplateBundleId);
            if (!String.IsNullOrEmpty(templateBundleId))
            {
                return templateBundleId;
            }

            // register new template bundle at Facebook
            String msg_user_liked_item = GetProperty(PropertyNameNewsfeedUserLikedItemMessage);

            templateBundleId = facebookLib.RegisterUserNewsfeedTemplateBundle(msg_user_liked_item);
            this.SetProperty(PropertyNameNewsfeedUserLikedItemTemplateBundleId, templateBundleId);
            return templateBundleId;
        }

        /// <summary>
        /// Registers template bundle for users' feed stories
        /// </summary>
        /// <param name="storyTemplate"></param>
        /// <returns>The ID of the newly registered template bundle, that needs to be stored and reused.</returns>
        public String RegisterUserNewsfeedTemplateBundle(String storyTemplate)
        {
            // register new template bundle at Facebook
            String templateBundleId = facebookLib.RegisterUserNewsfeedTemplateBundle(storyTemplate);
            return templateBundleId;
        }

        /// <summary>
        /// Clears(deactivates) all registered application template bundles
        /// Is useful because the number of registered template bundles per application is limited
        /// </summary>
        /// <returns>Number of template bundles deleted</returns>
        public virtual int ClearAllRegisteredApplicationTemplateBundles()
        {
            int numberOfBundlesDeleted = facebookLib.ClearAllRegisteredApplicationTemplateBundles();

            // remove registered template bundles at Facebook 
            foreach (String templateBundle in AllTemplateBundleNames)
            {
                SetProperty(templateBundle, "");
            }
            return numberOfBundlesDeleted;
        }
    }
}