﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Web;
using ExternalSystemIntegration;
using EMIC.Design;
using System.Globalization;
using MyMediaProject.RecommenderSystem.Core;


namespace FacebookIntegrationComponent
{
    /// <summary>
    /// This class provides lower level methods to communicate with Facebook REST API
    /// </summary>
    /// <remarks>The Facebook Application must be already registered at Facebook, and then at MyMedia.</remarks>
    internal class FacebookRest
    {
        private String ApiKey;
        private String AppSecret;
        private String AuthenticationCallbackDefaultUri;
        private String FbLoginUrl;
        private String FbRestUrl;

        private XmlNamespaceManager NamespaceManager; // to hold Facebook's namespace registered

        /// <summary>
        /// Constructor. Requires some Facebook application's data.
        /// </summary>
        /// <param name="apiKey">Application Key generated by Facebook platform for the application.</param>
        /// <param name="appSecret">Application Secret generated by Facebook platform for the application.</param>
        /// <param name="fbLoginUrl">Facebook URL to authenticate users</param>
        /// <param name="fbRestUrl">Facebook REST URL</param>
        /// <param name="authenticationCallbackUri">The authentication callback URI.</param>
        internal FacebookRest(String apiKey, String appSecret, String fbLoginUrl, String fbRestUrl, String authenticationCallbackUri)
        {
            this.ApiKey = apiKey;
            this.AppSecret = appSecret;
            this.AuthenticationCallbackDefaultUri = authenticationCallbackUri;
            this.FbLoginUrl = fbLoginUrl;
            this.FbRestUrl = fbRestUrl;

            // define common XML namespace
            NamespaceManager = new XmlNamespaceManager(new NameTable());
            NamespaceManager.AddNamespace("fb", "http://api.facebook.com/1.0/");
            TraceLog.Write(LogEventType.Information, "Initializing FacebookRest('{0}','{1}','{2}','{3}','{4}')", new object[]{apiKey, appSecret, fbLoginUrl, fbRestUrl, authenticationCallbackUri});
        }

        internal static String GenerateSignature(SortedDictionary<String, String> dictionaryParameters, String secret)
        {
            String result = "";
            foreach (String key in dictionaryParameters.Keys)
            {
                result += key + "=" + dictionaryParameters[key];
            }

            result += secret;

            return Utils.Md5(result);
        }

        /// <summary>
        /// Composes URL that user's WEB browser should be redirected to for Facebook's authentication
        /// </summary>
        /// <param name="callbackUri">The callback URI.</param>
        /// <returns>
        /// Returns the URL that user's WEB browser should be redirected to for Facebook's authentication
        /// </returns>
        internal virtual String ComposeAuthenticationUrl(String callbackUri)
        {
            if (String.IsNullOrEmpty(callbackUri)){
                callbackUri = AuthenticationCallbackDefaultUri;
                TraceLog.Write(LogEventType.Information, "ComposeAuthenticationUrl(): Using default call back URI on authentication.");
            }

            String result = FbLoginUrl + "?v=1.0&api_key=" + ApiKey;
            if (!String.IsNullOrEmpty(callbackUri))
            {
                result += "&next=" + HttpUtility.UrlEncode(callbackUri);
            }
            TraceLog.Write(LogEventType.Verbose, "get_login_url() returns : '" + result + "';");
            return result;
        }

        /// <summary>
        /// Composes URL that user's WEB browser should be redirected to for Facebook's authentication
        /// </summary>
        /// <returns>Returns the URL that user's WEB browser should be redirected to for Facebook's authentication</returns>
        internal virtual String ComposeAuthenticationUrl()
        {
            return ComposeAuthenticationUrl(AuthenticationCallbackDefaultUri);
        }

        /// <summary>
        /// Composes URL to redirect user's browser for authentication at Facebook
        /// </summary>
        /// <param name="methodName"></param>
        /// <param name="cgiParameters"></param>
        /// <param name="sessionKey"></param>
        /// <param name="authToken">Can be null, if not applicable</param>
        /// <returns></returns>
        internal String ComposeRequestUrl(String methodName, Dictionary<String, String> cgiParameters, String sessionKey, String authToken)
        {
            cgiParameters.Add("v", "1.0");
            cgiParameters.Add("method", methodName);
            if (sessionKey != null && sessionKey.Length > 0)
            {
                cgiParameters.Add(FacebookIntegrator.CgiParameterNameSessionKey, sessionKey);
            }
            if (authToken != null){
                cgiParameters.Add(FacebookIntegrator.CgiParameterNameAuthenticationToken, authToken);
            }

            cgiParameters.Add("api_key", this.ApiKey);

            TimeSpan span = DateTime.UtcNow - new DateTime(1970, 1, 1);
            double millis = span.TotalMilliseconds;

            cgiParameters.Add("call_id", millis.ToString(CultureInfo.CurrentCulture));
            cgiParameters.Add("sig", GenerateSignature(new SortedDictionary<String, String>(cgiParameters), this.AppSecret));

            String result = this.FbRestUrl + "?" + Utils.JoinDictionary(cgiParameters, true);

            TraceLog.Write(LogEventType.Verbose, "compose_request_url() returns : '" + result + "';");

            return result;
        }

        /// <summary>
        /// Alias for 'call_fb_method' without parameters
        /// </summary>
        /// <param name="method_name">The method_name.</param>
        /// <returns></returns>
        internal virtual XmlDocument CallFbMethod(String method_name)
        {
            return CallFbMethod(method_name, new Dictionary<String, String>(), null, null);
        }

        /// <summary>
        /// Alias for 'call_fb_method' without parameters
        /// </summary>
        /// <param name="method_name">The method_name.</param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        internal virtual XmlDocument CallFbMethod(String method_name, String sessionKey)
        {
            return CallFbMethod(method_name, new Dictionary<String, String>(), sessionKey);
        }

        /// <summary>
        /// Performs the actual HTTP(s) request to Facebook API and returns requested XML tag value from the FB's response
        /// </summary>
        /// <param name="methodName">Facebook's method name (as defined by Facebook's API)</param>
        /// <param name="cgiParameters">parameters (as defined by Facebook's API) to be sent with the request</param>
        /// <param name="responseValueXPath">The name of the tag(referenced by xPath expression) which value should be returned</param>
        /// <param name="sessionKey"></param>
        /// <param name="authToken"></param>
        /// <returns>XMLDocument with Facebook's response (as defined by Facebook's REST API)</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal virtual String CallFbMethod(String methodName, Dictionary<String, String> cgiParameters, String responseValueXPath, String sessionKey, String authToken)
        {
            XmlDocument xmlResponse = CallFbMethod(methodName, cgiParameters, sessionKey, authToken);

            XmlElement xmlElem = xmlResponse.SelectSingleNode(responseValueXPath, NamespaceManager) as XmlElement;
            if (xmlElem != null){
                return xmlElem.InnerText;
            }

            throw new ArgumentException("The requested tag '" +  responseValueXPath + "' does not exist in the Facebook's response.");
        }

        /// <summary>
        /// Performs the actual HTTP(s) request to Facebook API
        /// </summary>
        /// <param name="methodName">Facebook's method name (as defined by Facebook's API)</param>
        /// <param name="cgiParameters">parameters (as defined by Facebook's API) to be sent with the request</param>
        /// <param name="sessionKey"></param>
        /// <returns>XMLDocument with Facebook's response (as defined by Facebook's REST API)</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal XmlDocument CallFbMethod(String methodName, Dictionary<String, String> cgiParameters, String sessionKey)
        {
            return CallFbMethod(methodName, cgiParameters, sessionKey, null);
        }


        /// <summary>
        /// Performs the actual HTTP(s) request to Facebook API
        /// </summary>
        /// <param name="methodName">Facebook's method name (as defined by Facebook's API)</param>
        /// <param name="cgiParameters">parameters (as defined by Facebook's API) to be sent with the request</param>
        /// <param name="sessionKey"></param>
        /// <param name="authToken">Can be null, if not applicable.</param>
        /// <returns>XMLDocument with Facebook's response (as defined by Facebook's REST API)</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal XmlDocument CallFbMethod(String methodName, Dictionary<String, String> cgiParameters, String sessionKey, String authToken)
        {
            String url2facebook = ComposeRequestUrl(methodName, cgiParameters, sessionKey, authToken);
            String response = Utils.MakeHttpsRequest(new Uri(url2facebook));
            if (response == null || response.Length < 10)
            {
                throw new ExternalSystemIntegratorException("Facebook did not reply correctly to the request: '" + url2facebook + "'.");
            }
            XmlDocument LastXmlDoc = new XmlDocument();

            try
            {
                LastXmlDoc.LoadXml(response);
                String xPath = "//fb:error_response";

                XmlElement xmlElem = LastXmlDoc.SelectSingleNode(xPath, NamespaceManager) as XmlElement;
                // if there is an error response, try to extract the error code & msg
                if (xmlElem != null)
                {
                    TraceLog.Write(LogEventType.Verbose, "Facebook returned an error. Processing..");
                    String error_code = (xmlElem.SelectSingleNode("//fb:error_response/fb:error_code", NamespaceManager) as XmlElement).InnerText;
                    String error_msg = (xmlElem.SelectSingleNode("//fb:error_response/fb:error_msg", NamespaceManager) as XmlElement).InnerText;

                    throw new ExternalSystemIntegratorException("Facebook Communication Error: " + error_code + " : " + error_msg);
                }
            }
            catch (System.Xml.XmlException ex)
            {
                throw new ExternalSystemIntegratorException(ex.ToString());
            }

            return LastXmlDoc;
        }


        /// <summary>
        /// Performs HTTP request to Facebook REST API(by calling 'call_method(..)') and returns session_key that is 
        /// an identifier of the authenticated user's session on Facebook
        /// </summary>
        /// <param name="auth_token"></param>
        /// <returns>session_key's value communicated by Facebook (as defined by Facebook's API)</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal virtual Dictionary<String, String> RetrieveSessionKeyAndUserId(String auth_token)
        {
            Dictionary<String, String> result = new Dictionary<String, String>();
            XmlDocument LastXmlDoc = CallFbMethod("facebook.auth.getSession", new Dictionary<string,string>(), null, auth_token);

            String xPath = "//fb:auth_getSession_response/fb:" + FacebookIntegrator.CgiParameterNameSessionKey;
            XmlElement xmlSessionKeyElem = LastXmlDoc.SelectSingleNode(xPath, NamespaceManager) as XmlElement;
            if (xmlSessionKeyElem == null)
            {
                throw new ExternalSystemIntegratorException("No valid response received from Facebook.");
            }
            String sessionKey = xmlSessionKeyElem.InnerText;
           
            // extract the 'uid' from the response
            xPath = "//fb:auth_getSession_response/fb:" + FacebookIntegrator.CgiParameterNameFacebookUserId;

            XmlElement xmlUidElem = LastXmlDoc.SelectSingleNode(xPath, NamespaceManager) as XmlElement;
            if (xmlUidElem == null)
            {
                throw new ExternalSystemIntegratorException("No valid response received from Facebook.");
            }
            String uid = xmlUidElem.InnerText;


            result.Add(FacebookIntegrator.CgiParameterNameSessionKey, sessionKey);
            result.Add(FacebookIntegrator.CgiParameterNameFacebookUserId, uid);
            // include authentication token into the result for convenience
            result.Add(FacebookIntegrator.CgiParameterNameAuthenticationToken, auth_token);

            return result;
        }

        /// <summary>
        /// Returns array of user's friends' Facebook IDs
        /// </summary>
        /// <param name="sessionKey"></param>
        /// <returns>Array of friends' Facebook IDs of the authenticated user</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal virtual String[] RequestListOfFriends(String sessionKey)
        {
            List<String> result = new List<String>();
            XmlDocument LastXmlDoc = CallFbMethod("facebook.friends.get", new Dictionary<String, String>(), sessionKey);

            String xPath = "//fb:friends_get_response/fb:uid";

            XmlNodeList xmlFriendsNodes = LastXmlDoc.SelectNodes(xPath, NamespaceManager) as XmlNodeList;
            if (xmlFriendsNodes == null)
            {
                throw new ExternalSystemIntegratorException("No valid response received from Facebook.");
            }

            foreach (XmlNode friend in xmlFriendsNodes)
            {
                result.Add(friend.InnerText);
            }
            return result.ToArray<String>();
        }

        /// <summary>
        /// Returns array of initilized ExternalSystemUser objects holding basic profile information.
        /// </summary>
        /// <param name="externalSystemId">External system Id (needed to initialize the objects)</param>
        /// <param name="externalUsersIdentifiers">Array of Facebook Ids of the users whose profile data is requested</param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal virtual ExternalSystemUser[] GetUserProfile(int externalSystemId, String[] externalUsersIdentifiers, String sessionKey)
        {
            List<ExternalSystemUser> fbLResult = new List<ExternalSystemUser>();
            String[] requestedFieldsAsArray = { "first_name", "last_name" }; // which data to retrieve from Facebook

            Dictionary<String, String> dParams = new Dictionary<String, String>();
            dParams.Add("uids", Utils.JoinArray(externalUsersIdentifiers, ","));
            dParams.Add("fields", Utils.JoinArray(requestedFieldsAsArray, ","));
            XmlDocument LastXmlDoc = CallFbMethod("facebook.users.getInfo", dParams, sessionKey);

            String xPath = "//fb:users_getInfo_response/fb:user";

            XmlNodeList xmlFriendsNodes = LastXmlDoc.SelectNodes(xPath, NamespaceManager) as XmlNodeList;
            if (xmlFriendsNodes == null)
            {
                throw new ExternalSystemIntegratorException("No valid response received from Facebook.");
            }

            TraceLog.Write(LogEventType.Verbose, "Number of users = " + xmlFriendsNodes.Count);

            for (int i = 0; i < xmlFriendsNodes.Count; i++)
            {
                XmlNodeList friendFields = xmlFriendsNodes.Item(i).ChildNodes;
                ExternalSystemUser externalSystemUser = new ExternalSystemUser();

                String FirstName = "";
                String LastName = "";
                String UserId = "";
                for (int j = 0; j < friendFields.Count; j++)
                {
                    XmlNode friendField = friendFields.Item(j);
                    //TraceLog.Write(LogEventType.Verbose, " Name-Value === '" + friendField.Name + "' = " + friendField.InnerXml);
                    if (friendField.Name == FacebookIntegrator.CgiParameterNameFacebookUserId) externalSystemUser.UserExternalIdentifier = friendField.InnerXml;
                    if (friendField.Name == "first_name") FirstName = friendField.InnerXml;
                    if (friendField.Name == "last_name") LastName = friendField.InnerXml;
                    if (friendField.Name == "uid") UserId = friendField.InnerXml;
                }
                externalSystemUser.ExternalSystemId = externalSystemId;
                externalSystemUser.FirstName = FirstName.Trim();
                externalSystemUser.LastName = LastName.Trim();
                externalSystemUser.UserExternalIdentifier = UserId.Trim();

                fbLResult.Add(externalSystemUser);
            }
            return fbLResult.ToArray<ExternalSystemUser>();
        }

        /// <summary>
        /// Registers template bundle for users' feed stories
        /// </summary>
        /// <param name="storyTemplate"></param>
        /// <returns>The ID of the newly registered template bundle, that needs to be stored and reused.</returns>
        internal virtual String RegisterUserNewsfeedTemplateBundle(String storyTemplate)
        {
            TraceLog.Write(LogEventType.Information, "Registering new feed template bundle: '{0}'.", new object[] { storyTemplate });

            Dictionary<String, String> dParams = new Dictionary<String, String>();

            //  = "[\"{*actor*} has been playing.\"]";
            String templateBundle = "[\"" + storyTemplate + "\"]";
            
            dParams.Add("one_line_story_templates", templateBundle);

            String xPath = "//fb:Feed_registerTemplateBundle_response";

            // if it fails, exception will be thrown
            String templateBundleId = CallFbMethod("Feed.registerTemplateBundle", dParams, xPath, null, null);

            if (templateBundleId != null && templateBundleId.Length > 0)
            {
                return templateBundleId;
            }

            throw new ExternalSystemIntegratorException("Failed to register a template bundle at Facebook.");
        }

        /// <summary>
        /// Sends a notification to the specified user.
        /// </summary>
        /// <param name="uids"></param>
        /// <param name="messageText"></param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal virtual bool SendMessageToAUser(String uids, String messageText, String sessionKey)
        {
            Dictionary<String, String> dParams = new Dictionary<String, String>();

            dParams.Add("to_ids", uids);

            dParams.Add("notification", messageText);

            // if it fails, exception will be thrown
            CallFbMethod("facebook.notifications.send", dParams, sessionKey);
            return true;
        }

        /// <summary>
        /// Returns list of the groups the specified user is a member of
        /// </summary>
        /// <param name="externalSystemUserId"></param>
        /// <param name="sessionKey"></param>
        /// <returns>Dictionary style pairs where key is the external identifier of a group, and value is it's name</returns>
        public Dictionary<String, String> GetListOfUserGroups(String externalSystemUserId, String sessionKey)
        {
            Dictionary<String, String> dParams = new Dictionary<String, String>();

            dParams.Add("uid", externalSystemUserId);

            Dictionary<String, String> groupsDictionary = new Dictionary<String, String>();
            XmlDocument LastXmlDoc = CallFbMethod("facebook.groups.get", dParams, sessionKey);

            String xPath = "//fb:groups_get_response/fb:group";

            XmlNodeList xmlGroupNodes = LastXmlDoc.SelectNodes(xPath, NamespaceManager) as XmlNodeList;
            if (xmlGroupNodes == null)
            {
                throw new ExternalSystemIntegratorException("No valid response received from Facebook.");
            }

            foreach (XmlNode group_item in xmlGroupNodes)
            {
                String groupId = group_item.SelectSingleNode("fb:gid", NamespaceManager).InnerText;
                String groupName = group_item.SelectSingleNode("fb:name", NamespaceManager).InnerText;
                if (groupsDictionary.ContainsKey(groupId))
                {
                    continue;
                }
                groupsDictionary.Add(groupId, groupName);
            }

            return groupsDictionary;
        }


        /// <summary>
        /// Adds a story to user's news feed about user favourite content item
        /// </summary>
        /// <param name="templateBundleId"></param>
        /// <param name="contentItemString"></param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        internal virtual bool PublishUserFavoriteInTheNewsfeed(String templateBundleId, String contentItemString, String sessionKey)
        {
            Dictionary<String, String> dParams = new Dictionary<String, String>();

            dParams.Add("template_bundle_id", templateBundleId);
            dParams.Add("template_data", "{\"item_name\":\"" + contentItemString + "\"}");

            // if it fails, exception will be thrown
            CallFbMethod("Feed.publishUserAction", dParams, sessionKey);
            return true;
        }

        /// <summary>
        /// Adds a story to user's news feed about user favourite content item
        /// </summary>
        /// <param name="templateBundleId"></param>
        /// <param name="storyParameters">String Dictionary holding named values of the parameters that should be substituted in the template bundle</param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        internal virtual bool PublishUserStoryInTheNewsfeed(String templateBundleId, Dictionary<String, String> storyParameters, String sessionKey)
        {
            Dictionary<String, String> dParams = new Dictionary<String, String>();

            dParams.Add("template_bundle_id", templateBundleId);

            String sTemplateDataFieldValue = "";
            for (int i = 0; i < storyParameters.Keys.Count(); i++)
            {
                String sKey = storyParameters.Keys.ElementAt(i);
                String sValue = storyParameters.Values.ElementAt(i);
                if (sValue != null && sValue.Length > 0)
                {
                    sValue = sValue.Replace("\"", "\\\""); // replacing:  "  ->  \"
                }
                if (sTemplateDataFieldValue.Length > 0)
                {
                    sTemplateDataFieldValue += ",";
                }
                sTemplateDataFieldValue += "\"" + sKey + "\":\""+ sValue+ "\"";
            }

            sTemplateDataFieldValue = "[{" + sTemplateDataFieldValue + "}]";

            dParams.Add("template_data", sTemplateDataFieldValue);

            // if it fails, exception will be thrown
            CallFbMethod("Feed.publishUserAction", dParams, sessionKey);
            return true;
        }


        /// <summary>
        /// Adds a story to user's news feed about user favourite content item
        /// </summary>
        /// <param name="templateBundleId"></param>
        /// <param name="storyJsonEncodedParameters">JSON String holding named values of the parameters that should be substituted in the template bundle</param>
        /// <param name="sessionKey"></param>
        /// <returns></returns>
        internal virtual bool PublishUserStoryInTheNewsfeed(String templateBundleId, String storyJsonEncodedParameters, String sessionKey)
        {
            Dictionary<String, String> dParams = new Dictionary<String, String>();

            dParams.Add("template_bundle_id", templateBundleId);
            dParams.Add("template_data", storyJsonEncodedParameters);

            // if it fails, exception will be thrown
            CallFbMethod("Feed.publishUserAction", dParams, sessionKey);
            return true;
        }



        /// <summary>
        /// Asks Facebook whether the Facebook application (with the APP_KEY registered for this component) is installed by the user
        /// </summary>
        /// <param name="sessionKey"></param>
        /// <returns>true, if application is installed, false otherwise</returns>
        /// <exception cref="ExternalSystemIntegratorException">Throws exception if some error occures while communicating to the Facebook API</exception>
        internal virtual bool IsAppAdded(String sessionKey)
        {
            // if it fails, exception will be thrown
            XmlDocument LastXmlDoc = CallFbMethod("Users.isAppAdded", sessionKey);
            String xPath = "//fb:Users_isAppAdded_response";

            XmlElement xmlElem = LastXmlDoc.SelectSingleNode(xPath, NamespaceManager) as XmlElement;
            if (xmlElem == null)
            {
                throw new ExternalSystemIntegratorException("No valid response received from Facebook.");
            }
            String response = xmlElem.InnerText;

            return (response == "1") ? true : false;
        }

        /// <summary>
        /// Retrieves all template bundles registered for the application
        /// </summary>
        /// <returns>List of IDs, or empty List if no bundles are registered</returns>
        internal IEnumerable<String> GetRegisteredTemplateBundleIds()
        {
            List<String> templateBundlesIds = new List<String>();
            XmlDocument xmlResponse = CallFbMethod("feed.getRegisteredTemplateBundles");
            String xPath = "//fb:feed_getRegisteredTemplateBundles_response/fb:template_bundle/fb:template_bundle_id";

            XmlNodeList xmlTemplateBundlesIds = xmlResponse.SelectNodes(xPath, NamespaceManager) as XmlNodeList;
            if (xmlTemplateBundlesIds == null)
            {
                return templateBundlesIds;
            }

            TraceLog.Write(LogEventType.Verbose, "Number of template bundles retrieved: {0}" , new object[]{xmlTemplateBundlesIds.Count});

            for (int i = 0; i < xmlTemplateBundlesIds.Count; i++)
            {
                templateBundlesIds.Add(xmlTemplateBundlesIds.Item(i).InnerText);
            }

            return templateBundlesIds;
        }

        /// <summary>
        /// Clears(deactivates) all registered application template bundles
        /// Is useful because the number of registered template bundles per application is limited
        /// </summary>
        /// <returns>Number of template bundles deleted</returns>
        internal int ClearAllRegisteredApplicationTemplateBundles(){
            IEnumerable<String> templateBundlesIds = GetRegisteredTemplateBundleIds();
            foreach (String templateBundleId in templateBundlesIds)
            {
                Dictionary<string, string> dictionaryParameters = new Dictionary<string, string>();
                dictionaryParameters.Add("template_bundle_id", templateBundleId);
                TraceLog.Write(LogEventType.Information, "Deactivating application template bundle with id='{0}'.", new object[] { templateBundleId });
                CallFbMethod("feed.deactivateTemplateBundleByID", dictionaryParameters, null);
            }

            return templateBundlesIds.Count();
        }
    }
}
