﻿using System;
using System.Web;
using System.Collections.Specialized;
using System.Collections.ObjectModel;
using EMIC.Design;
using ExternalSystemIntegration;
using System.Collections.Generic;
using ExtremeSwank.OpenId;
using ExtremeSwank.OpenId.PlugIns.Extensions;
using MyMediaProject.RecommenderSystem.Core;

namespace OpenIdIntegrationComponent
{
    /// <summary>
    /// Component providing OpenID Authentication and authenticated user's data retrieval
    /// </summary>
    public class OpenIdIntegrator : AbstractIntegrationDecorator
    {
        // component name's 
        private const String integratorsName = "OpenIdIntegrator";

        /// <summary>
        /// Configuration property name for IsExtensionsUsageDisabled
        /// </summary>
        public const String PropertyNameIsExtensionsUsageDisabled = "IsExtensionsUsageDisabled";


        /// <summary>
        /// Name of the corresponding configuration setting (as used in the configuration files)
        /// </summary>
        public const String PropertyNameAuthenticationCallbackDefaultUri = "AUTHENTICATION_CALLBACK_DEFAULT_URI";

        /// <summary>
        /// Name of the corresponding configuration setting (as used in the configuration files)
        /// </summary>
        public const String PropertyNameAuthenticationCallbackDefaultTrustedHostUri = "AUTHENTICATION_CALLBACK_DEFAULT_TRUSTED_HOST_URI";


        /// <summary>
        /// Determines whether OpenID extensions usage is disabled by the global configuration
        /// </summary>
        /// <returns></returns>
        private bool IsExtensionsUsageDisabled()
        {
            String extensionsDisabledString = this.GetProperty(OpenIdIntegrator.PropertyNameIsExtensionsUsageDisabled);

            if (String.Equals("true", extensionsDisabledString))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Constructor that makes call to its abstract parent's constructor
        /// </summary>
        /// <exception cref="ExternalSystemIntegratorException">if the component is not registered properly</exception>
        public OpenIdIntegrator()
            : base(integratorsName)
        {
            TraceLog.Write(LogEventType.Verbose, "OpenID's integratoin component initialized.");
        }

        /// <summary>
        /// Returns an ExternalSystemIntegratorResult that holds a redirection request(WebBrowserRedirectRequest)
        /// The client application should then call a method specified in the returned 'WebBrowserRedirectRequest.MethodNameToPassCallback'.
        /// </summary>
        /// <param name="dictionaryParameters">Parameter specifying claimed openID identity(PerformAuthenticationParameterNameIdentity)
        /// is expected as a dictionary key=value pair.
        /// The redirection URL to user's OpenID provider will be returned.
        /// </param>
        /// <returns>ExternalSystemIntegratorResult holding user's WEB browser's redirection request</returns>
        public override ExternalSystemIntegratorResult PerformExternalAuthentication(Dictionary<String, Object> dictionaryParameters)
        {
            List<String> requiredFields = null;
            List<String> optionalFields = null;

            if (!IsExtensionsUsageDisabled())
            {
                if (dictionaryParameters.ContainsKey(PerformAuthenticationParameterKeyNames.RequiredSimpleRegistrationFieldsList))
                {
                    requiredFields = (List<String>)dictionaryParameters[PerformAuthenticationParameterKeyNames.RequiredSimpleRegistrationFieldsList];
                }

                if (dictionaryParameters.ContainsKey(PerformAuthenticationParameterKeyNames.OptionalSimpleRegistrationFieldsList))
                {
                    optionalFields = (List<String>)dictionaryParameters[PerformAuthenticationParameterKeyNames.OptionalSimpleRegistrationFieldsList];
                }
            }

            ReadOnlyCollection<String> readOnlyRequiredFields;
            if (requiredFields != null)
            {
                readOnlyRequiredFields = new ReadOnlyCollection<String>(requiredFields);
            }
            else
            {
                readOnlyRequiredFields = new ReadOnlyCollection<String>(new List<String>());
            }

            ReadOnlyCollection<String> readOnlyOptionalFields;
            if (optionalFields != null)
            {
                readOnlyOptionalFields = new ReadOnlyCollection<String>(optionalFields);
            }
            else
            {
                readOnlyOptionalFields = new ReadOnlyCollection<String>(new List<String>());
            }


            OpenIdClient openid = GetClient(new NameValueCollection(), readOnlyRequiredFields, readOnlyOptionalFields);

            // set the identity to be verified
            openid.Identity = NormalizeOpenIdIdentity((String)dictionaryParameters[PerformAuthenticationParameterKeyNames.Identity]);

            // Discover the OpenID Provider from the Identity and
            // return the redirect request for the browser.
            return new ExternalSystemIntegratorResult(new WebBrowserRedirectRequest(openid.CreateRequest(false, false), "ProcessProvidersAuthenticationCallback"));
        }

        /// <summary>
        /// Bringing identities to a common format (without trailing slash '/')
        /// </summary>
        /// <returns></returns>
        private static String NormalizeOpenIdIdentity(String identity)
        {
            String initialIdentity = identity;

            if (String.IsNullOrEmpty(identity)){
                return null;
            }
            
            identity = identity.Trim();

            // ensure protocol prefix is present
            if (!identity.StartsWith("http://", StringComparison.CurrentCulture) && !identity.StartsWith("https://", StringComparison.CurrentCulture))
            {
                identity = "http://" + identity;
            }

            // remove trailing slash
            if (identity.EndsWith("/", StringComparison.CurrentCulture))
            {
                identity = identity.Substring(0, identity.Length - 1);
            }

            if (!String.Equals(initialIdentity, identity, StringComparison.CurrentCulture)){
                TraceLog.Write(LogEventType.Verbose, "OpenId identity normalized: '{0}' --> '{1}'.", new object[] { initialIdentity , identity});
            }

            return identity;
        }


        /// <summary>
        /// This method needs to be called after calling 'PerformExternalAuthentication(..)' to process 
        /// external authentication's result
        /// </summary>
        /// <param name="claimedIdentity">claimed OpenId identity of the user</param>
        /// <param name="rawCallbackUri">The complete unchanged OpenID provider's callback URL as String (passing as URI would decode it)</param>
        /// <param name="requiredFields"></param>
        /// <param name="optionalFields"></param>
        /// <returns></returns>
        public ExternalSystemIntegratorResult ProcessProvidersAuthenticationCallback(String claimedIdentity, Uri rawCallbackUri, ReadOnlyCollection<String> requiredFields, ReadOnlyCollection<String> optionalFields)
        {
            return ProcessProvidersAuthenticationCallback(claimedIdentity, rawCallbackUri.ToString(), requiredFields, optionalFields);
        }

        /// <summary>
        /// This method needs to be called after calling 'PerformExternalAuthentication(..)' to process 
        /// external authentication's result
        /// </summary>
        /// <param name="claimedIdentity">claimed OpenId identity of the user</param>
        /// <param name="rawCallbackUri">The complete unchanged OpenID provider's callback URL as String (passing as URI would decode it)</param>
        /// <param name="requiredFields"></param>
        /// <param name="optionalFields"></param>
        /// <returns></returns>
        public ExternalSystemIntegratorResult ProcessProvidersAuthenticationCallback(String claimedIdentity, String rawCallbackUri, ReadOnlyCollection<String> requiredFields, ReadOnlyCollection<String> optionalFields)
        {
            OpenIdUser lastAuthenticatedUser = null;
            TraceLog.Write(LogEventType.Information, "ProcessProvidersAuthenticationCallback called with claimedIdentity='{0}' and callbackUri = '{1}'.", new object[] { claimedIdentity, rawCallbackUri });
            NameValueCollection arguments = GetUrlArguments(rawCallbackUri);

            //foreach (String key in arguments.AllKeys)
            //{
            //    TraceLog.Write(LogEventType.Verbose, "GetUrlArguments: '" + key + "' = '" + arguments[key] + "'.");
            //}

            OpenIdClient openid = null;
            claimedIdentity = NormalizeOpenIdIdentity(claimedIdentity);

            try
            {
                openid = GetClient(arguments, requiredFields, optionalFields);

                if (openid.ValidateResponse())
                {
                    // bring the result to common format
                    openid.Identity = NormalizeOpenIdIdentity(openid.Identity);
                    
                    // if claimed identity does not coincide with provider's response
                    if (!String.Equals(claimedIdentity, openid.Identity, StringComparison.CurrentCultureIgnoreCase))
                    {
                        throw new ExternalSystemIntegratorException("The claimed identity ('" + claimedIdentity + "') does not coincide with the one determined by the OpenID provider('" + openid.Identity + "').");
                    }

                    lastAuthenticatedUser = openid.RetrieveUser();
                }
            }
            catch (ExternalSystemIntegratorException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new ExternalSystemIntegratorException("User authentication cancelled by user (or incorrect fallback from OpenID provider): \r\n" + ex.ToString());
            }

            if (lastAuthenticatedUser == null)
            {
                throw new ExternalSystemIntegratorException("OpenID authentication failed.");
            }

            return new ExternalSystemIntegratorResult(lastAuthenticatedUser);
        }


        /// <summary>
        /// Utility method to parse URL extracting CGI parameters
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        private static NameValueCollection GetUrlArguments(String uriAsString)
        {
            if (uriAsString == null)
            {
                return new NameValueCollection();
            }

            int paramsPosition = uriAsString.IndexOf('?');
            if (paramsPosition < 0)
            {
                return new NameValueCollection();
            }

            String queryString = uriAsString.Substring(paramsPosition);

            return HttpUtility.ParseQueryString(queryString);
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="arguments"></param>
        /// <param name="requiredSimpleRegistrationFields">List of mandatory fields of the SimpleReigstraionExtension to be asked from the user</param>
        /// <param name="optionalSimpleRegistrationFields">List of optional fields of the SimpleReigstraionExtension to be asked from the user</param>
        /// <returns></returns>
        private OpenIdClient GetClient(NameValueCollection arguments, ReadOnlyCollection<String> requiredSimpleRegistrationFields, ReadOnlyCollection<String> optionalSimpleRegistrationFields)
        {
            OpenIdClient openid = new OpenIdClient(arguments);

            openid.ReturnUrl = new Uri(GetProperty(PropertyNameAuthenticationCallbackDefaultUri));
            openid.TrustRoot = GetProperty(PropertyNameAuthenticationCallbackDefaultTrustedHostUri);

            // add simple registration extension fields request, if any wanted
            if (requiredSimpleRegistrationFields != null && requiredSimpleRegistrationFields.Count > 0
                || optionalSimpleRegistrationFields != null && optionalSimpleRegistrationFields.Count > 0)
            {
                SimpleRegistration simpleRegistrationFields = new SimpleRegistration(openid);

                if (requiredSimpleRegistrationFields != null && requiredSimpleRegistrationFields.Count > 0)
                {
                    foreach (String field in requiredSimpleRegistrationFields)
                    {
                        simpleRegistrationFields.AddRequiredFields(field);
                    }
                }

                if (optionalSimpleRegistrationFields != null && optionalSimpleRegistrationFields.Count > 0)
                {
                    foreach (String field in optionalSimpleRegistrationFields)
                    {
                        simpleRegistrationFields.AddOptionalFields(field);
                    }
                }
            }

            return openid;
        }
    }
}
