﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EMIC.Design;
using System.Data;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;
using VirtualProfiles;
using ExternalSystemIntegration;

namespace VirtualProfileTests
{
    /// <summary>
    /// Summary description for VirtualProfilesTests
    /// </summary>
    [TestClass]
    public class VirtualProfilesTests
    {
        static VirtualProfilesManager vpManager;
        static RecommendationsManager RecommendationsMngr;
        private static MyMediaDataContext Context;
        private static int tempModuleId;
        private static String publicProfileName;
        private TestContext testContextInstance;

        // Called once before executing the tests
        [ClassInitialize]
        public static void TestsInitialization(TestContext testContext)
        {
            TraceLog.Write(LogEventType.Verbose, "BeforeAllTests called.");
            publicProfileName = "[Test @ " + DateTime.Now + "] (" + DateTime.Now.Ticks + ")";
            TraceLog.Write(LogEventType.Verbose, "VirtualProfilesTests(): constructor called.");

            Context = new MyMediaDataContext(ConfigManager.TestDBConnectionString);
            
            // The DBConnectionString' value in memory has to be changed, because VirtualProfilesManager uses configuration directly
            ConfigManager.Modules[ConfigConstants.DefaultModuleName].Properties[ConfigConstants.PropertyNameDBConnectionString].Value = ConfigManager.TestDBConnectionString;
            if (!Context.DatabaseExists())
            {
                Context.CreateDatabase();
            }

            vpManager = new VirtualProfilesManager();
            RecommendationsMngr = new RecommendationsManager(Context);
        }

        //[TestMethod]
        public void TestVirtualProfileInvitations()
        {
            User ownerUser = createTestUser();
            User otherUser = createTestUser();
            ExternalSystemUser extUser = CreateExternalTestUser("tester");


            // test ext users properties
            using (UsersManager usersMngr = new UsersManager())
            {
                ExternalSystemUser user2 = usersMngr.GetExternalSystemUserById(extUser.ExternalSystemUserId);
                Assert.IsTrue(extUser.ExternalSystemUserId == user2.ExternalSystemUserId);

                String somePropertyName = "test setting 1";
                String someValue = "some test value";
                String someValue2 = "updated test value 2";
                ExternalSystemUserProperty property = usersMngr.StoreExternalSystemUserProperty(user2.ExternalSystemUserId, somePropertyName, someValue);
                Assert.IsNotNull(property);
                Assert.IsTrue(property.ExternalSystemUserId == user2.ExternalSystemUserId && property.ExternalSystemUserPropertyId > 0 && property.Name == somePropertyName && property.Value == someValue);

                Assert.IsTrue(someValue.Equals(usersMngr.GetExternalSystemUserPropertyValue(user2.ExternalSystemUserId, somePropertyName)));

                property = usersMngr.StoreExternalSystemUserProperty(user2.ExternalSystemUserId, somePropertyName, someValue2);
                Assert.IsNotNull(property);
                Assert.IsTrue(property.ExternalSystemUserId == user2.ExternalSystemUserId 
                                && property.ExternalSystemUserPropertyId > 0 
                                && property.Name == somePropertyName 
                                && property.Value == someValue2);

                Assert.IsTrue(someValue2.Equals(usersMngr.GetExternalSystemUserPropertyValue(user2.ExternalSystemUserId, somePropertyName)));
            }

            VirtualProfile publicProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + "_TestVirtualProfileInvitations", "description " + publicProfileName, true, true, "http://designyoutrust.com/wp-content/uploads2/bla.jpg", "http://dipity.s3.amazonaws.com/uploads/timelines/8bf8400b92eb09da6f5cf0de27e2503c_1s.png");

            VirtualProfileUsageInvitation invitation = (VirtualProfileUsageInvitation)RecommendationsMngr.RegisterNewRecommendation(default(int), publicProfile.VirtualProfileId, ownerUser, otherUser);

            TraceLog.Write(LogEventType.Verbose, "\r\nInvited to id=" + invitation.ContentRecommendationId + ", vp=" + invitation.VirtualProfileId + "\r\n");

            // test rec-n to external user
            VirtualProfileUsageInvitation invitation2 = (VirtualProfileUsageInvitation)RecommendationsMngr.RegisterNewRecommendation(default(int), publicProfile.VirtualProfileId, ownerUser, extUser);
            Assert.IsTrue(invitation2.ContentRecommendationId > 0 && invitation2.ToUserId == null && invitation2.VirtualProfileId == publicProfile.VirtualProfileId && invitation2.ToExternalSystemUserId == extUser.ExternalSystemUserId);

            IEnumerable<ContentRecommendation> invitations2Ext = RecommendationsMngr.GetActiveVirtualProfileInvitations(extUser);
            Assert.IsTrue(invitations2Ext != null && invitations2Ext.Count() > 0);
            foreach (ContentRecommendation activeInvitation in invitations2Ext)
            {
                VirtualProfileUsageInvitation vpInvitation = (VirtualProfileUsageInvitation)activeInvitation;
                Assert.IsTrue(vpInvitation.ToExternalSystemUserId == extUser.ExternalSystemUserId);
                Assert.IsTrue(RecommendationsMngr.MarkRecommendationAsRejected(vpInvitation.ContentRecommendationId));
            }
            IEnumerable<ContentRecommendation> invitations2ExtAfterRejection = RecommendationsMngr.GetActiveVirtualProfileInvitations(extUser);
            Assert.IsTrue(invitations2ExtAfterRejection == null || invitations2ExtAfterRejection.Count() == 0);
                        

            IEnumerable<VirtualProfileUsageInvitation> invitations = RecommendationsMngr.GetActiveVirtualProfileInvitations(ownerUser);
            Assert.IsTrue(invitations == null || invitations.Count() == 0, "Invitation has been registered to VP owner??!!");


            invitations = RecommendationsMngr.GetActiveVirtualProfileInvitations(otherUser);
            TraceLog.Write(LogEventType.Verbose, "\r\n# of invitations found: " + invitations.Count());
            Assert.IsTrue(invitations != null && invitations.Count() > 0);

            foreach (VirtualProfileUsageInvitation activeInvitation in invitations)
            {
                Assert.IsTrue(RecommendationsMngr.MarkRecommendationAsRead(activeInvitation.ContentRecommendationId), "Failed to inactivate an invitation.");
            }

            invitations = RecommendationsMngr.GetActiveVirtualProfileInvitations(otherUser);
            Assert.IsTrue(invitations == null || invitations.Count() == 0, "Active invitations should not exists already. Active invitations nr: " + invitations.Count());
        }

        [TestMethod]
        public void TestRecommendations()
        {
            User ownerUser = createTestUser();
            User otherUser = createTestUser();

            IEnumerable<CatalogItemRecommendation> recommendations = RecommendationsMngr.GetActiveCatalogItemRecommendations(otherUser);
            Assert.IsTrue(recommendations == null || recommendations.Count() == 0, "Invitations should not have been sent to the recently created user.");

            int dummyCatalogItemId = 1;
            CatalogItemRecommendation recommendation = (CatalogItemRecommendation)RecommendationsMngr.RegisterNewRecommendation(dummyCatalogItemId, default(int), ownerUser, otherUser);

            // ensure that duplicate recommendations are not created
            CatalogItemRecommendation duplicateRecommendation = (CatalogItemRecommendation)RecommendationsMngr.RegisterNewRecommendation(dummyCatalogItemId, default(int), ownerUser, otherUser);
            Assert.IsTrue(duplicateRecommendation.ContentRecommendationId == recommendation.ContentRecommendationId);

            Assert.IsTrue(dummyCatalogItemId == recommendation.CatalogItemId);
            Assert.IsTrue(recommendation.ContentRecommendationId > 0);
            Assert.IsTrue(recommendation.IsActive);
            Assert.IsFalse(recommendation.IsAccepted);

            IEnumerable<CatalogItemRecommendation> activeRecs = RecommendationsMngr.GetActiveCatalogItemRecommendations(otherUser);
            Assert.IsTrue(activeRecs != null && activeRecs.Count() > 0);
            Assert.IsTrue(activeRecs.Contains(recommendation));

            RecommendationsMngr.MarkRecommendationAsRead(recommendation.ContentRecommendationId);
            ContentRecommendation rec2 = RecommendationsMngr.GetRecommendationById(recommendation.ContentRecommendationId);
            Assert.IsFalse(rec2.IsActive);
            Assert.IsTrue(rec2.ContentRecommendationId == recommendation.ContentRecommendationId);

            // test acceptance& rejection
            Assert.IsFalse(RecommendationsMngr.MarkRecommendationAsAccepted(recommendation.ContentRecommendationId)); // must be active

            VirtualProfile publicProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + "_TestVirtualProfileInvitations4recTest", "description " + publicProfileName, true, true, "http://designyoutrust.com/wp-content/uploads2/bla.jpg", "http://dipity.s3.amazonaws.com/uploads/timelines/8bf8400b92eb09da6f5cf0de27e2503c_1s.png");
            VirtualProfileUsageInvitation rec3 = (VirtualProfileUsageInvitation)RecommendationsMngr.RegisterNewRecommendation(default(int), publicProfile.VirtualProfileId, ownerUser, otherUser);
            Assert.IsFalse(rec3.IsAccepted);
            Assert.IsTrue(rec3.IsActive);

            Assert.IsTrue(RecommendationsMngr.MarkRecommendationAsAccepted(rec3.ContentRecommendationId));
            Assert.IsTrue(rec3.IsAccepted);
            Assert.IsFalse(rec3.IsActive);
            // the same recommendation can be made now, because there is no active rec-n with these params anymore
            VirtualProfileUsageInvitation rec3_new = (VirtualProfileUsageInvitation)RecommendationsMngr.RegisterNewRecommendation(default(int), publicProfile.VirtualProfileId, ownerUser, otherUser);
            Assert.IsTrue(rec3_new.ContentRecommendationId > 0);
            Assert.IsTrue(rec3_new.ContentRecommendationId != rec3.ContentRecommendationId);
            Assert.IsFalse(rec3_new.IsAccepted);
            Assert.IsTrue(rec3_new.IsActive);


            int iDummyCatalogItemId2 = 2;
            CatalogItemRecommendation rec4 = (CatalogItemRecommendation)RecommendationsMngr.RegisterNewRecommendation(iDummyCatalogItemId2, default(int), ownerUser, otherUser);
            Assert.IsTrue(RecommendationsMngr.MarkRecommendationAsRejected(rec4.ContentRecommendationId));
            Assert.IsFalse(rec4.IsAccepted);
            Assert.IsFalse(rec4.IsActive);

        }

        [TestMethod]
        public void TestPublicVirtualProfile()
        {
            User ownerUser = createTestUser();
            User otherUser = createTestUser();

            VirtualProfile publicProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + " TestPublicVirtualProfile", "description " + publicProfileName, true, true, "http://designyoutrust.com/wp-content/uploads2/bla.jpg", "http://dipity.s3.amazonaws.com/uploads/timelines/8bf8400b92eb09da6f5cf0de27e2503c_1s.png");

            TraceLog.Write(LogEventType.Verbose, "TestPublicVpCreation(): test method called.");
            // Create public profile
            Assert.IsNotNull(publicProfile);
            Assert.IsNotNull(publicProfile.VirtualProfileId);
            Assert.IsTrue(publicProfile.IsPublicUsageAllowed);
            Assert.IsTrue(publicProfile.IsModificationAllowed);

            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, publicProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, publicProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, publicProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, publicProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, publicProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, publicProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            // nothing should happen here
            vpManager.GrantAccessRightToLocalUser(ownerUser, publicProfile, VirtualProfilesManager.AccessRightAdministrate);
            vpManager.GrantAccessRightToLocalUser(ownerUser, publicProfile, VirtualProfilesManager.AccessRightSubmitFeedback);
            vpManager.GrantAccessRightToLocalUser(ownerUser, publicProfile, VirtualProfilesManager.AccessRightGetRecommendations);
            vpManager.GrantAccessRightToLocalUser(otherUser, publicProfile, VirtualProfilesManager.AccessRightSubmitFeedback);
            vpManager.GrantAccessRightToLocalUser(otherUser, publicProfile, VirtualProfilesManager.AccessRightGetRecommendations);

            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, publicProfile, VirtualProfilesManager.AccessRightAdministrate));
            vpManager.GrantAccessRightToLocalUser(otherUser, publicProfile, VirtualProfilesManager.AccessRightAdministrate);
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, publicProfile, VirtualProfilesManager.AccessRightAdministrate));
            vpManager.RemoveAccessRight(otherUser, publicProfile, VirtualProfilesManager.AccessRightId_AdministerProfile);
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, publicProfile, VirtualProfilesManager.AccessRightAdministrate));
        }

        [TestMethod]
        public void TestVPPagination()
        {
            User ownerUser = createTestUser();
            VirtualProfile vProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + " keywords test", "private description " + publicProfileName, false, false);

            HashSet<String> keywords = new HashSet<String>();
            String keyword1 = "hello";
            String keyword2 = "pop";
            keywords.Add(keyword1);
            keywords.Add(keyword2);

            vpManager.AddVPKeywords(vProfile, keywords, ownerUser.UserId);

            ReadOnlyDictionary<int, String> vpDictionaryByKeywords = vpManager.FindPublicVpsByKeywords(keywords, 0, 2);
            Assert.IsTrue(vpDictionaryByKeywords != null && vpDictionaryByKeywords.Count > 0);
            int iFoundVpId = vpDictionaryByKeywords.ElementAt(0).Key;
            Assert.IsTrue(iFoundVpId > 0 && vpDictionaryByKeywords.ElementAt(0).Value.Length > 0);

            HashSet<String> realKeywords = vpManager.GetVPKeywords(iFoundVpId);
            Assert.IsTrue(realKeywords.Contains(keyword1) || realKeywords.Contains(keyword2));
        }

        [TestMethod]
        public void TestVPKeywords()
        {
            User ownerUser = createTestUser();

            VirtualProfile vProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + " HELLO keyword!", "private description " + publicProfileName, false, false);
            HashSet<String> keywords = VirtualProfilesManager.GetVPKeywords(vProfile);
            Assert.IsTrue(keywords.Count == 0);

            keywords.Add("hello");
            keywords.Add("music");
            keywords.Add("pop");
            keywords.Add("bla");
            keywords.Add("hello");
            keywords.Add("HELLO");
            
            vpManager.AddVPKeywords(vProfile, keywords);
            HashSet<String> newKeywords = VirtualProfilesManager.GetVPKeywords(vProfile);
            Assert.IsTrue(newKeywords.Count == 4);
            Assert.IsTrue(newKeywords.Contains("pop"));
        }

        [TestMethod]
        public void TestPrivateVirtualProfile()
        {
            User ownerUser = createTestUser();
            User otherUser = createTestUser();

            VirtualProfile privateProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + " TestPrivateVirtualProfile", "private description " + publicProfileName, false, false);

            TraceLog.Write(LogEventType.Verbose, "TestPrivateVpCreation called.");

            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, privateProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, privateProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, privateProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            try
            {
                // rights can't be removed from VP's owner
                vpManager.RemoveAccessRight(ownerUser, privateProfile, VirtualProfilesManager.AccessRightId_GetRecommendations);
                Assert.Fail("Access rights can't be removed from the VP owner.");
            }
            catch (ArgumentException)
            {
                // it's what expected
            }
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, privateProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, privateProfile, VirtualProfilesManager.AccessRightGetRecommendations));


            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            vpManager.GrantAccessRightToLocalUser(otherUser, privateProfile, VirtualProfilesManager.AccessRightGetRecommendations);
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightGetRecommendations));
            vpManager.RemoveAccessRight(otherUser, privateProfile, VirtualProfilesManager.AccessRightId_GetRecommendations);
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            // admin right does not mean everything is allowed (metadata change only)
            vpManager.GrantAccessRightToLocalUser(otherUser, privateProfile, VirtualProfilesManager.AccessRightAdministrate);
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightGetRecommendations));
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, privateProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
        }


        //[TestMethod]
        public void TestCombinedAccessRights()
        {
            User ownerUser = createTestUser();
            User otherUser = createTestUser();
            ExternalSystemUser extUser1 = CreateExternalTestUser("test user");
            ExternalSystemUser extUser2 = CreateExternalTestUser("test user2");

            VirtualProfile vp = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + " TestCombinedAccessRights", "description", false, false);
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, vp, VirtualProfilesManager.AccessRightId_SubmitFeedback));
            vpManager.GrantAccessRightToExternalUser(extUser1, vp, VirtualProfilesManager.AccessRightSubmitFeedback);
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, vp, VirtualProfilesManager.AccessRightId_SubmitFeedback));

            using (UsersManager usersMngr = new UsersManager())
            {
                usersMngr.AssociateExternalAndLocalUser(extUser1, otherUser);

                Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, vp, VirtualProfilesManager.AccessRightId_SubmitFeedback));
                Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(extUser1, vp, VirtualProfilesManager.AccessRightId_SubmitFeedback));
                Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, vp, VirtualProfilesManager.AccessRightId_GetRecommendations));

                vpManager.GrantAccessRightToExternalUser(extUser2, vp, VirtualProfilesManager.AccessRightGetRecommendations);
                Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(extUser2, vp, VirtualProfilesManager.AccessRightId_GetRecommendations));

                usersMngr.AssociateExternalAndLocalUser(extUser2, otherUser);
                Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, vp, VirtualProfilesManager.AccessRightId_GetRecommendations));

                usersMngr.DissociateExternalAndLocalUser(extUser2, otherUser);
                Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, vp, VirtualProfilesManager.AccessRightId_GetRecommendations));
                Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(extUser2, vp, VirtualProfilesManager.AccessRightId_GetRecommendations));

                Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(extUser2, vp, VirtualProfilesManager.AccessRightId_AdministerProfile));
                vpManager.GrantAccessRightToLocalUser(otherUser, vp, VirtualProfilesManager.AccessRightAdministrate);
                usersMngr.AssociateExternalAndLocalUser(extUser2, otherUser);
                Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(extUser2, vp, VirtualProfilesManager.AccessRightId_AdministerProfile));
            }

        }

        [TestMethod]
        public void TestRestrictedVirtualProfile()
        {
            User ownerUser = createTestUser();
            User otherUser = createTestUser();

            VirtualProfile restrictedProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + "_restricted TestRestrictedVirtualProfile", "restricted description " + publicProfileName, true, false);

            TraceLog.Write(LogEventType.Verbose, "TestRestrictedVpCreation called.");

            // owner can do everything
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, restrictedProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, restrictedProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, restrictedProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            // rights can't be removed from VP's owner
            try
            {
                // rights can't be removed from VP's owner
                vpManager.RemoveAccessRight(ownerUser, restrictedProfile, VirtualProfilesManager.AccessRightId_GetRecommendations);
                Assert.Fail("Access rights can't be removed from the VP owner.");
            }
            catch (ArgumentException)
            {
                // it's what expected
            }
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, restrictedProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(ownerUser, restrictedProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightGetRecommendations));

            vpManager.GrantAccessRightToLocalUser(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightSubmitFeedback);
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightAdministrate));
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightSubmitFeedback));
            vpManager.RemoveAccessRight(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightId_SubmitFeedback);
            Assert.IsFalse(VirtualProfilesManager.IsAccessRightGranted(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightSubmitFeedback));

            // get recommendations right can't be removed
            vpManager.RemoveAccessRight(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightId_GetRecommendations);
            Assert.IsTrue(VirtualProfilesManager.IsAccessRightGranted(otherUser, restrictedProfile, VirtualProfilesManager.AccessRightGetRecommendations));

        }

        [TestMethod]
        public void TestVpManagement()
        {
            User ownerUser = createTestUser();
            TraceLog.Write(LogEventType.Verbose, "TestVpManagement called.");

            VirtualProfile publicProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + "TestVpManagement", "description " + publicProfileName, true, true, "http://designyoutrust.com/wp-content/uploads2/bla.jpg", "http://dipity.s3.amazonaws.com/uploads/timelines/8bf8400b92eb09da6f5cf0de27e2503c_1s.png");

            VirtualProfile virtualProfile = vpManager.GetActiveVirtualProfileByName(publicProfile.Name);
            Assert.IsNotNull(virtualProfile);
            Assert.IsTrue(virtualProfile.IsModificationAllowed);
            Assert.IsTrue(virtualProfile.IsPublicUsageAllowed);
        }


        #region Additional test attributes
        //
        // You can use the following additional attributes as you write your tests:
        //
        // Use ClassInitialize to run code before running the first test in the class
        // [ClassInitialize()]
        // public static void MyClassInitialize(TestContext testContext) { }
        //
        // Use ClassCleanup to run code after all tests in a class have run
        // [ClassCleanup()]
        // public static void MyClassCleanup() { }
        //
        // Use TestInitialize to run code before running each test 
        // [TestInitialize()]
        // public void MyTestInitialize() { }
        //
        // Use TestCleanup to run code after each test has run
        // [TestCleanup()]
        // public void MyTestCleanup() { }
        //
        #endregion

        private static ExternalSystemUser CreateExternalTestUser(String firstname)
        {
            ConfigBasedDeployer componentDeployer = new ConfigBasedDeployer(ConfigManager.TestDBConnectionString);
            ExternalSystem extSystem = componentDeployer.DeployOrInitializeExternalSystem("FacebookIntegrator");


            ExternalSystemUser extUser = new ExternalSystemUser();
            extUser.FirstName = firstname;
            extUser.LastName = "TestUser";
            extUser.ExternalSystemId = extSystem.ExternalSystemId;
            extUser.UserExternalIdentifier = "12"; // dummy value
            extUser.Created = DateTime.Now;
            extUser.Modified = DateTime.Now;

            Context.ExternalSystemUsers.InsertOnSubmit(extUser);
            Context.SubmitChanges();

            return extUser;
        }

        private static User createTestUser(String firstname)
        {
            User user = new User();
            user.FirstName = firstname;

            user.ModuleId = getExistingDummyModuleId();
            user.IsDisabled = false;

            user.Created = DateTime.Now;
            user.Modified = DateTime.Now;

            Context.Users.InsertOnSubmit(user);
            Context.SubmitChanges();

            return user;
        }

        private static User createTestUser()
        {
            return createTestUser("test user");
        }

        private static int getExistingDummyModuleId()
        {
            if (tempModuleId == default(int))
            {
                Module module = null;
                try
                {
                    module = Context.Modules.FirstOrDefault<Module>();
                }
                catch (InvalidOperationException)
                {
                    TraceLog.Write(LogEventType.Verbose, "getExistingDummyModuleId() :: No modules exist.");
                }
                if (module == null)
                {
                    tempModuleId = 1; // maybe it would work (if there are no DB constraints)
                }
                else
                {
                    tempModuleId = module.ModuleId;
                }
            }

            return tempModuleId;
        }

        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext
        {
            get
            {
                return testContextInstance;
            }
            set
            {
                testContextInstance = value;
            }
        }

        [TestMethod]
        public void TestSubscriptions()
        {
            User ownerUser = createTestUser();

            VirtualProfile privateProfile = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + "_private TestSubscriptions", "private description " + publicProfileName, false, false);

            User localTestUser = createTestUser("local user to test subscriptions");
            User localTestOwner = createTestUser("local owner to test subscriptions");
            VirtualProfile localVProfile = vpManager.CreateVirtualProfile(localTestOwner.UserId, publicProfileName + "_local TestSubscriptions", "description " + publicProfileName + "_local", false, false);

            vpManager.SubscribeToVirtualProfile(privateProfile.VirtualProfileId, localTestUser.UserId);
            Assert.IsTrue(vpManager.GetVirtualProfileUsersCount(privateProfile.VirtualProfileId) == 2);
        }

        [TestMethod]
        public void TestVPNameUniqueness()
        {
            User ownerUser = createTestUser();
            VirtualProfile vp = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + "_local TestVPNameUniqueness", "vp description ", true, false);
            String vpName = vp.Name;
            Assert.IsTrue(vpManager.IsVPNameTaken(vp.Name));
            vpManager.DeleteVirtualProfile(vp.VirtualProfileId);

            Assert.IsFalse(vpManager.IsVPNameTaken(vpName));
        }

        [TestMethod]
        public void TestVPRatingsRetrieval(){
            User ownerUser = createTestUser();
            VirtualProfile vp = vpManager.CreateVirtualProfile(ownerUser.UserId, publicProfileName + "_TestVPRatingsRetrieval", "vp description ", true, false);
            IEnumerable<RatingAction> ratings = vpManager.GetVPRatings(vp.VirtualProfileId, 0, 100);
            Assert.IsTrue(ratings == null || ratings.Count() == 0);
        }

        [ClassCleanup]
        public static void TearDown()
        {
            TraceLog.Write(LogEventType.Verbose, "TearDown called.");
        }
    }
}
