﻿using System;
using System.Collections.Generic;
using System.Data.SqlClient;
using System.Globalization;
using System.Linq;
using EMIC.Design;
using ExternalSystemIntegration;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.ActionLogging;
using MyMediaProject.RecommenderSystem.Core.Config;
using MyMediaProject.RecommenderSystem.Core.Databases;

namespace VirtualProfiles
{
    public partial class VirtualProfilesManager
    {
        /// <summary>
        /// Retrieves all keywords associated with the given VP
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <returns>HashSet with all words associated with the specified virtual profile</returns>
        public HashSet<String> GetVPKeywords(int virtualProfileId)
        {
            return GetVPKeywords(GetActiveVirtualProfileById(virtualProfileId));
        }

        /// <summary>
        /// Retrieves all keywords associated with the given VP
        /// </summary>
        /// <param name="vp"></param>
        /// <returns>HashSet with all words associated with the specified virtual profile</returns>
        public static HashSet<String> GetVPKeywords(VirtualProfile vp)
        {
            HashSet<String> hsResult = new HashSet<String>();
            if (vp == null || vp.VPKeywords == null || vp.VPKeywords.Count == 0)
            {
                return hsResult;
            }

            for (int i = 0; i < vp.VPKeywords.Count; i++)
            {
                hsResult.Add(vp.VPKeywords.ElementAt(i).Keyword.Word);
            }

            return hsResult;
        }

        /// <summary>
        /// Sets or updates keywords associated with the given virtual profile.
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="keywords"></param>
        public void AddVPKeywords(int virtualProfileId, HashSet<String> keywords)
        {
            AddVPKeywords(GetActiveVirtualProfileById(virtualProfileId), keywords);
        }

        /// <summary>
        /// Sets or updates keywords associated with the given virtual profile.
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="keywords"></param>
        /// <param name="localUserId">local userId that assigns the keyword to the VP</param>
        public void AddVPKeywords(int virtualProfileId, HashSet<String> keywords, int localUserId)
        {
            AddVPKeywords(GetActiveVirtualProfileById(virtualProfileId), keywords, localUserId);
        }

        /// <summary>
        /// Sets or updates keywords associated with the given virtual profile.
        /// </summary>
        /// <param name="vp"></param>
        /// <param name="keywords"></param>
        public void AddVPKeywords(VirtualProfile vp, HashSet<String> keywords)
        {
            AddVPKeywords(vp, keywords, default(int));
        }

        /// <summary>
        /// Adds new keywords associated with the given virtual profile.
        /// </summary>
        /// <param name="vp"></param>
        /// <param name="keywords"></param>
        /// <param name="localUserId">local userId that assigns the keyword to the VP</param>
        public void AddVPKeywords(VirtualProfile vp, HashSet<String> newKeywords, int localUserId)
        {
            HashSet<String> existingKeywords = GetVPKeywords(vp);
            CatalogItemStoreDataContext itemContext = new CatalogItemStoreDataContext(Context.Connection.ConnectionString);
            using (itemContext)
            {
                for (int i = 0; i < newKeywords.Count; i++)
                {
                    String word = newKeywords.ElementAt(i).Trim();
                    if (existingKeywords.Contains(word))
                    {
                        // this keyword is assigned to the specified VP already, don't add it again
                        continue;
                    }

                    if (word.Length > MAX_KEYWORD_LENGTH)
                    {
                        String truncatedWord = word.Substring(0, MAX_KEYWORD_LENGTH);
                        TraceLog.Write(LogEventType.Verbose, "Truncating a virtual profile's(vpId='{0}') keyword from '{1}' to '{2}'.", new object[] { vp.VirtualProfileId, word, truncatedWord });
                        word = truncatedWord;
                    }
                    Keyword keyword = itemContext.CreateKeyword(word, CultureInfo.CurrentCulture);
                    // if this is a new keyword, it needs to be inserted to the DB first
                    if (keyword.KeywordId == default(int))
                    {
                        itemContext.SubmitChanges();
                    }

                    VPKeyword vpKeyword = new VPKeyword();
                    vpKeyword.KeywordId = keyword.KeywordId;
                    vpKeyword.VirtualProfileId = vp.VirtualProfileId;
                    vpKeyword.Created = DateTime.Now;
                    if (localUserId > 0)
                    {
                        vpKeyword.AssignedByUserId = localUserId;
                    }
                    else
                    {
                        vpKeyword.AssignedByUser = null;
                    }

                    vp.VPKeywords.Add(vpKeyword);
                }

                Context.SubmitChanges();
            }
        }

        /// <summary>
        /// Removes specified VP keyword
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="keyword"></param>
        /// <param name="userId"></param>
        public void RemoveVPKeyword(int virtualProfileId, String keyword, int userId)
        {
            RemoveVPKeyword(GetActiveVirtualProfileById(virtualProfileId), keyword, userId);
        }

        /// <summary>
        /// Removes specified VP keyword
        /// </summary>
        /// <param name="vp"></param>
        /// <param name="keyword"></param>
        /// <param name="userId"></param>
        public void RemoveVPKeyword(VirtualProfile vp, String keyword, int userId)
        {
            if (vp == null)
            {
                throw new ArgumentException("Specified VP does not exist");
            }
            if (keyword == null)
            {
                throw new ArgumentException("No keyword specified.");
            }

            keyword = keyword.Trim();

            VPKeyword vpKeyword = vp.VPKeywords.FirstOrDefault<VPKeyword>(k => k.Keyword.Word == keyword);

            if (vpKeyword == null)
            {
                throw new ArgumentException("The specified keyword does not exist.");
            }

            Context.VPKeywords.DeleteOnSubmit(vpKeyword);
            Context.SubmitChanges();

            User user = null;
            if (userId > 0)
            {
                UsersManager usersMngr = new UsersManager();
                user = usersMngr.GetLocalUserById(userId);
            }

            ActionLogger.LogVirtualProfileAction("VP/Keywords/Removed", keyword, vp, user);
        }


        /// <summary>
        /// Delete all associated keywords of the specified VP
        /// </summary>
        /// <param name="vp"></param>
        public void ClearVPKeywords(int virtualProfileId, int userId)
        {
            ClearVPKeywords(GetActiveVirtualProfileById(virtualProfileId), userId);
        }

        /// <summary>
        /// Delete all associated keywords of the specified VP
        /// </summary>
        /// <param name="vp"></param>
        public void ClearVPKeywords(VirtualProfile vp, int userId)
        {
            int keywordsCount = vp.VPKeywords.Count;
            if (keywordsCount == 0)
            {
                return;
            }
            Context.VPKeywords.DeleteAllOnSubmit(vp.VPKeywords);
            Context.SubmitChanges();

            User user = null;
            if (userId > 0)
            {
                UsersManager usersMngr = new UsersManager();
                user = usersMngr.GetLocalUserById(userId);
            }

            ActionLogger.LogVirtualProfileAction("VP/Keywords/DeletedAll", keywordsCount.ToString(CultureInfo.CurrentCulture), vp, user);
        }

        /// <summary>
        /// Retrieves Ids of virtual profiles by the keywords specified
        /// </summary>
        /// <param name="keywords"></param>
        /// <param name="elementsToSkip"></param>
        /// <param name="maximalElementsCount"></param>
        /// <returns>ReadOnlyDictionary structure holding found public VPs (where keys are the Ids of the found VPs, and names of the VPs are the values).</returns>
        public ReadOnlyDictionary<int, String> FindPublicVpsByKeywords(HashSet<String> keywords, int elementsToSkip, int maximalElementsCount)
        {
            List<String> keywordsList = keywords.ToList<String>();

            IEnumerable<VirtualProfile> results = from vp in Context.VirtualProfiles
                                              from vpk in Context.VPKeywords
                                              where vp.IsDisabled == false
                                              && vpk.VirtualProfileId == vp.VirtualProfileId
                                              && keywordsList.Contains(vpk.Keyword.Word)
                                              select vp;
            List<VirtualProfile> vpList = results.OrderByDescending(vp => vp.Name).ThenByDescending(vp => vp.VirtualProfileId).Skip(elementsToSkip).Take(maximalElementsCount).Distinct<VirtualProfile>().ToList();

            Dictionary<int, String> vpIdsAndNamesDict = new Dictionary<int, String>();
            foreach (VirtualProfile vp in vpList)
            {
                vpIdsAndNamesDict.Add(vp.VirtualProfileId, vp.Name);
            }

            return new ReadOnlyDictionary<int, String>(vpIdsAndNamesDict);
        }
    }
}
