﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Linq;
using EMIC.Design;
using MyMediaProject.RecommenderSystem.Core;
using MyMediaProject.RecommenderSystem.Core.Config;

namespace VirtualProfiles
{
    /// <summary>
    /// Class provides the functionality to access and manage virtual profiles.
    /// It is the responsibility of the calling client to ensure that the user has right to 
    /// perform certain operation on the virtual profile
    /// </summary>
    public partial class VirtualProfilesManager : IDisposable
    {
        private MyMediaDataContext Context;
        private bool ownsDataContext;
        private int MAX_KEYWORD_LENGTH = 256;

        /// <summary>
        /// Default constructor
        /// </summary>
        public VirtualProfilesManager()
        {
            String dbConnectionString = ConfigManager.DBConnectionString;
            TraceLog.Write(LogEventType.Information, "VirtualProfilesManager() : using dbConnectionString='{0}'.", new object[] { dbConnectionString });
            this.Context = new MyMediaDataContext(dbConnectionString);
            ownsDataContext = true;

            // initialize access rights (ensure they are registered in the DB & initialize properties)
            InitializeAccessRights();
        }

        /// <summary>
        /// Constructor that uses given DataContext. This DataContext will not be disposed by this class.
        /// </summary>
        /// <param name="context">Initialized data context that this class will use</param>
        public VirtualProfilesManager(MyMediaDataContext context)
        {
            this.Context = context;
        }


        /// <summary>
        /// Retrieves active(not deleted) virtual profile by its ID
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <returns>Virtual Profile(if exists) or null</returns>
        public VirtualProfile GetActiveVirtualProfileById(int virtualProfileId)
        {
            VirtualProfile vp = null;
            vp = Context.VirtualProfiles.SingleOrDefault<VirtualProfile>(v => v.VirtualProfileId == virtualProfileId && v.IsDisabled == false);

            if (vp == null)
            {
                TraceLog.Write(LogEventType.Verbose, "GetActiveVirtualProfileById('{0}') not found.", new object[] { virtualProfileId });
            }
            else
            {
                TraceLog.Write(LogEventType.Verbose, "GetActiveVirtualProfileById('{0}') found: '{1}'.", new object[] { virtualProfileId, vp.Name });
            }
            return vp;
        }

        /// <summary>
        /// Retrieves active(not deleted) virtual profile
        /// </summary>
        /// <param name="name"></param>
        /// <returns>Virtual Profile(if exists) or null</returns>
        public VirtualProfile GetActiveVirtualProfileByName(String name)
        {
            return Context.VirtualProfiles.SingleOrDefault<VirtualProfile>(v => v.Name == name && v.IsDisabled == false);
        }

        /// <summary>
        /// Checks if active(not deleted) virtual profile with the given name exists
        /// </summary>
        /// <returns>true if active virtual profile with the specified name exists already, false otherwise</returns>
        public bool IsVPNameTaken(String name)
        {
            name = name.Trim();
            return Context.VirtualProfiles.Any<VirtualProfile>(v => v.Name == name && v.IsDisabled == false);
        }

        /// <summary>
        /// Registers user as virtual profile subscriber
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="userId"></param>
        public bool SubscribeToVirtualProfile(int virtualProfileId, int userId)
        {
            // if the user has subscribed to VP already, do nothing
            if (Context.VirtualProfileUsers.Any<VirtualProfileUser>(u => u.UserId == userId && u.VirtualProfileId == virtualProfileId))
            {
                return false;
            }

            VirtualProfile vp = GetActiveVirtualProfileById(virtualProfileId);
            if (vp == null)
            {
                throw new ArgumentException("Active VP with id='" + virtualProfileId + "' has not been found.");
            }

            // if the user is VP's owner, s/he can't subscribe(because s/he is always subscribed to this VP)
            if (vp.CreatorUserId == userId)
            {
                return false;
            }

            VirtualProfileUser vpUser = new VirtualProfileUser();
            vpUser.VirtualProfileId = virtualProfileId;
            vpUser.UserId = userId;
            vpUser.Created = DateTime.Now;
            vpUser.Modified = DateTime.Now;

            TraceLog.Write(LogEventType.Information, "About to register VP User: vp={0}, user= {1}", new object[] { virtualProfileId, userId });

            Context.VirtualProfileUsers.InsertOnSubmit(vpUser);
            Context.SubmitChanges();
            TraceLog.Write(LogEventType.Information, "User (userId=" + userId + ") subscribed to the VP(virtualProfileId=" + virtualProfileId + ").");
            return true;
        }

        /// <summary>
        /// Retrieves all users(user objects) of the specified virtual profile, including its creator
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <returns></returns>
        public IEnumerable<User> GetVirtualProfileSubscribers(int virtualProfileId)
        {
            IEnumerable<User> users =
                ((from vpUser in Context.VirtualProfileUsers
                where vpUser.VirtualProfileId == virtualProfileId
                  select vpUser.User).Union(from vp in Context.VirtualProfiles
                                            where vp.VirtualProfileId == virtualProfileId
                                            select vp.CreatorUser)).Distinct();

            return users;
        }

        /// <summary>
        /// Calculates the number of subscribers of the VP
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <returns></returns>
        public int GetVirtualProfileUsersCount(int virtualProfileId)
        {
            // VP's creator's "subscription" to the VP is not registered in the 'VirtualProfileUsers' table
            return Context.VirtualProfileUsers.Count(u => u.VirtualProfileId == virtualProfileId) + 1;
        }

        /// <summary>
        /// Terminate user's subscription to the virtual profile
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="userId"></param>
        /// <returns>true if unsubscribed successfully; false otherwise.</returns>
        public bool UnsubscribeFromVirtualProfile(int virtualProfileId, int userId)
        {
            VirtualProfile vp = GetActiveVirtualProfileById(virtualProfileId);
            if (vp == null)
            {
                throw new ArgumentException("Active VP with id='" + virtualProfileId + "' has not been found.");
            }
            if (vp.CreatorUserId == userId)
            {
                throw new ArgumentException("User " + userId + " is creator of the VP with id='" + virtualProfileId + "' and can't be unsubscribed.");
            }

            VirtualProfileUser vpUser = Context.VirtualProfileUsers.FirstOrDefault<VirtualProfileUser>(u => u.UserId == userId && u.VirtualProfileId == virtualProfileId);

            // if the user is not subscribed to VP, do nothing
            if (vpUser == null)
            {
                return false;
            }

            Context.VirtualProfileUsers.DeleteOnSubmit(vpUser);
            Context.SubmitChanges();
            TraceLog.Write(LogEventType.Information, "User's(userId=" + userId + ") subscription to the VP(virtualProfileId=" + virtualProfileId + ") has been terminated.");
            return true;
        }

        /// <summary>
        /// Creates new virtual profile by the given fields and returns it
        /// Throws ArgumentException in case an active virtual profile with the given name exists already
        /// </summary>
        /// <param name="userId"></param>
        /// <param name="name"></param>
        /// <param name="description"></param>
        /// <param name="isPublicUsageAllowed"></param>
        /// <param name="isModificationAllowed"></param>
        /// <returns>Initialized Virtual Profile object</returns>
        public VirtualProfile CreateVirtualProfile(int userId, String name, String description, bool isPublicUsageAllowed, bool isModificationAllowed)
        {
            return CreateVirtualProfile(userId, name, description, isPublicUsageAllowed, isModificationAllowed, null, null);
        }


        /// <summary>
        /// Creates new virtual profile by the given fields and returns it
        /// Throws ArgumentException in case an active virtual profile with the given name exists already
        /// </summary>
        /// <param name="userId"></param>
        /// <param name="name"></param>
        /// <param name="description"></param>
        /// <param name="isPublicUsageAllowed"></param>
        /// <param name="isModificationAllowed"></param>
        /// <returns>Initialized Virtual Profile object</returns>
        public VirtualProfile CreateVirtualProfile(int userId, String name, String description, bool isPublicUsageAllowed, bool isModificationAllowed, String urlToBigLogo, String urlToSmallLogo)
        {
            name = name.Trim();
            if (IsVPNameTaken(name))
            {
                throw new ArgumentException("A virtual profile with the name '" + name + "' exists already.");
            }
            VirtualProfile virtualProfile = new VirtualProfile();
            virtualProfile.CreatorUserId = userId;
            virtualProfile.Name = name;
            virtualProfile.Description = description;
            virtualProfile.IsPublicUsageAllowed = isPublicUsageAllowed;
            virtualProfile.IsModificationAllowed = isModificationAllowed;
            virtualProfile.Created = DateTime.Now;
            virtualProfile.Modified = DateTime.Now;
            virtualProfile.IsDisabled = false;
            virtualProfile.UrlToSmallLogo = urlToSmallLogo;
            virtualProfile.UrlToBigLogo = urlToBigLogo;

            Context.VirtualProfiles.InsertOnSubmit(virtualProfile);
            Context.SubmitChanges();
            TraceLog.Write(LogEventType.Information, "New Virtual Profile '{0}' with id='{1}' has been created.", new object[] { virtualProfile.Name, virtualProfile.VirtualProfileId });

            // register the creator as a VP 'user' / 'subscriber'
            SubscribeToVirtualProfile(virtualProfile.VirtualProfileId, userId);

            return virtualProfile;
        }


        /// <summary>
        /// Updates the description of the virtual profile
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="newDescription"></param>
        public void UpdateVirtualProfileDescription(int virtualProfileId, String newDescription)
        {
            VirtualProfile virtualProfile = GetActiveVirtualProfileById(virtualProfileId);
            if (virtualProfile == null)
            {
                throw new PreconditionViolationException("An active VP with the Id= '" + virtualProfileId + "' could not be found.");
            }

            // if the description is not different, do nothing
            if (String.Compare(newDescription, virtualProfile.Description, StringComparison.Ordinal) == 0)
            {
                return;
            }
            virtualProfile.Description = newDescription;
            virtualProfile.Modified = DateTime.Now;
            Context.SubmitChanges();
        }

        /// <summary>
        /// Deletes the specified virtual profile (sets IsDisabled flag to 'true')
        /// </summary>
        /// <param name="virtualProfile"></param>
        public void DeleteVirtualProfile(int virtualProfileId)
        {
            VirtualProfile virtualProfile = GetActiveVirtualProfileById(virtualProfileId);
            if (virtualProfile == null)
            {
                return;
            }

            // TraceLog.Write(LogEventType.Information, "Virtual profile '{0}' with id='{1}' has been deleted(disabled).", new object[] { virtualProfile.Name, virtualProfile.VirtualProfileId });
            virtualProfile.IsDisabled = true;
            virtualProfile.Modified = DateTime.Now;
            Context.SubmitChanges();
        }

        /// <summary>
        /// Retrieves list of virtual profiles that user subscribed to, or created
        /// </summary>
        /// <param name="userId"></param>
        /// <returns></returns>
        public IEnumerable<VirtualProfile> GetUsersVirtualProfiles(int userId)
        {
            IEnumerable<VirtualProfile> userVirtualProfiles =
                ((from vpUser in Context.VirtualProfileUsers
                  where vpUser.UserId == userId
                  select vpUser.VirtualProfile).Union(from vp in Context.VirtualProfiles
                                                      where vp.CreatorUserId == userId
                                                      select vp)).Distinct();

            TraceLog.Write(LogEventType.Information, "GetUsersVirtualProfiles('{0}').Count = '{1}'.", new object[] { userId, userVirtualProfiles.Count()});
            return userVirtualProfiles;
        }

        /// <summary>
        /// Returns ALL ratings given on behalf of the virtual profile
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <returns></returns>
        public IEnumerable<RatingAction> GetVPRatings(int virtualProfileId)
        {
            return 
                from ratingAction in Context.UserActions
                where ratingAction is RatingAction
                && ratingAction.VirtualProfileId == virtualProfileId
                select ratingAction as RatingAction;
        }

        /// <summary>
        /// Returns ratings given on behalf of the virtual profile
        /// </summary>
        /// <param name="virtualProfileId"></param>
        /// <param name="skip"></param>
        /// <param name="take"></param>
        /// <returns></returns>
        public IEnumerable<RatingAction> GetVPRatings(int virtualProfileId, int skip, int take)
        {
            IEnumerable<RatingAction> ratings = GetVPRatings(virtualProfileId);

            ratings = ratings.OrderByDescending(ra => ra.Timestamp).ThenByDescending(ra => ra.Key).Skip(skip).Take(take).ToList();

            return ratings;
        }




        #region IDisposable Members

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (this.ownsDataContext && disposing)
            {
                this.Context.Connection.Dispose();
                this.Context.Dispose();
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

    }
}
